/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*-
 *
 * Copyright (c) 1997 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Header: /nfs/jade/vint/CVSROOT/ns-2/mac/mac-802_11.h,v 1.20 2002/03/14 01:12:53 haldar Exp $
 *
 * Ported from CMU/Monarch's code, nov'98 -Padma.
 * wireless-mac-802_11.h
 */
// HCF v3.0
//------------------------------------------------------------------------------------------------------------------------------//EDCFpatch

#ifndef mac_80211_h
#define mac_80211_h

#include "mac-timers.h"
#include "marshall.h"
#include "drop-tail.h"

#define GET_ETHER_TYPE(x)              GET2BYTE((x))
#define SET_ETHER_TYPE(x,y)            {u_int16_t t = (y); STORE2BYTE(x,&t);}

/* ======================================================================
   Frame Formats
   ====================================================================== */

#define	MAC_ProtocolVersion			0x00

#define MAC_Type_Management			0x00
#define MAC_Type_Control			0x01
#define MAC_Type_Data				0x02
#define MAC_Type_Reserved			0x03

// ---------- HCFstart
#define MAC_Subtype_Beacon			0x08
#define MAC_Subtype_RR				0x04
#define MAC_Subtype_CC				0x05
// ---------- HCFend
#define MAC_Subtype_RTS				0x0B
#define MAC_Subtype_CTS				0x0C
#define MAC_Subtype_ACK				0x0D
#define MAC_Subtype_Data			0x00
// ---------- HCFstart
#define MAC_Subtype_CFEnd			0x0E
#define MAC_Subtype_QoS_Data			0x08
#define MAC_Subtype_QoS_Data_CFAck		0x09
#define MAC_Subtype_QoS_Data_CFPoll		0x0A
#define MAC_Subtype_QoS_Data_CFAck_CFPoll	0x0B
#define MAC_Subtype_QoS_Null			0x0C
#define MAC_Subtype_QoS_CFAck			0x0D
#define MAC_Subtype_QoS_CFPoll			0x0E
#define MAC_Subtype_QoS_CFAck_CFPoll		0x0F
// ---------- HCFend


struct frame_control {
	u_char		fc_subtype		: 4;
	u_char		fc_type			: 2;
	u_char		fc_protocol_version	: 2;

	u_char		fc_order		: 1;
	u_char		fc_wep			: 1;
	u_char		fc_more_data		: 1;
	u_char		fc_pwr_mgt		: 1;
	u_char		fc_retry		: 1;
	u_char		fc_more_frag		: 1;
	u_char		fc_from_ds		: 1;
	u_char		fc_to_ds		: 1;
};

// ---------- HCFstart
struct cf_param_set {
	u_int8_t	elementID;		// must be 4
	u_int8_t	length;			// must be 8 (octects)
	u_int8_t	CFPCount;		// 0: CFP starts now
	u_int8_t	CFPPeriod;		// CFPRate
	u_int16_t	CFPMaxDuration;
	u_int16_t	CFPDurRemaining;
};

struct tim {
	u_int8_t	elementID;		// must be 5
	u_int8_t	length;			// must be 4 (octects)
	u_int8_t	DTIMCount;		// 0: this is a DTIM
	u_int8_t	DTIMPeriod;		// 1: all TIMs are DTIMs
};

struct beacon_frame {
	struct frame_control	bf_fc;
	u_int16_t		bf_duration;
	u_char			bf_da[ETHER_ADDR_LEN];
	u_char			bf_sa[ETHER_ADDR_LEN];
	u_char			bf_bssid[ETHER_ADDR_LEN];
	u_int16_t		bf_scontrol;
	struct cf_param_set	bf_cf_pset;
	struct tim		bf_tim;
};

struct cc_frame {
	struct frame_control	cc_fc;
	u_int16_t		cc_duration;
	u_char			cc_ra[ETHER_ADDR_LEN];
	u_char			cc_bssid[ETHER_ADDR_LEN];
	u_char			cc_fcs[ETHER_FCS_LEN];
	u_int8_t		cc_cci_length;		// Nccop
};

struct rr_frame {
	struct frame_control	rr_fc;
	u_int16_t		rr_duration;
	u_char			rr_ra[ETHER_ADDR_LEN];
	u_char			rr_bssid[ETHER_ADDR_LEN];
	u_char			rr_fcs[ETHER_FCS_LEN];
	u_int16_t		rr_aid;
	u_int16_t		rr_txop_duration;
};
// ---------- HCFend

struct rts_frame {
	struct frame_control	rf_fc;
	u_int16_t		rf_duration;
	u_char			rf_ra[ETHER_ADDR_LEN];
	u_char			rf_ta[ETHER_ADDR_LEN];
	u_char			rf_fcs[ETHER_FCS_LEN];
};

struct cts_frame {
	struct frame_control	cf_fc;
	u_int16_t		cf_duration;
	u_char			cf_ra[ETHER_ADDR_LEN];
	u_char			cf_fcs[ETHER_FCS_LEN];
};

struct ack_frame {
	struct frame_control	af_fc;
	u_int16_t		af_duration;
	u_char			af_ra[ETHER_ADDR_LEN];
	u_char			af_fcs[ETHER_FCS_LEN];
};

// ---------- HCFstart
struct cfend_frame {
	struct frame_control	ef_fc;
	u_int16_t		ef_duration;			// must be 0 (TU)
	u_char			ef_ra[ETHER_ADDR_LEN];		// MAC broadcast
	u_char			ef_bssid[ETHER_ADDR_LEN];	// PC MAC address
	u_char			ef_fcs[ETHER_FCS_LEN];
};

struct null_frame {
	struct frame_control	nf_fc;
	u_int16_t		nf_duration;
	u_char			nf_ra[ETHER_ADDR_LEN];
	u_char			nf_fcs[ETHER_FCS_LEN];
};

struct qos_cfpoll_frame {
	struct frame_control	pf_fc;
	u_int16_t		pf_duration;
	u_char			pf_ra[ETHER_ADDR_LEN];
	u_char			rr_bssid[ETHER_ADDR_LEN];
	u_char			rr_fcs[ETHER_FCS_LEN];
	u_int16_t		pf_txop_duration;
};
// ---------- HCFend

// This header does not have its header access function because it shares
// the same header space with hdr_mac.
struct hdr_mac802_11 {
	struct frame_control	dh_fc;
	u_int16_t		dh_duration;
	u_char			dh_da[ETHER_ADDR_LEN];
	u_char			dh_sa[ETHER_ADDR_LEN];
	u_char			dh_bssid[ETHER_ADDR_LEN];
	u_int16_t		dh_scontrol;
	u_char			dh_body[0];
};

/* ======================================================================
   Definitions
   ====================================================================== */

#define PLCP_HDR_LEN				\
  	((phymib_->PreambleLength >> 3) +	\
 	 (phymib_->PLCPHeaderLength >> 3))

#define PLCP_HDR_LEN2				\
  	((DSSS_PreambleLength >> 3) +		\
 	 (DSSS_PLCPHeaderLength >> 3))

#define ETHER_HDR_LEN11				\
	(PLCP_HDR_LEN +				\
	 sizeof(struct hdr_mac802_11) +		\
	 ETHER_FCS_LEN)

// XXX v2.0Start
#define ETHER_HDR_LEN11_2			\
	(PLCP_HDR_LEN2 +			\
	 sizeof(struct hdr_mac802_11) +		\
	 ETHER_FCS_LEN)
// XXX v2.0End

// ---------- HCFstart
#define ETHER_BCN_LEN				\
        (PLCP_HDR_LEN +				\
	 sizeof(struct beacon_frame))

#define ETHER_BCN_LEN2				\
        (PLCP_HDR_LEN2 +			\
	 sizeof(struct beacon_frame))

#define ETHER_CC_LEN				\
        (PLCP_HDR_LEN +				\
	 sizeof(struct cc_frame))

#define ETHER_CC_LEN2				\
        (PLCP_HDR_LEN2 +			\
	 sizeof(struct cc_frame))

#define ETHER_RR_LEN				\
        (PLCP_HDR_LEN +				\
	 sizeof(struct rr_frame))

#define ETHER_RR_LEN2				\
        (PLCP_HDR_LEN2 +			\
	 sizeof(struct rr_frame))
// ---------- HCFend

#define ETHER_RTS_LEN				\
        (PLCP_HDR_LEN +				\
	 sizeof(struct rts_frame))

#define ETHER_RTS_LEN2				\
        (PLCP_HDR_LEN2 +			\
	 sizeof(struct rts_frame))

#define ETHER_CTS_LEN				\
         (PLCP_HDR_LEN +			\
         sizeof(struct cts_frame))

#define ETHER_CTS_LEN2				\
         (PLCP_HDR_LEN2 +			\
         sizeof(struct cts_frame))

#define ETHER_ACK_LEN				\
         (PLCP_HDR_LEN +			\
	 sizeof(struct ack_frame))

#define ETHER_ACK_LEN2				\
         (PLCP_HDR_LEN2 +			\
	 sizeof(struct ack_frame))

// ---------- HCFstart
#define ETHER_CFEND_LEN				\
        (PLCP_HDR_LEN +				\
	 sizeof(struct cfend_frame))

#define ETHER_CFEND_LEN2			\
        (PLCP_HDR_LEN2 +			\
	 sizeof(struct cfend_frame))

#define ETHER_NULL_LEN				\
        (PLCP_HDR_LEN +				\
	 sizeof(struct null_frame))

#define ETHER_NULL_LEN2			\
        (PLCP_HDR_LEN2 +			\
	 sizeof(struct null_frame))

#define ETHER_QOS_CFPOLL_LEN			\
        (PLCP_HDR_LEN +				\
	 sizeof(struct qos_cfpoll_frame))

#define ETHER_QOS_CFPOLL_LEN2			\
        (PLCP_HDR_LEN2 +			\
	 sizeof(struct qos_cfpoll_frame))
// ---------- HCFend

/*
 * IEEE 802.11 Spec, section 15.3.2
 *	- default values for the DSSS PHY MIB
 */

#define DSSS_CWMin			31              // Replaced by Mac802_11e::CWmin in order to allow different Access Categories
#define DSSS_CWMax			1023            // Replaced by Mac802_11e::CWmax in order to allow different Access Categories
#define DSSS_SlotTime			0.000020        // secs
#define	DSSS_CCATime			0.000015        // secs
#define DSSS_RxTxTurnaroundTime		0.000005        // secs
#define DSSS_SIFSTime			0.000010        // secs
#define DSSS_PreambleLength		144             // bits
#define DSSS_PLCPHeaderLength		48              // bits
#define DSSS_PLCPDataRate		1.0e6           // bps

/* Must account for propagation delays added by the channel model when
 * calculating tx timeouts (as set in tcl/lan/ns-mac.tcl).
 *   -- Gavin Holland, March 2002
 */

#define DSSS_MaxPropagationDelay        0.000002        // secs

class PHY_MIB {
public:
	u_int32_t	CWMin;         // Replaced by Mac802_11e::CWmin in order to allow different Access Categories
	u_int32_t	CWMax;         // Replaced by Mac802_11e::CWmax in order to allow different Access Categories
	double		SlotTime;
	double		CCATime;
	double		RxTxTurnaroundTime;
	double		SIFSTime;
	u_int32_t	PreambleLength;
	u_int32_t	PLCPHeaderLength;
	double		PLCPDataRate;
};

/*
 * IEEE 802.11 Spec, section 11.4.4.2
 *      - default values for the MAC Attributes
 */

#define MAC_RTSThreshold		40000     // no RTS-CTS!!!
#define MAC_ShortRetryLimit		7	// Replaced by Mac802_11e::ShortRL in order to allow different Access Categories
#define MAC_LongRetryLimit		4	// Replaced by Mac802_11e::LongRL in order to allow different Access Categories
#define MAC_FragmentationThreshold	2346	// bytes
#define MAC_MaxTransmitMSDULifetime	512	// time units
#define MAC_MaxReceiveLifetime		512	// time units
// ---------- HCFstart
#define MAC_MaxMPDUSize			2312	// bytes
#define MAC_MinMPDUSize			34	// bytes

// IEEE 802.11 Spec, section 9.3.1
// CF Parameter Set for Point Coordinator
#define	MAC_TIM				0.01024	// time units
// #define MAC_CFPRate			400		// TIMs == DTIMs
#define MAC_CFPMaxDuration		0.39		// time units
// #define MAC_CFPDurRemaining		0.39		// time units

// max number of Access Categories within the 802.11e node
#define MAX_TC				8
// max number of [node address/Traffic Category] elements in the HC Polling Lists
#define MAX_CF_POLL			640
#define MAX_RR_POLL			8

// XXX v2.0Start
// max number of stations in the network for each Access Point
#define MAX_STA				64
// XXX v2.0End
// ---------- HCFend

class MAC_MIB {
public:
	u_int32_t	RTSThreshold;
	u_int32_t	ShortRetryLimit;        // Replaced by Mac802_11e::ShortRL in order to allow different Access Categories
	u_int32_t	LongRetryLimit;         // Replaced by Mac802_11e::LongRL in order to allow different Access Categories
	u_int32_t	FragmentationThreshold;
	u_int32_t	MaxTransmitMSDULifetime;
	u_int32_t	MaxReceiveLifetime;

	u_int32_t	TransmittedFragmentCount;
	u_int32_t	MulticastTransmittedFrameCount;
	u_int32_t	FailedCount;
	u_int32_t	RetryCount;
	u_int32_t	MultipleRetryCount;
	u_int32_t	FrameDuplicateCount;
	u_int32_t	RTSSuccessCount;
	u_int32_t	RTSFailureCount;
	u_int32_t	ACKFailureCount;
	u_int32_t	ReceivedFragmentCount;
	u_int32_t	MulticastReceivedFrameCount;
	u_int32_t	FCSErrorCount;
};


/* ======================================================================
   The following destination class is used for duplicate detection.
   ====================================================================== */

class Host {
public:
	LIST_ENTRY(Host) link;
	u_int32_t	index;
	u_int32_t	seqno;
};

/* ======================================================================
   The basic 802.11 MAC class.
   ====================================================================== */

// XXX v3.0Start
// Element of TSPEC management list
struct TSPECElem {
   // TSPEC fields
   int         NominalMSDUSize;           // bytes
   int         MaximumMSDUSize;           // bytes
   double      MaximumServiceInterval;    // sec
   double      DelayBound;                // sec
   int         MinimumDataRate;           // bit/sec
   int         MeanDataRate;              // bit/sec   (es. 128 Kbps = (128*1024) = 131072)
   int         PeakDataRate;              // bit/sec

   // Not really belonging to TSPEC
   double      CFP_TXOPLmt;            // Last TXOP duration used to poll the station
   int         RemainingData;          // Data not sent yet in bytes
   double      TimeNotUsed;            // Time not used during the last Poll-TXOP
   double      MinimumTXOPGuaranteed;  // Note: computed at the first schedule
   double      MaximumTXOPGuaranteed;  // (actually when TSPECs are received by QAP)
   double      MeanDataRate_TXOP;      // The TXOP evaluated using MeanDataRate (useful to limit the scheduler computation)
};
// XXX v3.0End

class Mac802_11e;

class Mac802_11 : public Mac {
   friend class IFTimer;
   friend class NavTimer;
   friend class RxTimer;
	// ---------- HCFstart
	friend class CfbTimer;
	friend class CfpTimer;
	friend class CapTimer;
	friend class CfpStartTimer;
	friend class CapStartTimer;
   friend class BS_DropTail;
   friend class Mac802_11e;
   friend class Mac802_11HC;
	friend class Mac802_11eHC;
	// ---------- HCFend
   friend class MyErrorModel;

public:
   Mac802_11();
   void recv(Packet *p, Handler *h);
   void set_qos(int n, double f, int w, int srl, int lrl, int wmax);
   virtual void set_queue(int n);
   void set_callback(Handler *h);
   void set_burst(int n, int len);
   void reset_txop(int n);
   void decr_ptosend(int n);

   /* ======================================================================
           Packet Headers Routines
   ========================================================================= */

        inline int hdr_dst(char* hdr, int dst = -2) {
                struct hdr_mac802_11 *dh = (struct hdr_mac802_11*) hdr;
                if(dst > -2)
                        STORE4BYTE(&dst, (dh->dh_da));
                return ETHER_ADDR(dh->dh_da);
        }

        inline int hdr_src(char* hdr, int src = -2) {
                struct hdr_mac802_11 *dh = (struct hdr_mac802_11*) hdr;
                if(src > -2)
                        STORE4BYTE(&src, (dh->dh_sa));
                return ETHER_ADDR(dh->dh_sa);
        }

        inline int hdr_type(char* hdr, u_int16_t type = 0) {
                struct hdr_mac802_11 *dh = (struct hdr_mac802_11*) hdr;
                if(type)
                        STORE2BYTE(&type,(dh->dh_body));
                return GET2BYTE(dh->dh_body);
        }

        Mac802_11e* tc[MAX_TC];         // A vector of Traffic Categories

protected:
   int command(int argc, const char*const* argv);
   void transmit (Packet* p, double timeout, int n);
   void discard(Packet *p, const char* why);
   void trace_pkt(Packet *p);
   int initialized(int n);
   void txHandler(void);
   virtual void navHandler(void);
   virtual void recvHandler(void);
   // ---------- HCFstart
   virtual void cfpStartHandler(void) { exit(1); };
   virtual void cfpHandler(void) { exit(1); };
   virtual void capStartHandler(void) { exit(1); };
   virtual void capHandler(void);
   virtual void cfbHandler(void);
   // ---------- HCFend
   void set_rx_state(MacState x);
   void check_backoff_timer(int n);
   double txtime(Packet *p);
   double txtime(double psz, double drt);
   void capture(Packet *p);
   void collision(Packet *p, int nt);
   void recv_timer(void);
   void rx_resume(void);
   void mac_log(Packet *p) {
      logtarget_->recv(p, (Handler*) 0);
   }
   void upsend(Packet *p, int n);
   virtual void dump_tc();
   void need_pkt(int n);
   int resolve_coll(int n);
   // ---------- HCFstart
   int queueLength(int ac);
   int queueByteLength(int ac);
   u_int32_t HCmacAddress(void);
   int dataToSend(int dst, int ac);
   void recvBCN(Packet *p);
   void recvCC(Packet *p);
   void recvCFEND(Packet *p);
   // ---------- HCFend

   /* ======================================================================
           Network Allocation Vector (NAV) routines
   ====================================================================== */
   inline double sec(double t) { return(t *= 1.0e-6); } // Converts micro-seconds in seconds

	inline u_int16_t usec(double t) {                    // Converts seconds in micro-seconds
		u_int16_t us = (u_int16_t)floor((t *= 1e6) + 0.5);
		return us;
	}

	inline void set_nav(u_int16_t us) { // Updates the NAV only if the new adjournment 'us' is larger
		double now = Scheduler::instance().clock();     // than the previous
		double t = us * 1e-6;
		if((now + t) > nav_) {
			nav_ = now + t;
			if(mhNav_.busy())
				mhNav_.stop();
			mhNav_.start(t);
		}
	}

	// ---------- HCFstart
	inline void set_nav(double us) {
		double now = Scheduler::instance().clock();
		double t = us * 1e-6;

		if((now + t) > nav_) {
			nav_ = now + t;
			if(mhNav_.busy())
				mhNav_.stop();
			mhNav_.start(t);
		}
	}
	// ---------- HCFend

   double basicRate_;              // Control frames transmission rate
   double dataRate_;               // Data frames transmission rate for this station

   NsObject* logtarget_;
   u_int16_t sta_seqno_;	        // The next MAC sequence number that must be used
   int cache_node_count_;          // Wireless nodes counter
   Host *cache_;

   IFTimer mhIF_;      // Network Interface timer, started when the node transmits on the channel
   NavTimer mhNav_;    // NAV timer
   RxTimer mhRecv_;	  // Timer used to emulate the receiving time of the incoming packet (from the first to the last bit)

   double nav_;		        // Network Allocation Vector (NAV)
   MacState rx_state_;	        // Incoming state (MAC_RECV, MAC_COLL or MAC_IDLE)
   int tx_active_;	                // Transmitter is active (>=0) or not (-1)
   int num_tc;                     // The number of 802.11e EDCF Access Categories

   bool err_frame;                 // 0 if the last received frame was error-free, 1 otherwise
   double eifs_;		        // Extended Interframe Space
   double sifs_;		        // Short Interframe Space
   // ---------- HCFstart
   double pifs_;			// PCF Interframe Space
   bool capture_enabled;		// enable/disable packet capture when collisions occur
   // ---------- HCFend

   int txop;                       // Identifier of the queue that has gained access to the channel
   int sta_coll;                   // Number of colliding stations
   bool detect_coll;               // True if a collision has been detected
   TracedInt virtcoll;             // Virtual collision counter (for statistics)

   double txopLimit_;  // During bursting or CFB the time at which the first frame was transmitted

   double TXOPLmt;   // The maximum time for which a station can transmit on the channel

   // ---------- HCFstart
   Packet *pktMNG_;     // Outgoing Beacon, CC or CF-End frame
                        // that is sent by the HC with TC=0

	bool legacy_;			// True if this is a legacy station

	int cap_;			// -1 if not in CAP, CAP's TC otherwise
	bool capWaiting_;		// True if waiting to start the CAP
	CapTimer mhCap_;		// Controlled Access Period (CAP) timer
	CapStartTimer mhCapStart_;	// Timer that counts PIFS before the CAP

	bool cfp_;			// True if we're during the CFP
	bool cfpWaiting_;		// True if waiting to start the CFP
	double cfpDurRemaining_;	// Shortened CFP duration
	CfpTimer mhCfp_;		// Contention Free Period (CFP) timer
	CfpStartTimer mhCfpStart_;	// Timer that counts PIFS before the CFP

	int cfb_;			// -1 if not in CFB, CFB's TC otherwise
	double txopDuration_;		// The max duration of the current CFB
	bool cfbFailed_;		// True if tx failed during the CFB
	CfbTimer mhCfb_;		// Contention Free Burst (CFB) timer

	// XXX v3.0Start
	// NOTE: the following attributes belong
	//       only to the HC. However, they
	//       have to be defined here for
	//       inheritance reasons
	TSPECElem	tspecs_[MAX_STA];       // TSPEC management:
                                       // for each station the HC stores
                                       // several informations.
                                       // NOTE: the station MAC address
                                       //       is the index to access
                                       //       to array information for
                                       //       that particular station

   // XXX v3.0Start
   // NOTE: the following attributes belong
   //       only to the HC. However, they
   //       have to be defined here for
   //       inheritance reasons
   double bitRates_[MAX_STA];          // bitarates of stations (mac-address is the index!)
   // XXX v3.0End
   // ---------- HCFend
};

/* ======================================================================
   The 802.11e MAC class, one for each EDCF Access Category
   ====================================================================== */

class Mac802_11e : public Connector {
	friend class DeferTimer;
	friend class BackoffTimer;
	friend class Mac802_11;
	// ---------- HCFstart
	friend class Mac802_11HC;
	friend class Mac802_11eHC;
	// ---------- HCFend
        friend class TxTimer;
        friend class IntNavTimer;

public:
	Mac802_11e(Mac802_11* upmac, PHY_MIB* p, MAC_MIB *m);

protected:
	void backoffHandler(void);
	void deferHandler(void);
        virtual void sendHandler(void);
        void int_navHandler(void);
	void send_timer(void);
	int check_pktCTRL();
	int check_pktRTS();
	int check_pktTx();
	void send(Packet *p);
	void sendRTS(int dst);
	void sendCTS(int dst, double duration);
	void sendACK(int dst);
	void sendDATA(Packet *p);
	// ---------- HCFstart
	void capSendRR(int dst);
	void cfbSendCFACK(int dst, int tc);
	void cfbSendDATA(int dst);
	void cfbSendDATA_CFACK(int dst, int tc);
	void cfbSendNULL(int dst);
	virtual void cfbSendCFPOLL(int dst) { exit(1); };
	virtual void cfbSendDATA_CFPOLL(int dst, int tc) { exit(1); };
	// ---------- HCFend
	void RetransmitRTS();
	void RetransmitDATA();
	void recvRTS(Packet *p);
	void recvCTS(Packet *p);
	virtual void recvACK(Packet *p);
	virtual void recvDATA(Packet *p);
	// ---------- HCFstart
	void stopTC(void);
	void recvCC(Packet* p);
	virtual void recvCFACK(Packet* p);
	void recvCFPOLL(Packet* p);
	void recvDATA_CFPOLL(Packet* p);
	virtual void recvDATA_CFACK(Packet* p);
	void recvDATA_CFACK_CFPOLL(Packet* p);
	virtual void recvNULL(Packet* p) { exit(1); };
	void resumeTC(void);
	void edcaTX(bool enabled);	// enable/disable transmission of data during EDCA access.
	bool edcaTXenabled();
	// ---------- HCFend
	virtual void tx_resume(void);
   void set_tx_state(MacState x);

	void dump(char* fname);

        inline int is_idle() {                                  // Returns 0 if the medium is busy, 1 if it is idle
                if(mac_target_->rx_state_ != MAC_IDLE)
                        return 0;
                if(tx_state_ != MAC_IDLE)                                   // The channel is assumed to be busy if the node is receiving
                        return 0;                                           // or transmitting a frame on the channel (physical sensing), or
                if(mac_target_->nav_ > Scheduler::instance().clock())       // if the NAV at present time is still counting (virtual
                        return 0;                                           // sensing), or if another Access Category has gained access
                if(int_nav > Scheduler::instance().clock())                 // to the channel
                        return 0;
                return 1;
        }

	inline void inc_cw() {                  // This method doubles the contention window, limiting its value to 'CWmax'
		cw_ = (cw_ << 1) + 1;
		if(cw_ > CWmax)
			cw_ = CWmax;
	}

	inline void rst_cw() { cw_ = CWmin; }  // This method resets the contention window to 'CWmin'

   // Updates the "internal" NAV only if the new adjournment 'us' is larger
   inline void set_int_nav(u_int16_t us) {
		double now = Scheduler::instance().clock();     // than the previous
		double t = us * 1e-6;
		if((now + t) > int_nav) {
			int_nav = now + t;
			if(mhintNav_.busy())
				mhintNav_.stop();
			mhintNav_.start(t);
		}
	}

   PHY_MIB *phymib_;
   MAC_MIB *macmib_;
   Mac802_11* mac_target_; // Pointer to the main Mac802_11 class

   DeferTimer mhDefer_;       // Defer timer
   BackoffTimer mhBackoff_;   // Backoff timer
   TxTimer mhSend_;           // Timer used to wait for the transmission timeout
   IntNavTimer mhintNav_;     // Internal NAV timer for 802.11e EDCF

	MacState tx_state_;	// Outgoing state (MAC_RTS, MAC_SEND, MAC_CTS, MAC_ACK or MAC_IDLE)
	Packet *pktRTS_;	// Outgoing RTS packet
	Packet *pktCTRL_;	// Outgoing CTS, ACK or Null packet
	Packet *pktTx_;         // Outgoing DATA packet
	// ---------- HCFstart
	Packet *pktPOLL_;	// Outgoing CF-Poll packet
	Packet *pktBackup_;	// Backup DATA packet
	// ---------- HCFend

   u_int32_t cw_;		// Contention Window
   u_int32_t ssrc_;	// STA Short Retry Counter
   u_int32_t slrc_;	// STA Long Retry Counter
   u_int32_t CWmin;        // Minimum value for the contention window
   u_int32_t CWmax;        // Maximum value for the contention window
   u_int32_t ShortRL;      // Short Retry Limit
   u_int32_t LongRL;       // Long Retry Limit
   double difs_;		// DCF Interframe Space (AIFS in 802.11e)

   int nq;                 // The identifier of this virtual queue
   bool virtual_coll;      // True during a virtual collision
   bool find_idle;         // True if the channel is found idle when a new frame has come
   double int_nav;         // Internal NAV : allows to stop backoff when another virtual queue is transmitting
   int bursting;           // During bursting, the maximum number of frames to transmit
   int ptosend;            // During bursting, a counter for the frames not sent yet
   // ---------- HCFstart
   bool backoffOn_;	// Saves the state of the Backoff timer
   bool edcaTX_status;     // Data trasmission enabled/disabled during EDCA channel access
	// ---------- HCFend

        TracedInt rcv_ifq_tcp;
        TracedInt rcv_ifq_ack;          // Other variables for Tcl statistics
        TracedInt rcv_ifq_message;
        TracedInt rcv_ifq_ARP;
        TracedInt rcv_ifq_udp;
	TracedInt rcv_ifq_voip;
        TracedInt rcv_ifq_mac_cts;
        TracedInt rcv_ifq_mac_rts;
        TracedInt rcv_ifq_mac_ack;
        TracedInt send_erm_tcp;
        TracedInt send_erm_ack;
        TracedInt send_erm_message;
        TracedInt send_erm_ARP;
        TracedInt send_erm_udp;
	TracedInt send_erm_voip;
        TracedInt send_erm_mac_cts;
        TracedInt send_erm_mac_rts;
        TracedInt send_erm_mac_ack;
        TracedInt rcv_chn_tcp;
        TracedInt rcv_chn_ack;
        TracedInt rcv_chn_message;
        TracedInt rcv_chn_ARP;
        TracedInt rcv_chn_udp;
	TracedInt rcv_chn_voip;
        TracedInt rcv_chn_mac_cts;
        TracedInt rcv_chn_mac_rts;
        TracedInt rcv_chn_mac_ack;
        TracedInt rcv_chn_my_tcp;
        TracedInt rcv_chn_my_ack;
        TracedInt rcv_chn_my_message;
        TracedInt rcv_chn_my_ARP;
        TracedInt rcv_chn_my_udp;
	TracedInt rcv_chn_my_voip;
        TracedInt rcv_chn_my_mac_cts;
        TracedInt rcv_chn_my_mac_rts;
        TracedInt rcv_chn_my_mac_ack;
        TracedInt rcv_chn_other_tcp;
        TracedInt rcv_chn_other_ack;
        TracedInt rcv_chn_other_message;
        TracedInt rcv_chn_other_ARP;
        TracedInt rcv_chn_other_udp;
	TracedInt rcv_chn_other_voip;
        TracedInt rcv_chn_other_mac_cts;
        TracedInt rcv_chn_other_mac_rts;
        TracedInt rcv_chn_other_mac_ack;
        TracedInt rcv_chn_err_tcp;
        TracedInt rcv_chn_err_ack;
        TracedInt rcv_chn_err_message;
        TracedInt rcv_chn_err_ARP;
        TracedInt rcv_chn_err_udp;
	TracedInt rcv_chn_err_voip;
        TracedInt rcv_chn_err_mac_cts;
        TracedInt rcv_chn_err_mac_rts;
        TracedInt rcv_chn_err_mac_ack;
        TracedInt rcv_chn_col_tcp;
        TracedInt rcv_chn_col_ack;
        TracedInt rcv_chn_col_message;
        TracedInt rcv_chn_col_ARP;
        TracedInt rcv_chn_col_udp;
	TracedInt rcv_chn_col_voip;
        TracedInt rcv_chn_col_mac_cts;
        TracedInt rcv_chn_col_mac_rts;
        TracedInt rcv_chn_col_mac_ack;
        TracedDouble col;
        TracedInt virt_col;
        TracedInt busy_tcp;
        TracedInt busy_ack;
        TracedInt busy_message;
        TracedInt busy_ARP;
        TracedInt busy_udp;
	TracedInt busy_voip;
        TracedInt busy_mac_cts;
        TracedInt busy_mac_rts;
        TracedInt busy_mac_ack;
        TracedInt max_rtx_tcp;
        TracedInt max_rtx_ack;
        TracedInt max_rtx_message;
        TracedInt max_rtx_ARP;
        TracedInt max_rtx_udp;
	TracedInt max_rtx_voip;
        TracedInt max_rtx_mac_cts;
        TracedInt max_rtx_mac_rts;
        TracedInt max_rtx_mac_ack;
        TracedInt rtx_tcp;
        TracedInt rtx_ack;
        TracedInt rtx_message;
        TracedInt rtx_ARP;
        TracedInt rtx_udp;
        TracedInt rtx_voip;
        TracedInt rtx_mac_cts;
        TracedInt rtx_mac_rts;
        TracedInt rtx_mac_ack;
        TracedInt send_ll_tcp;
        TracedInt send_ll_ack;
        TracedInt send_ll_message;
        TracedInt send_ll_ARP;
        TracedInt send_ll_udp;
	TracedInt send_ll_voip;
        TracedInt send_ll_mac_cts;
        TracedInt send_ll_mac_rts;
        TracedInt send_ll_mac_ack;
        TracedInt dupl_tcp;
        TracedInt dupl_ack;
        TracedInt dupl_message;
        TracedInt dupl_ARP;
        TracedInt dupl_udp;
	TracedInt dupl_voip;
        TracedInt dupl_mac_cts;
        TracedInt dupl_mac_rts;
        TracedInt dupl_mac_ack;
	// ---------- HCFstart
   TracedDouble cfb_time;
   TracedInt cfb_num;
   TracedInt cfb_num_bad_end;
	// ---------- HCFend
	// XXX v2.0Start
	TracedInt cfb_data_fast_recv;
	TracedInt cfb_data_slow_recv;
	TracedInt cfb_voice_fast_recv;
	TracedInt cfb_voice_slow_recv;
	TracedInt cfb_noDataToSend;
	// XXX v2.0End
};

//---------- HCFstart
/* ======================================================================
   The basic 802.11 Hybrid Coordinator MAC class.
   ====================================================================== */

// Elements of the HC Polling Lists
struct cfPollElem {
   Mac802_11*  mac;
   u_int32_t   addr;    // QSTA address
   int         tc;      // Traffic Category
};

struct rrPollElem {
	u_int32_t	addr;			// QSTA address
	int		tc;			// Traffic Category
	double		txop_duration;		// Requested TXOP duration
};

// Scheduler being used
typedef enum {
   SCHED_DRAFT = 0,
   SCHED_ENHANCED,
   SCHED_MMFAIR
} schedule_t;
// XXX v3.0End

class Mac802_11eHC;

class Mac802_11HC : public Mac802_11 {
   friend class IFTimer;
   friend class NavTimer;
   friend class RxTimer;
   friend class TbttTimer;
   friend class CapRateTimer;
   friend class CfbTimer;
   friend class CfpTimer;
   friend class CapTimer;
   friend class CfpStartTimer;
   friend class CapStartTimer;
   friend class Mac802_11eHC;
   friend class MyErrorModel;

public:
	Mac802_11HC();
   virtual void set_queue(int n);

private:
   int command(int argc, const char*const* argv);
   virtual void navHandler(void);
   virtual void recvHandler(void);

   void capRateHandler(void);
   virtual void capStartHandler(void);
   virtual void capHandler(void);

   void tbttHandler(void);
   virtual void cfpStartHandler(void);
   virtual void cfpHandler(void);

   virtual void cfbHandler(void);

   void sendCC(void);
   void sendBCN(void);
   void sendCFEND(void);
   bool recv_pktCheck(int src, int nt, char* type);
   void recv_timer(void);
   void recvRR(Packet *p);

   virtual void dump_tc(void);
   int is_idle(void);
   int bursting(void);
   double pcf(void);
   double simDuration(void);

  // XXX v3.0Start ------------------
   // get from a station the size of the remaining data for a TC
   void getRemainingData(int poll_index, int tc, u_int32_t addr);
   // schedules per-station TXOPLIMIT and and CFPrate
   void schedule(bool first_time = false);
   // normalizes the scheduled TXOP in order to fit the cfpDuration
   void normalizeTXOPs();
   // utility: given a size and a tx_bitrate returns the time needed for
   //  transmitting it (including the headers of several protocols)
   double dataTxTime(int payload, double bitrate);
   // tells what scheduler is being used (either SCHED_DRAFT or SCHED_ENHANCED or SCHED_MMFAIR)
   // default: SCHED_DRAFT
   schedule_t scheduler;
   bool second_schedule; // true during second schedule of the CFP (used only with SCHED_MMFAIR)
  // XXX v3.0End ---------------------

   TbttTimer mhTbtt_;		// Target Beacon Tx Time (TBTT) timer
   CapRateTimer mhCapRate_;	// Counts the interval between two CAPs

   double   capRate_;                     // Time interval between two CAPs
   int      nccop_;                       // Number of CCOPs during CCI
   rrPollElem	rrPollList_[MAX_RR_POLL];  // Reservation Rqst List
   int      rrPollNum_;                   // Number of elements
   int      rrPollIndex_;                 // RR Poll List index

   double     cfpRate_;      // CP + CFP duration
   // XXX v2.0Start
   double      cfpDuration_;			// CFP duration
   // XXX v2.0End
   cfPollElem	cfPollList_[MAX_CF_POLL];	// Polling List
   int         cfPollNum_;		     // Number of elements
   int         cfPollIndex_;        // Current polled node/TC

	double     cfStartTime_;   // Start time of CAP/CFP (for statistics)
	TracedDouble  cap_time;    // Total duration of CAPs
	TracedInt  cap_num;        // Number of CAPs
	TracedDouble  cfp_time;    // Total duration of CFPs
	TracedInt  cfp_num;        // Number of CFPs

   int  nSTAs_;        // Number of stations
   // XXX v2.0End
};

/* ===========================================================================
   The 802.11e Hybrid Coordinator MAC class, one for each EDCF Access Category
   =========================================================================== */

class Mac802_11eHC : public Mac802_11e {
   friend class DeferTimer;
   friend class BackoffTimer;
   friend class Mac802_11HC;
   friend class TxTimer;
   friend class IntNavTimer;

public:
   Mac802_11eHC(Mac802_11* upmac, PHY_MIB* p, MAC_MIB *m);

private:
   virtual void sendHandler(void);
   void send_timer(void);
   virtual void cfbSendCFPOLL(int dst);
   virtual void cfbSendDATA_CFPOLL(int dst, int tc);
   void cfbSendDATA_CFACK_CFPOLL(int dst, int tc);
   virtual void recvACK(Packet *p);
   virtual void recvDATA(Packet *p);
   virtual void recvCFACK(Packet* p);
   virtual void recvDATA_CFACK(Packet* p);
   virtual void recvNULL(Packet* p);
   virtual void tx_resume(void);
};
//---------- HCFend

#endif /* mac_80211_h */

//----------------------------------------------------------------------------------------------------------------------------//EndEDCFpatch

