/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 1997 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/* Ported from CMU/Monarch's code, nov'98 -Padma.*/


#include <delay.h>
#include <connector.h>
#include <packet.h>
#include <random.h>
 
#include <arp.h>
#include <ll.h>
#include <mac.h>
#include <mac-timers.h>
#include <mac-802_11.h> 

/*
 * Force timers to expire on slottime boundries.
 */
// #define USE_SLOT_TIME

#define ROUND_TIME()	\
	{								\
		assert(slottime);					\
		double rmd = remainder(s.clock() + rtime, slottime);	\
		if(rmd > 0.0)						\
			rtime += (slottime - rmd);			\
		else							\
			rtime += (-rmd);				\
	}


/* ======================================================================
   Timers
   ====================================================================== */

void
MacTimer::start(double time)
{
	Scheduler &s = Scheduler::instance();
	assert(busy_ == 0);

	busy_ = 1;
	paused_ = 0;
	stime = s.clock();
	rtime = time;
	assert(rtime >= 0.0);


	s.schedule(this, &intr, rtime);
}

void
MacTimer::stop(void)
{
	Scheduler &s = Scheduler::instance();

	assert(busy_);

	if(paused_ == 0)
		s.cancel(&intr);

	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;
}

/* ======================================================================
   Defer Timer
   ====================================================================== */
void
DeferTimer::start(double time)
{
	Scheduler &s = Scheduler::instance();

	assert(busy_ == 0);

	busy_ = 1;
	paused_ = 0;
	stime = s.clock();
	rtime = time;
#ifdef USE_SLOT_TIME
	ROUND_TIME();
#endif
	assert(rtime >= 0.0);
	s.schedule(this, &intr, rtime);
}


void    
DeferTimer::handle(Event *)
{       
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;



	mac->deferHandler();
}

/* ======================================================================
   NAV Timer
   ====================================================================== */
void
NavTimer::handle(Event *)
{
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;

//------------------------------------------------------------------------------------------------------------------------------//EDCFpatch

	init_mac->navHandler();         // navHandler is a method of the 802_11 class

//----------------------------------------------------------------------------------------------------------------------------//EndEDCFpatch

}

//------------------------------------------------------------------------------------------------------------------------------//EDCFpatch

void
IntNavTimer::handle(Event *)            // A new timer performing as internal NAV for 802.11e
{
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;
	mac->int_navHandler();
}

//----------------------------------------------------------------------------------------------------------------------------//EndEDCFpatch

/* ======================================================================
   Receive Timer
   ====================================================================== */
void    
RxTimer::handle(Event *)
{       
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;

//------------------------------------------------------------------------------------------------------------------------------//EDCFpatch

	init_mac->recvHandler();        // recvHandler is a method of the 802_11 class

//----------------------------------------------------------------------------------------------------------------------------//EndEDCFpatch

}


/* ======================================================================
   Send Timer
   ====================================================================== */
void    
TxTimer::handle(Event *)
{       
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;

	mac->sendHandler();
}


/* ======================================================================
   Interface Timer
   ====================================================================== */
void
IFTimer::handle(Event *)
{
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;

//------------------------------------------------------------------------------------------------------------------------------//EDCFpatch

        init_mac->txHandler();          // txHandler is a method of the 802_11 class

//----------------------------------------------------------------------------------------------------------------------------//EndEDCFpatch

}


/* ======================================================================
   Backoff Timer
   ====================================================================== */
void
BackoffTimer::handle(Event *)
{
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;
	difs_wait = 0.0;

	mac->backoffHandler();
}

void
BackoffTimer::start(int cw, int idle)
{                                                 
	Scheduler &s = Scheduler::instance();

	assert(busy_ == 0);

	busy_ = 1;
	paused_ = 0;
	stime = s.clock();
	
//------------------------------------------------------------------------------------------------------------------------------//EDCFpatch

        if (cw > 0) {
	        rtime = (Random::random() % cw) * mac->phymib_->SlotTime;
        } else {
                rtime = 0.0;    // When cw is equal to 0, the backoff timer expires immediately. This is used by 802.11e
        }                       //    to handle virtual collisions (see mac-802_11.cc)

//----------------------------------------------------------------------------------------------------------------------------//EndEDCFpatch

#ifdef USE_SLOT_TIME
	ROUND_TIME();
#endif
	difs_wait = 0.0;

	if(idle == 0)
		paused_ = 1;
	else {                         
		assert(rtime >= 0.0);
		s.schedule(this, &intr, rtime);
	}
}


void
BackoffTimer::pause()
{
	Scheduler &s = Scheduler::instance();

	//the caculation below make validation pass for linux though it
	// looks dummy

	double st = s.clock();
	double rt = stime + difs_wait;
	double sr = st - rt;
        double mst = (mac->phymib_->SlotTime);

        int slots = int (sr/mst);

	if(slots < 0)
		slots = 0;
	assert(busy_ && ! paused_);

	paused_ = 1;
	rtime -= (slots * mac->phymib_->SlotTime);
	assert(rtime >= 0.0);

	difs_wait = 0.0;

	s.cancel(&intr);
}


void
BackoffTimer::resume(double difs)
{
	Scheduler &s = Scheduler::instance();

	assert(busy_ && paused_);

	paused_ = 0;
	stime = s.clock();

	/*
	 * The media should be idle for DIFS time before we start
	 * decrementing the counter, so I add difs time in here.
	 */
	difs_wait = difs;
	/*
#ifdef USE_SLOT_TIME
	ROUND_TIME();
#endif
	*/
	assert(rtime + difs_wait >= 0.0);
       	s.schedule(this, &intr, rtime + difs_wait);
}

//------------------------------------------------------------------------------------------------------------------------------//EDCFpatch

/*###################################################################
#                                                                   #
# remaining: returns the number of remaining slots before the       #
#            backoff expiration. Used in mac-802_11.cc to           #
#            detect virtual collisions                              #
#                                                                   #
###################################################################*/

double BackoffTimer::remaining(void) {
        return ((stime + rtime + difs_wait) - Scheduler::instance().clock())/(double)(mac->phymib_->SlotTime);
}

// XXX HCFstart
/* ======================================================================
   Target Beacon Transmission Time (TBTT) Timer
   ====================================================================== */

/*###################################################################
#                                                                   #
# TbttTimer: used by the Hybrid Coordinator (HC), its expiration    #
#            starts the CFP, invoking Mac802_11HC::tbttHandler()    #
#                                                                   #
###################################################################*/

void    
TbttTimer::handle(Event *)
{       
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;

	init_macHC->tbttHandler();
}

double
TbttTimer::startTime()
{
	if(busy_)
		return(stime);
	else
		return(-1.0);
}

/* ======================================================================
   Controlled Access Period (CAP) rate Timer
   ====================================================================== */

/*###################################################################
#                                                                   #
# CapRateTimer: used by the Hybrid Coordinator (HC), its            #
#               expiration starts a CAP, invoking                   #
#               Mac802_11HC::capRateHandler()                       #
#                                                                   #
###################################################################*/

void    
CapRateTimer::handle(Event *)
{       
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;

	init_macHC->capRateHandler();
}

/* ======================================================================
   Contention Free Period (CFP) Timer
   ====================================================================== */

/*###################################################################
#                                                                   #
# CfpTimer: used by the Hybrid Coordinator (HC) during              #
#           the CFP to coordinate the PCF mechanism                 #
#                                                                   #
#           passed() returns the time that has passed               #
#                    since the timer was started. Used              #
#                    to evaluate the duration of the                #
#                    foreshortened CFP                              #
#                                                                   #
###################################################################*/

void    
CfpTimer::handle(Event *)
{       
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;

	// init_mac is used instead of init_macHC due to
	// inheritance limitations that force the CfpTimer
	// to be declared in the Mac802_11 class.
	// In fact it is used by the Mac802_11HC class
	init_mac->cfpHandler();
}

double
CfpTimer::passed()
{
	if(busy_)
		return (NOW - stime);
	else
		return(-1.0);
}

/* ======================================================================
   Contention Free Burst (CFB) Timer
   ====================================================================== */

/*###################################################################
#                                                                   #
# CfbTimer: used during the CFB to coordinate the frame exchanges   #
#                                                                   #
###################################################################*/

void    
CfbTimer::handle(Event *)
{       
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;

	init_mac->cfbHandler();
}

/* ======================================================================
   Controlled Access Period (CAP) Timer
   ====================================================================== */

/*###################################################################
#                                                                   #
# CapTimer: used by the Hybrid Coordinator (HC), it coordinates     #
#           the frame exchanges during the CAPs.                    #
#           When invoked by QSTAs, it is used to transmit the       #
#           Reservation Request frame during a CCI                  #
#                                                                   #
###################################################################*/

void    
CapTimer::handle(Event *)
{       
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;

	init_mac->capHandler();
}

/* ======================================================================
   CFP start Timer
   ====================================================================== */

/*###################################################################
#                                                                   #
# CfpStartTimer: used by the Hybrid Coordinator (HC), it is         #
#                started a PIFS before the CFP begins. Its          #
#                expiration invokes Mac802_11HC::cfpStartHandler()  #
#                that sets up the HC for the PCF mechanism          #
#                                                                   #
###################################################################*/

void    
CfpStartTimer::handle(Event *)
{       
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;

	// init_mac is used instead of init_macHC due to
	// inheritance limitations that force the CfpStartTimer
	// to be declared in the Mac802_11 class.
	// In fact it is used by the Mac802_11HC class
	init_mac->cfpStartHandler();
}

/* ======================================================================
   CAP start Timer
   ====================================================================== */

/*###################################################################
#                                                                   #
# CapStartTimer: used by the Hybrid Coordinator (HC), it is         #
#                started a PIFS before the CAP begins. Its          #
#                expiration invokes Mac802_11HC::capStartHandler()  #
#                that sets up the HC for the CAP mechanism          #
#                                                                   #
###################################################################*/

void    
CapStartTimer::handle(Event *)
{       
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;

	// init_mac is used instead of init_macHC due to
	// inheritance limitations that force the CfpStartTimer
	// to be declared in the Mac802_11 class.
	// In fact it is used by the Mac802_11HC class
	init_mac->capStartHandler();
}
// XXX HCFend

//----------------------------------------------------------------------------------------------------------------------------//EndEDCFpatch

