/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 1997 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/* Ported from CMU/Monarch's code, nov'98 -Padma.*/

#ifndef __mac_timers_h__
#define __mac_timers_h__

/* ======================================================================
   Timers
   ====================================================================== */

//------------------------------------------------------------------------------------------------------------------------------//EDCFpatch

//   In the timer constructors, "Mac802_11" has been changed in "Mac802_11e", "Mac802_11eHC" or "Mac802_11HC" when needed !!

class Mac802_11e;       // The new 802.11e class
// XXX HCFstart
class Mac802_11eHC;
class Mac802_11HC;
// XXX HCFend

//----------------------------------------------------------------------------------------------------------------------------//EndEDCFpatch

class Mac802_11;

class MacTimer : public Handler {
public:

//------------------------------------------------------------------------------------------------------------------------------//EDCFpatch

	MacTimer(Mac802_11e* m, double s = 0) : mac(m), slottime(s) {           // Constructor for timers used
		busy_ = paused_ = 0; stime = rtime = 0.0;                       // in the new Mac802_11e class
	}

	// XXX HCFstart
	MacTimer(Mac802_11eHC* m, double s = 0) : macHC(m), slottime(s) {       // Constructor for timers used
		busy_ = paused_ = 0; stime = rtime = 0.0;                       // in the new Mac802_11eHC class
	} 

	MacTimer(Mac802_11HC* m, double s = 0) : init_macHC(m), slottime(s) {	// Constructor for timers used
		busy_ = paused_ = 0; stime = rtime = 0.0;			// in the new Mac802_11HC class
	}
	// XXX HCFend

//----------------------------------------------------------------------------------------------------------------------------//EndEDCFpatch

        MacTimer(Mac802_11* m, double s = 0) : init_mac(m), slottime(s) {
		busy_ = paused_ = 0; stime = rtime = 0.0;
	}

	virtual void handle(Event *e) = 0;

	virtual void start(double time);
	virtual void stop(void);
	virtual void pause(void) { assert(0); }
	virtual void resume(void) { assert(0); }

	inline int busy(void) { return busy_; }
	inline int paused(void) { return paused_; }
	inline double expire(void) {
		return ((stime + rtime) - Scheduler::instance().clock());
	}

protected:

//------------------------------------------------------------------------------------------------------------------------------//EDCFpatch

        Mac802_11e	*mac;           // The new 802.11e class
        Mac802_11	*init_mac;      // The original 802.11 class
	// XXX HCFstart
	Mac802_11eHC	*macHC;		// The new 802.11e class for HC
	Mac802_11HC	*init_macHC;	// The new 802.11 class for HC
	// XXX HCFend

//----------------------------------------------------------------------------------------------------------------------------//EndEDCFpatch

	int		busy_;
	int		paused_;
	Event		intr;
	double		stime;	// start time
	double		rtime;	// remaining time
	double		slottime;
};


class BackoffTimer : public MacTimer {
public:
	BackoffTimer(Mac802_11e *m, double s) : MacTimer(m, s), difs_wait(0.0) {}

	void	start(int cw, int idle);
	void	handle(Event *e);
	void	pause(void);
	void	resume(double difs);

//------------------------------------------------------------------------------------------------------------------------------//EDCFpatch

        double  remaining(void);        // Returns the number of remaining slots

//----------------------------------------------------------------------------------------------------------------------------//EndEDCFpatch


private:
	double	difs_wait;
};

class DeferTimer : public MacTimer {
public:
	DeferTimer(Mac802_11e *m, double s) : MacTimer(m,s) {}

	void	start(double);
	void	handle(Event *e);
};

class IFTimer : public MacTimer {
public:
	IFTimer(Mac802_11 *m) : MacTimer(m) {}

	void	handle(Event *e);
};

class NavTimer : public MacTimer {
public:
	NavTimer(Mac802_11 *m) : MacTimer(m) {}

	void	handle(Event *e);
};

class IntNavTimer : public MacTimer {
public:
	IntNavTimer(Mac802_11e *m) : MacTimer(m) {}

	void	handle(Event *e);
};

class RxTimer : public MacTimer {
public:
	RxTimer(Mac802_11 *m) : MacTimer(m) {}

	void	handle(Event *e);
};

class TxTimer : public MacTimer {
public:
	TxTimer(Mac802_11e *m) : MacTimer(m) {}

	void	handle(Event *e);
};

// XXX HCFstart
class TbttTimer : public MacTimer {
public:
	TbttTimer(Mac802_11HC *m) : MacTimer(m) {}

	void	handle(Event *e);
	double	startTime(void);
};

class CapRateTimer : public MacTimer {
public:
	CapRateTimer(Mac802_11HC * m) : MacTimer(m) {}

	void	handle(Event *e);
};

class CfpTimer : public MacTimer {
public:
	CfpTimer(Mac802_11 *m) : MacTimer(m) {}

	void	handle(Event *e);
	double	passed();
};

class CfbTimer : public MacTimer {
public:
	CfbTimer(Mac802_11 *m) : MacTimer(m) {}

	void	handle(Event *e);
};

class CapTimer : public MacTimer {
public:
	CapTimer(Mac802_11 * m) : MacTimer(m) {}

	void	handle(Event *e);
};

class CfpStartTimer : public MacTimer {
public:
	CfpStartTimer(Mac802_11 *m) : MacTimer(m) {}

	void	handle(Event *e);
};

class CapStartTimer : public MacTimer {
public:
	CapStartTimer(Mac802_11 *m) : MacTimer(m) {}

	void	handle(Event *e);
};
// XXX HCFend

#endif /* __mac_timers_h__ */

