/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
// HCF v3.0

#ifndef lint
static const char rcsid[] =
    "@(#) $Header: /nfs/jade/vint/CVSROOT/ns-2/queue/drop-tail.cc,v 1.14 2002/05/07 18:28:27 haldar Exp $ (LBL)";
#endif

#include "drop-tail.h"

static class DropTailClass : public TclClass {
 public:
	DropTailClass() : TclClass("Queue/DropTail") {}
	TclObject* create(int, const char*const*) {
		return (new DropTail);
	}
} class_drop_tail;

void DropTail::reset()
{
	Queue::reset();
}

int
DropTail::command(int argc, const char*const* argv) {

	if (argc==2) {
		if (strcmp(argv[1], "printstats") == 0) {
			print_summarystats();
			return (TCL_OK);
		}
	}
	if (argc == 3) {
		if (!strcmp(argv[1], "packetqueue-attach")) {
			delete q_;
			if (!(q_ = (PacketQueue*) TclObject::lookup(argv[2])))
				return (TCL_ERROR);
			else {
				pq_ = q_;
				return (TCL_OK);
			}
		}
	}
	return Queue::command(argc, argv);
}

/*
 * drop-tail
 */
void DropTail::enque(Packet* p)
{
	if (summarystats) {
                Queue::updateStats(qib_?q_->byteLength():q_->length());
	}
	Packet* pt = q_->enque(p);
	int qlimBytes = qlim_ * mean_pktsize_;
	if ((!qib_ && q_->length() >= qlim_) ||
	    (qib_ && q_->byteLength() >= qlimBytes)) {
		if (drop_front_) { /* remove from head of queue */
			Packet *pp = q_->deque();
			drop(pp);
		} else {

//------------------------------------------------------------------------------------------------------------------------------//EDCFpatch

                        switch (HDR_CMN(p)->ptype()) {          // Statistics: dropped packets
                                case PT_TCP:
                                        ++(q_->drop_tcp);
                                        break;
                                case PT_ACK:
                                        ++(q_->drop_ack);
                                        break;
                                case PT_MESSAGE:
                                        ++(q_->drop_message);
                                        break;
                                case PT_ARP:
                                        ++(q_->drop_ARP);
                                        break;
                                case PT_VBR:
                                case PT_CBR:
                                        ++(q_->drop_udp);
                                        break;
                                case PT_VOIP:
                                        ++(q_->drop_voip);
                                        break;	
                        }

//----------------------------------------------------------------------------------------------------------------------------//EndEDCFpatch

			q_->remove(p, pt);
			drop(p);
		}
	}
}

Packet* DropTail::deque()
{
        if (summarystats && &Scheduler::instance() != NULL) {
                Queue::updateStats(qib_?q_->byteLength():q_->length());
        }
	return q_->deque();
}

void DropTail::print_summarystats()
{
   //double now = Scheduler::instance().clock();
   printf("True average queue: %5.3f", true_ave_);
   if (qib_)
      printf(" (in bytes)");
   printf(" time: %5.3f\n", total_time_);
}

//------------------------------------------------------------------------------------------------------------------------------//EDCFpatch

//#define MYIFQDBG
#define MYIFQTIME 20000.0

static class BS_DropTailClass : public TclClass {
 public:
	BS_DropTailClass() : TclClass("Queue/DropTail/BS") {}
	TclObject* create(int, const char*const*) {
		return (new BS_DropTail);
	}
} class_bs_drop_tail;

int BS_DropTail::command(int argc, const char*const* argv) {

	if (argc==2) {
		if (strcmp(argv[1], "dump_queue") == 0) {       // A command for reading some variables at
			dump_stat();                                 // the end of the simulation
			return (TCL_OK);
		}
		else if (strcmp(argv[1], "dump_avg_size") == 0) {	// A command for getting queue avg_size 
			dump_avg_size();
			return (TCL_OK);
		}
		else if (strcmp(argv[1], "reset_avg_size") == 0) {	// A command for resetting avg_size counting
			reset_avg_size();
			return (TCL_OK);
		}
   } else if (argc==3) {
      if (strcmp(argv[1], "init-llc") == 0)
      {  // This command sets the bursting modality:
         bursting = atoi(argv[2]);
        ((Mac802_11*)target_)->set_callback(&qbs_);  // 0 in order to disability bursting
         return (TCL_OK);                 //   n to transmit with bursts of n frames
      }                                   //  -1 to transmit as many frames as the TXOPlimit allows
   } else if (argc==4) {
      if (strcmp(argv[1], "set-delaybound") == 0)
      {  // This command sets the delay after which a packet in queue is discarded:
        int ac = atoi(argv[2]);
        if (ac > num_queue)
           return TCL_ERROR;
        tab[ac].timeout = atof(argv[3]);
        return (TCL_OK);
      }
   } else if (argc==5) {
      if (strcmp(argv[1], "add-queue") == 0)
      {                                   // This command is called every time a new queue
          int i = atoi(argv[2]);          // is added in a node. It is responsable for
          tab[i].que_  = new PacketQueue; // the queue initialization
          tab[i].tx_burst = 0;
          tab[i].timeout = atof(argv[3]);
          tab[i].limit = atoi(argv[4]);
          num_queue++;
          block[i] = 0;
          return (TCL_OK);
      }
   }

	return Queue::command(argc, argv);
}

/*###################################################################
#                                                                   #
# dump_stats: dumps some variables and calls Tcl for statistics     #
#                                                                   #
###################################################################*/

void BS_DropTail::dump_stat() {
        for (int i=0 ; i<num_queue ; i++) {
                Tcl::instance().evalf("catch \"queue_stat %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\"",i,tab[i].limit, \
                        (int)tab[i].que_->q_arrv, (int)tab[i].que_->q_size, (int)tab[i].que_->rcv_tcp, \
                        (int)tab[i].que_->rcv_ack, (int)tab[i].que_->rcv_message, (int)tab[i].que_->rcv_ARP, (int)tab[i].que_->rcv_udp, (int)tab[i].que_->rcv_voip, \
                        (int)tab[i].que_->drop_tcp, (int)tab[i].que_->drop_ack, (int)tab[i].que_->drop_message, \
                        (int)tab[i].que_->drop_ARP, (int)tab[i].que_->drop_udp, (int)tab[i].que_->drop_voip, \
                        (int)tab[i].que_->tout_tcp, (int)tab[i].que_->tout_ack, (int)tab[i].que_->tout_message, \
                        (int)tab[i].que_->tout_ARP, (int)tab[i].que_->tout_udp, (int)tab[i].que_->tout_voip);
	}
}
/*###################################################################
#                                                                   #
# reset_avg_size: reset avg_size counting                           #
#                                                                   #
###################################################################*/

void BS_DropTail::reset_avg_size() {
   for (int i=0 ; i<num_queue ; i++)
		tab[i].que_->reset_avg_size();
}

/*###################################################################
#                                                                   #
# dump_avg_size: dump avg_size by calling Tcl funcion               #
#                                                                   #
###################################################################*/

void BS_DropTail::dump_avg_size() {
        for (int i=0 ; i<num_queue ; i++)
                Tcl::instance().evalf("catch \"queue_avgsize %d %f\"", i, tab[i].que_->avg_size());
}

/*###################################################################
#                                                                   #
# recv: the main receiving routine called every time a new frame    #
#       comes. It "enques" the new frame or passes it directly to   #
#       MAC if the queue was previously empty                       #
#                                                                   #
#         p: the new incoming packet                                #
#                                                                   #
###################################################################*/

void BS_DropTail::recv(Packet* p, Handler*) {

	HDR_CMN(p)->ts_queue = NOW;

	enque(p);       // After this call, 'qtoen' contains the identifier of the queue in which the new packet has been queued

	if (!block[qtoen]) {    // If this queue was previously empty, then we pass the new frame directly to the MAC layer

		p = force_deque(qtoen);         // Takes a packet from the 'qtoen' queue, i.e. the new incoming packet
                qtodq = qtoen;
                ((Mac802_11*)target_)->set_burst(qtodq,1);      // Informs the MAC to not use bursting
                tab[qtodq].tx_burst = 0;
		if (p != 0) {
			block[qtoen] = 1;       // 'qtoen' will be considered empty only if will be still empty when the MAC
                                                // will request a new frame. For now, block is set to 1

			target_->recv(p, &qbs_);        // Pass the new frame to the MAC

		}

#ifdef MYIFQDBG //=================================================================================
// la coda rimane vuota ... tutto OK se uso RED
else { fprintf(stderr,"Stai usando RED ???"); getchar();}
#endif //==========================================================================================

	}
}

/*###################################################################
#                                                                   #
# enque: this function "enques" the new frame, giving it priority   #
#        if it is a routing or ARP frame                            #
#                                                                   #
#         p: the new incoming packet                                #
#                                                                   #
###################################################################*/

void BS_DropTail::enque(Packet* p)
{
        PacketQueue* temp_ = select(p);         // This call selects the queue to be used

        switch (HDR_CMN(p)->ptype()) {          // Statistics: frames received by the LLC layer
                case PT_TCP:
                        ++(temp_->rcv_tcp);
                        break;
                case PT_ACK:
                        ++(temp_->rcv_ack);
                        break;
                case PT_MESSAGE:
                        ++(temp_->rcv_message);
                        break;
                case PT_ARP:
                        ++(temp_->rcv_ARP);
                        break;
                case PT_CBR:
                case PT_VBR:
                        ++(temp_->rcv_udp);
                        break;
                case PT_VOIP:
                        ++(temp_->rcv_voip);
                        break;	
        }
        switch(HDR_CMN(p)->ptype()) {
                case PT_MESSAGE:
                case PT_ARP:
                        // If 'p' is a priority frame, it will be put at the head of the queue,
                        // so it will be served for first
                        temp_->enqueHead(p);
                        if (!qinb_ && temp_->length() >= tab[qtoen].limit || qinb_ && temp_->byteLength() >= tab[qtoen].limit) {
                                Packet *to_drop = temp_->lookup(temp_->length()-1);
                                switch (HDR_CMN(to_drop)->ptype()) {
                                        case PT_MESSAGE:
                                                ++(temp_->drop_message);
                                                break;
                                        case PT_ARP:
                                                ++(temp_->drop_ARP);
                                                break;
                                        case PT_TCP:
                                                ++(temp_->drop_tcp);
                                                break;
                                        case PT_ACK:
                                                ++(temp_->drop_ack);
                                                break;
                                        case PT_VBR:
                                        case PT_CBR:
                                                ++(temp_->drop_udp);
                                                break;
                                        case PT_VOIP:
                                                ++(temp_->drop_voip);
                                                break;	
                                }
                                temp_->remove(to_drop);         // If this queue has reached its maximum size, it is
                                drop(to_drop);                  // necessary to drop the frame in tail
                        }
                        break;

                default:
                        Packet* pt = temp_->enque(p);   // If 'p' is not priority, it is "enqued" in tail

//fprintf (stderr,"%d %d\n",temp_->length(),temp_->byteLength());

                        if (!qinb_ && temp_->length() >= tab[qtoen].limit || qinb_ && temp_->byteLength() >= tab[qtoen].limit) {
                                switch (HDR_CMN(p)->ptype()) {
                                        case PT_TCP:
                                                ++(temp_->drop_tcp);
                                                break;
                                        case PT_ACK:
                                                ++(temp_->drop_ack);
                                                break;
                                        case PT_CBR:
                                        case PT_VBR:
                                                ++(temp_->drop_udp);
                                                break;
                                        case PT_VOIP:
                                                ++(temp_->drop_voip);
                                                break;	
                                }

                                temp_->remove(p, pt);   // If this queue has reached its maximum size,
                                drop(p);                // the new frame is immediately discarded
                        }
        }
        temp_->q_size += temp_->length();
        temp_->q_arrv++;
}

/*###################################################################
#                                                                   #
# select: chooses the 802.11e Access Category at which 'p' should   #
#          be destined                                              #
#                                                                   #
#         p: the new incoming packet                                #
#                                                                   #
#   Returns a pointer to the selected queue                         #
#                                                                   #
###################################################################*/

PacketQueue* BS_DropTail::select(Packet* p) {
        struct hdr_cmn *hdr = HDR_CMN(p);

        // XXX HCFstart
	hdr->mac_tc_aux = num_queue;	// This value is initialized to the number of ACs
	// XXX HCFend
	
	if (num_queue == 1) {           // MAC 802.11b Standard: only one queue is available
                hdr->mac_tc = 0;
                qtoen = 0;
                return tab[0].que_;
        }

        switch (hdr->ptype()) {         
		case PT_MESSAGE:
                case PT_ARP:
                        hdr->mac_tc = num_queue - 1;
                        qtoen = num_queue - 1;
                        return tab[num_queue - 1].que_;
                        break;
                case PT_TCP:            
                        if(HDR_IP(p)->daddr() == 0) {
                                hdr->mac_tc = mapping_WStcp;
                                qtoen = mapping_WStcp;
                                return tab[mapping_WStcp].que_;
                        } else {
                                hdr->mac_tc = mapping_APtcp;
                                qtoen = mapping_APtcp;
                                return tab[mapping_APtcp].que_;
                        }
                        break;
                case PT_ACK:
                        if(HDR_IP(p)->daddr() == 0) {
                                hdr->mac_tc = mapping_WSack;
                                qtoen = mapping_WSack;
                                return tab[mapping_WSack].que_;
                        } else {
                                hdr->mac_tc = mapping_APack;
                                qtoen = mapping_APack;
                                return tab[mapping_APack].que_;
                        }
                        break;
                        case PT_CBR:
                        if(HDR_IP(p)->daddr() == 0) {
                                hdr->mac_tc = mapping_WScbr;
                                qtoen = mapping_WScbr;
                                return tab[mapping_WScbr].que_;
                        } else {
                                hdr->mac_tc = mapping_APcbr;
                                qtoen = mapping_APcbr;
                                return tab[mapping_APcbr].que_;
                        }
                        break;
                        case PT_VBR:
                        if(HDR_IP(p)->daddr() == 0) {
                                hdr->mac_tc = mapping_WSvbr;
                                qtoen = mapping_WSvbr;
                                return tab[mapping_WSvbr].que_;
                        } else {
                                hdr->mac_tc = mapping_APvbr;
                                qtoen = mapping_APvbr;
                                return tab[mapping_APvbr].que_;
                        }
                        break;
                        case PT_VOIP:
                        if(HDR_IP(p)->daddr() == 0) {
                                hdr->mac_tc = mapping_WSvoip;
                                qtoen = mapping_WSvoip;
                                return tab[mapping_WSvoip].que_;
                        } else {
                                hdr->mac_tc = mapping_APvoip;
                                qtoen = mapping_APvoip;
                                return tab[mapping_APvoip].que_;
                        }
                        break;			
        }

#ifdef MYIFQDBG //=================================================================================
// non ho trovato la coda 802.11e corrispondente
fprintf(stderr,"SELECT NOT FOUND: %s",packet_info.name(hdr->ptype())); getchar();
#endif //==========================================================================================

}

/*###################################################################
#                                                                   #
# force_deque: extracts the frame in head at the 'n' queue          #
#                                                                   #
#              n: the identifier of the queue to be "dequed"       #
#                                                                   #
#    Returns the extracted frame                                    #
#                                                                   #
###################################################################*/

Packet* BS_DropTail::force_deque(int n) {
        return tab[n].que_->deque();
}

/*###################################################################
#                                                                   #
# deque: extracts the frame in head at the 'qtodq' queue            #
#                                                                   #
#    Returns the extracted frame                                    #
#                                                                   #
###################################################################*/

Packet* BS_DropTail::deque() {
        //if (((Mac802_11*)target_)->index_ == 1)
        //    printf("(1) NOW %f   DEQUEUE \n", Scheduler::instance().clock());
        return tab[qtodq].que_->deque();
}

/*###################################################################
#                                                                   #
# resume: passes the next frame to the MAC layer for the current    #
#         Access Category                                           #
#                                                                   #
###################################################################*/

void BS_DropTail::resume() {

#ifdef MYIFQDBG //=================================================================================
if(NOW > MYIFQTIME) {fprintf(stderr,"%s %d CHIAMATA",((Mac802_11*)target_)->name(),qtodq);
getchar();}
#endif //==========================================================================================

        if (tab[qtodq].tx_burst > 0) {          // When bursting is activated, if there are still fames to be transmitted in
                                                // the current burst, we can pass the next frame to the MAC without doing other checks

#ifdef MYIFQDBG //=================================================================================
if (tab[qtodq].que_->length() == 0) {
fprintf(stderr,"Sto facendo bursting, ma la coda  vuota !!!"); getchar();}
if (tab[qtodq].tx_burst != ((Mac802_11*)target_)->tc[qtodq]->ptosend) {
fprintf(stderr,"tx_burst %d e ptosend %d diversi!!!!!",tab[qtodq].tx_burst,((Mac802_11*)target_)->tc[qtodq]->ptosend); getchar();}
#endif //==========================================================================================

                tab[qtodq].tx_burst--;          // The number of remaining frames must be decremented
        } else {

                tab[qtodq].tx_burst = calc_bursting();  // This call returns the length of the next burst
                if (tab[qtodq].tx_burst == 0) {
                        block[qtodq] = 0;               // If the queue is empty, no more frame can be sent and the 'block'
                                                        // flag must be deactivated
                } else {
                        ((Mac802_11*)target_)->set_burst(qtodq,tab[qtodq].tx_burst);    // Informs the MAC about the length
                        tab[qtodq].tx_burst--;                                          // of the next burst
                }
        }

	if (block[qtodq]) {
                Packet* p = deque();
                while (p && NOW - HDR_CMN(p)->ts_queue > tab[qtodq].timeout) {

//fprintf(stderr,"drop %10s %10.5f %10.5f, mac %5s tc %d",packet_info.name(HDR_CMN(p)->ptype()),NOW,HDR_CMN(p)->ts_queue,((Mac802_11*)target_)->name(),qtodq); //getchar();

                        switch (HDR_CMN(p)->ptype()) {
                                case PT_MESSAGE:
                                        ++(tab[qtodq].que_->tout_message);
                                        break;
                                case PT_ARP:
                                        ++(tab[qtodq].que_->tout_ARP);
                                        break;
                                case PT_TCP:
                                        ++(tab[qtodq].que_->tout_tcp);
                                        break;
                                case PT_ACK:
                                        ++(tab[qtodq].que_->tout_ack);
                                        break;
                                case PT_VBR:
                                case PT_CBR:
                                        ++(tab[qtodq].que_->tout_udp);
                                        break;
                                case PT_VOIP:
                                        ++(tab[qtodq].que_->tout_voip);
                                        break;	
                        }
//fprintf(stderr," ###### burst %d\n",tab[qtodq].tx_burst);
                        tab[qtodq].tx_burst--;
                        if (tab[qtodq].tx_burst >= 0) {
//fprintf(stderr,"before decr %d",((Mac802_11*)target_)->tc[qtodq]->ptosend);
                                ((Mac802_11*)target_)->decr_ptosend(qtodq);                     // serve se frame succesivo al primo (poco probabile)
                                ((Mac802_11*)target_)->set_burst(qtodq,tab[qtodq].tx_burst+1);  // serve se primo frame del burst
                        } else {
                                ((Mac802_11*)target_)->set_burst(qtodq,1);    // il frame successivo (se c'e') potrebbe avere indirizzo
                                ((Mac802_11*)target_)->reset_txop(qtodq);     // diverso, per cui stoppo il burst corrente
//fprintf(stderr," resettato txop\n");
                        }
			drop(p);
                        p = deque();
                }
                if (p != 0 ) {
                        target_->recv(p, &qbs_);        // If the queue is not empty, we pass the next frame to the MAC layer

                } else {
                        block[qtodq] = 0;
                }

        }
}

/*###################################################################
#                                                                   #
# calc_bursting: computes the length of the next burst              #
#                                                                   #
#    Returns the length of the next burst                           #
#                                                                   #
###################################################################*/

int BS_DropTail::calc_bursting() {

        if (!bursting) {                                // If bursting is not activated, return 1 if the queue
                if (tab[qtodq].que_->length() > 0) {    // is not empty, 0 otherwise
                        return 1;
                } else {
                        return 0;
                }
        } else {
                int i, same = 0;
                for (i=0 ; i < tab[qtodq].que_->length() ; i++) {       // This cycle checks how many packets are destined to the
                                                                        // same node, starting from the head, since the frames
                                                                        // belonging to the same burst must have a common destination

                        if ( (u_int32_t)ETHER_ADDR(HDR_MAC802_11(tab[qtodq].que_->lookup(0))->dh_da) != MAC_BROADCAST
                                        && (u_int32_t)ETHER_ADDR(HDR_MAC802_11(tab[qtodq].que_->lookup(i))->dh_da) ==
                                        (u_int32_t)ETHER_ADDR(HDR_MAC802_11(tab[qtodq].que_->lookup(0))->dh_da) ) {
                                same++;
                        } else {
                                break;          // Stop to increment the 'same' counter when a broadcast frame is found, or whether
                        }                       // the MAC destination address of the 'i' frame is different from that of the packet
                                                // at the head of the queue
                }
                if (bursting == -1) {           // In this modality, return 'same', because the burst length can be limited
                        return same;            // only by TXOPlimit
                } else {
                        if (same < bursting) {          // With a fixed burst length, return the minimum between 'same' and 'bursting'
                                return same;
                        } else {
                                return bursting;
                        }
                }
        }
}

/*###################################################################
#                                                                   #
# handle: called by Mac802_11 to obtain the next frame for this     #
#         Access Category                                           #
#                                                                   #
#        e: the number of the requesting Access Category            #
#                                                                   #
###################################################################*/

void BS_QueueHandler::handle(Event* e) {
   queue_.qtodq = (int)e;                     // Sets the queue (i.e. the 802.11 Access Category) to be served
   queue_.resume();
}

/*###################################################################
#                                                                   #
# stop_bursting: called by Mac802_11 when the current burst must be #
#                ended, due to an error transmission or to the      #
#                TXOPlimit overcoming                               #
#                                                                   #
#                n: the identifier 802.11e Access Category          #
#                                                                   #
###################################################################*/

void BS_QueueHandler::stop_bursting(int n) {
   queue_.tab[n].tx_burst = 0;
}

// ---------- HCFstart
/*###################################################################
#                                                                   #
# queueLength: called by the Mac802_11e virtual queues in order     #
#              to know the number of packets waiting to be sent.    #
#                                                                   #
#              ac: the AC requesting the information                #
#                                                                   #
###################################################################*/

int BS_QueueHandler::queueLength(int ac) {
   return(queue_.tab[ac].que_->length());
}

/*###################################################################
#                                                                   #
# queueLength: called by the Mac802_11e virtual queues in order     #
#              to know the number of packets waiting to be sent.    #
#                                                                   #
#              ac: the AC requesting the information                #
#                                                                   #
###################################################################*/

int BS_QueueHandler::queueByteLength(int ac) {
   return(queue_.tab[ac].que_->byteLength());
}

// ---------- HCFend

//----------------------------------------------------------------------------------------------------------------------------//EndEDCFpatch
