/************************************************************
 *  Universit degli Studi di Trento - Facolt di scienze
 *
 *  Author:  Paolo Larcheri
 *  Date:    May 2005
 *  
*************************************************************/
// HCF v3.0

#include <utility>
#include <vector>

using namespace std;

// This type is used to determine how interarrival time is chosen
typedef enum {
   INTERVAL_CONST    =  0,
   INTERVAL_EXPO
} t_interval;

// This type is used to determine how the packet size is chosen
typedef enum {
   PKTSIZE_EQDIST    =  0,
   PKTSIZE_EXPO,
   PKTSIZE_TWOVAL
} t_pktsize;

#define VBR_PKTSIZE_THRESHOLD    1460

#define VBR_DELTA  0.0001
#define VBR_EQ_ONE(x) ((x > 1.0-VBR_DELTA) && (x < 1.0+VBR_DELTA))  // our "equal to one"
#define VBR_GT_ONE(x) (x > 1.0+VBR_DELTA)                           // our "greater than one"
#define VBR_LT_ONE(x) (x < 1.0-VBR_DELTA)                           // our "less than one"

class VBR_state;
typedef struct{VBR_state* state; double prob;} VBR_transition;
typedef vector<VBR_transition>::iterator  transition_it;

class VBR_state {
   friend class VBR_Traffic;
public:
   VBR_state(int index_, long seed_);
   virtual ~VBR_state();

   // adds a new transition
   bool addTransition(VBR_state* sta, double prob);  
   // checks if the sum of the probability is less or equal to 1
   // and eventually evaluates the probability of the self-transition
   bool checkTransitions();
   // according to the transition probabilities 
   // returns the state we are moving to
   VBR_state* getNextState();
   // return the size of the packet that is about to be sent
   // according to the distribution of this state
   int getPktSize();
   // return the packet interval for the next packet
   // according to the distribution of this state
   double getInterval();
   // Return the mean data rate of this state
   int getMeanDataRate();

   // Return the avg interarrival time (sec) and packet size (bytes)
   double getAvgInterval();
   int    getAvgPktSize();

   // in Tcl once end-config has been called, no further change is allowed.
   // This variable must be set to false after end-config invocation.
   // Note: it's public in order to keep this class as simple as possible.
   bool writeable;

   // tells what distribution to use for interarrival and packet size
   // Note: are public in order to keep this class as simple as possible.
   t_interval interval_dist;
   t_pktsize  pktsize_dist;

   // variable used for interarrival
   // Note: are public in order to keep this class as simple as possible.
   double const_interval;  // used for INTERVAL_CONST
   double mean_interval;   // used for INTERVAL_EXPO

   // variable used for packetsize
   // Note: are public in order to keep this class as simple as possible.
   int size1;           // used for PKTSIZE_TWOVAL and PKTSIZE_EQDIST
   int size2;           // used for PKTSIZE_TWOVAL and PKTSIZE_EQDIST
   double prob1;        // used for PKTSIZE_TWOVAL
   double prob2;        // used for PKTSIZE_TWOVAL
   int mean_pktsize;    // used for PKTSIZE_EXPO

protected:
   int index;
   long seed;
   vector<VBR_transition> transitions;
};

