/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*-
 *
 * Copyright (c) 1997 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Header: /nfs/jade/vint/CVSROOT/ns-2/mac/mac-802_11.cc,v 1.40 2002/06/06 22:54:42 haldar Exp $
 *
 * Ported from CMU/Monarch's code, nov'98 -Padma.
 */
// HCF v3.0
//------------------------------------------------------------------------------------------------------------------------------//EDCFpatch
#include "delay.h"
#include "connector.h"
#include "packet.h"
#include "random.h"
#include "mobilenode.h"

#include "arp.h"
#include "ll.h"
#include "mac.h"
#include "mac-timers.h"
#include "mac-802_11.h"
#include "cmu-trace.h"

//#define DEBUGACTIVE
#define DEBUGTIME 1

#ifdef DEBUGACTIVE //=============================================================
FILE* macLog = fopen("./mac.log", "w");
#endif //=========================================================================

static PHY_MIB PMIB =
{
	DSSS_CWMin, DSSS_CWMax, DSSS_SlotTime, DSSS_CCATime,
	DSSS_RxTxTurnaroundTime, DSSS_SIFSTime, DSSS_PreambleLength,
	DSSS_PLCPHeaderLength, DSSS_PLCPDataRate
};

static MAC_MIB MMIB =
{
	MAC_RTSThreshold, MAC_ShortRetryLimit,
	MAC_LongRetryLimit, MAC_FragmentationThreshold,
	MAC_MaxTransmitMSDULifetime, MAC_MaxReceiveLifetime,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

static class Mac802_11Class : public TclClass {
public:
	Mac802_11Class() : TclClass("Mac/802_11") {}
	TclObject* create(int, const char*const*) {
		return (new Mac802_11());
	}
} class_mac802_11;

// ---------- HCFstart
Mac802_11::Mac802_11() : Mac(), mhIF_(this), mhNav_(this), mhRecv_(this), mhCfb_(this), mhCfp_(this), mhCap_(this), mhCfpStart_(this), mhCapStart_(this) {
// ---------- HCFend

   Tcl& tcl = Tcl::instance();
   tcl.evalf("Mac/802_11 set basicRate_");
   if (strcmp(tcl.result(), "0") != 0)
      bind_bw("basicRate_", &basicRate_);     // Control frames transmission rate: set in Tcl
   else
      basicRate_ = bandwidth_;

   // XXX v2.0Start
   // NOTE: for all the stations, the
   //       following is overridden by
   //       Tcl command 'set-bitrate'
   //       in order to implement multirate.
   //       This stays true for the HC, whose
   //       dataRate_ however depends on the
   //       frame destination data rate.
   // XXX v2.0End
   tcl.evalf("Mac/802_11 set dataRate_");
   if (strcmp(tcl.result(), "0") != 0)
      bind_bw("dataRate_", &dataRate_);       // Data frames transmission rate: set in Tcl
   else
      dataRate_ = bandwidth_;

        sta_seqno_ = 1;
   cache_ = 0;
   cache_node_count_ = 0;

   err_frame = 0;
   txop = -1;
   sta_coll = 2;
   detect_coll = 0;
   virtcoll = 0;

   tx_active_ = -1;
   nav_ = 0.0;
   rx_state_ = MAC_IDLE;
   pktRx_ = 0;
   num_tc = 0;
   sifs_ = DSSS_SIFSTime;          // 10us
	// ---------- HCFstart
	pifs_ = sifs_ + DSSS_SlotTime;	// 30us
	capture_enabled = true;
	// ---------- HCFend

	eifs_ = sifs_ + (8 * ETHER_ACK_LEN2 / DSSS_PLCPDataRate) + 0.000050;    // 802.11 Standard (9.2.10)

	// ---------- HCFstart
	pktMNG_ = 0;

	legacy_ = 0;
	
	cap_ = -1;
	capWaiting_ = 0;

	cfp_ = 0;
	cfpWaiting_ = 0;
	cfpDurRemaining_ = 0;

	cfb_ = -1;
	txopDuration_ = 0;
	cfbFailed_ = 0;

   // XXX v2.0Start
   for(int i=0; i<MAX_STA; i++) {
      bitRates_[i] = 0.0;                       // bit/sec
      tspecs_[i].CFP_TXOPLmt = -1.0;            // sec
      tspecs_[i].TimeNotUsed = -1.0;            // sec
      tspecs_[i].RemainingData = 0;             // # of MPDUs left in queue
      tspecs_[i].MinimumTXOPGuaranteed = -1.0;  // sec
      tspecs_[i].NominalMSDUSize = 0;           // bytes
      tspecs_[i].MaximumMSDUSize = 0;           // bytes
      tspecs_[i].MaximumServiceInterval = -1.0; // sec
      tspecs_[i].MinimumDataRate = 0;           // bit/sec
      tspecs_[i].MeanDataRate = 0;              // bit/sec
      tspecs_[i].PeakDataRate = 0;              // bit/sec
   }
   // XXX v2.0end
   // ---------- HCFend
}

int Mac802_11::command(int argc, const char*const* argv)
{
	if (argc==2) {
		if (strcmp(argv[1], "dump_tc") == 0) {          // A command for reading some variables at
			dump_tc();                              // the end of the simulation
			return (TCL_OK);
		}
		// ---------- HCFstart
		// This command returns this node's MAC address
		if(strcmp(argv[1], "mac-addr") == 0) {
			Tcl& tcl = Tcl::instance();
			tcl.resultf("%d", index_);
			return TCL_OK;
		}
		// This command identifies this node as
		// a legacy (non QoS) station, so it won't
		// answer to HCF Contention Control frames
		if(strcmp(argv[1], "set-legacy") == 0) {
			legacy_ = 1;			// Set the legacy flag
			return TCL_OK;
		}
		// ---------- HCFend
	}
   if (argc == 3) {
		if (strcmp(argv[1], "log-target") == 0) {
			logtarget_ = (NsObject*) TclObject::lookup(argv[2]);
			if(logtarget_ == 0)
				return TCL_ERROR;
			return TCL_OK;
		}
      // This command, invoked in 'ns-2.1b9a/tcl/lib/ns-mobilenode.tcl',
      // sets the number of wireless nodes (including the base station)
      else if(strcmp(argv[1], "nodes") == 0) {      
			if(cache_) return TCL_ERROR;             
			cache_node_count_ = atoi(argv[2]);
			cache_ = new Host[cache_node_count_ + 1];
			assert(cache_);
			bzero(cache_, sizeof(Host) * (cache_node_count_+1 ));
			return TCL_OK;
		}
		// XXX v2.0Start
		// Sets the station data transmission rate.
		// Overrides datarate_ binding with Tcl in
		// order to implement multirate
		if(strcmp(argv[1], "set-bitrate") == 0) {
			dataRate_ = atof(argv[2]);
			printf("STA %d bitrate: %f\n", index_, dataRate_);
			return TCL_OK;
		}
		// XXX v2.0End
		// --- HCF start -------------
		// This command enable/disable capture
		// at PHY layer
		if(strcmp(argv[1], "capture") == 0) {
			if (strcmp(argv[2], "0") == 0)
				capture_enabled = false;
			else if (strcmp(argv[2], "1") == 0)
				capture_enabled = true;
			return TCL_OK;
		}
		// --- HCF end ---------------
	}
   if (argc == 4) {
      // Command used in Tcl in order to set the number of virtual
      if (strcmp(argv[1], "init-mac") == 0)
      {         
         set_queue(atoi(argv[2])); // queues (i.e. the number of 802.11e Access Categories) and
         TXOPLmt = atof(argv[3]);  // the value for the 'TXOPLmt' parameter, used during bursting
         return (TCL_OK);
      }
      // XXX HCF v3.0 Start
      // Command used in Tcl in order to disable the transmission
      // of the data of a specified AC during EDCA access scheme
      if (strcmp(argv[1], "edca-tx") == 0) {
         int ac = atoi(argv[2]);
         if (ac < 0 || ac >= MAX_TC)
            return (TCL_ERROR);
         if (strcmp(argv[3], "off") == 0)
            tc[ac]->edcaTX(false);
         else if (strcmp(argv[3], "on") == 0)
            tc[ac]->edcaTX(true);
         else 
            return (TCL_ERROR);
         return (TCL_OK);
      }
      // XXX v3.0End
   }
   else if (argc == 8)
   {
      // This command sets the EDCF parameters of each virtual queue
      if (strcmp(argv[1], "set-qos") == 0)
      {
         // index      // AIFS       // CWmin      // SRL      // LRL          // CWmax
         set_qos(atoi(argv[2]),atof(argv[3]),atoi(argv[4]),atoi(argv[5]),atoi(argv[6]),atoi(argv[7]));
         return (TCL_OK);
      }
        }
   return Mac::command(argc, argv);
}

/*###################################################################
#                                                                   #
# dump_tc: dumps some variables associated to each virtual queue    #
#          and calls Tcl for statistics                             #
#                                                                   #
###################################################################*/

void Mac802_11::dump_tc() {
        for (int i=0 ; i<num_tc ; i++) {
		double cfb_avg;
		if(tc[i]->cfb_num > 0)
			cfb_avg = tc[i]->cfb_time / tc[i]->cfb_num;
		else
			cfb_avg = 0;
         Tcl::instance().evalf("catch \"802_11e_tc %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %f %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %f %d %d %d %d %d %d %d\"",
                        i, (int)tc[i]->rcv_ifq_tcp, (int)tc[i]->rcv_ifq_ack, (int)tc[i]->rcv_ifq_message, (int)tc[i]->rcv_ifq_ARP, (int)tc[i]->rcv_ifq_udp, (int)tc[i]->rcv_ifq_voip, (int)tc[i]->rcv_ifq_mac_cts, (int)tc[i]->rcv_ifq_mac_rts, (int)tc[i]->rcv_ifq_mac_ack,
                        (int)tc[i]->send_erm_tcp, (int)tc[i]->send_erm_ack, (int)tc[i]->send_erm_message, (int)tc[i]->send_erm_ARP, (int)tc[i]->send_erm_udp, (int)tc[i]->send_erm_voip, (int)tc[i]->send_erm_mac_cts, (int)tc[i]->send_erm_mac_rts, (int)tc[i]->send_erm_mac_ack,
                        (int)tc[i]->rcv_chn_tcp, (int)tc[i]->rcv_chn_ack, (int)tc[i]->rcv_chn_message, (int)tc[i]->rcv_chn_ARP, (int)tc[i]->rcv_chn_udp, (int)tc[i]->rcv_chn_voip, (int)tc[i]->rcv_chn_mac_cts, (int)tc[i]->rcv_chn_mac_rts, (int)tc[i]->rcv_chn_mac_ack,
                        (int)tc[i]->rcv_chn_my_tcp, (int)tc[i]->rcv_chn_my_ack, (int)tc[i]->rcv_chn_my_message, (int)tc[i]->rcv_chn_my_ARP, (int)tc[i]->rcv_chn_my_udp, (int)tc[i]->rcv_chn_my_voip, (int)tc[i]->rcv_chn_my_mac_cts, (int)tc[i]->rcv_chn_my_mac_rts, (int)tc[i]->rcv_chn_my_mac_ack,
                        (int)tc[i]->rcv_chn_other_tcp, (int)tc[i]->rcv_chn_other_ack, (int)tc[i]->rcv_chn_other_message, (int)tc[i]->rcv_chn_other_ARP, (int)tc[i]->rcv_chn_other_udp, (int)tc[i]->rcv_chn_other_voip, (int)tc[i]->rcv_chn_other_mac_cts, (int)tc[i]->rcv_chn_other_mac_rts, (int)tc[i]->rcv_chn_other_mac_ack,
                        (int)tc[i]->rcv_chn_err_tcp, (int)tc[i]->rcv_chn_err_ack, (int)tc[i]->rcv_chn_err_message, (int)tc[i]->rcv_chn_err_ARP, (int)tc[i]->rcv_chn_err_udp, (int)tc[i]->rcv_chn_err_voip, (int)tc[i]->rcv_chn_err_mac_cts, (int)tc[i]->rcv_chn_err_mac_rts, (int)tc[i]->rcv_chn_err_mac_ack,
                        (int)tc[i]->rcv_chn_col_tcp, (int)tc[i]->rcv_chn_col_ack, (int)tc[i]->rcv_chn_col_message, (int)tc[i]->rcv_chn_col_ARP, (int)tc[i]->rcv_chn_col_udp, (int)tc[i]->rcv_chn_col_voip, (int)tc[i]->rcv_chn_col_mac_cts, (int)tc[i]->rcv_chn_col_mac_rts, (int)tc[i]->rcv_chn_col_mac_ack,
                        (double)tc[i]->col, (int)tc[i]->virt_col, (int)virtcoll,
                        (int)tc[i]->busy_tcp, (int)tc[i]->busy_ack, (int)tc[i]->busy_message, (int)tc[i]->busy_ARP, (int)tc[i]->busy_udp, (int)tc[i]->busy_voip, (int)tc[i]->busy_mac_cts, (int)tc[i]->busy_mac_rts, (int)tc[i]->busy_mac_ack,
                        (int)tc[i]->max_rtx_tcp, (int)tc[i]->max_rtx_ack, (int)tc[i]->max_rtx_message, (int)tc[i]->max_rtx_ARP, (int)tc[i]->max_rtx_udp, (int)tc[i]->max_rtx_voip, (int)tc[i]->max_rtx_mac_cts, (int)tc[i]->max_rtx_mac_rts, (int)tc[i]->max_rtx_mac_ack,
                        (int)tc[i]->rtx_tcp, (int)tc[i]->rtx_ack, (int)tc[i]->rtx_message, (int)tc[i]->rtx_ARP, (int)tc[i]->rtx_udp, (int)tc[i]->rtx_voip, (int)tc[i]->rtx_mac_cts, (int)tc[i]->rtx_mac_rts, (int)tc[i]->rtx_mac_ack,
                        (int)tc[i]->send_ll_tcp, (int)tc[i]->send_ll_ack, (int)tc[i]->send_ll_message, (int)tc[i]->send_ll_ARP, (int)tc[i]->send_ll_udp, (int)tc[i]->send_ll_voip, (int)tc[i]->send_ll_mac_cts, (int)tc[i]->send_ll_mac_rts, (int)tc[i]->send_ll_mac_ack,
                        (int)tc[i]->dupl_tcp, (int)tc[i]->dupl_ack, (int)tc[i]->dupl_message, (int)tc[i]->dupl_ARP,
			(int)tc[i]->dupl_udp, (int)tc[i]->dupl_voip, (int)tc[i]->dupl_mac_cts, (int)tc[i]->dupl_mac_rts,
			(int)tc[i]->dupl_mac_ack, (double)cfb_avg, (int)tc[i]->cfb_num, (int)tc[i]->cfb_num_bad_end, (int)tc[i]->cfb_data_fast_recv, (int)tc[i]->cfb_data_slow_recv, (int)tc[i]->cfb_voice_fast_recv, (int)tc[i]->cfb_voice_slow_recv, (int)tc[i]->cfb_noDataToSend);
        }
}


/*###################################################################
#                                                                   #
# recv: the main receiving routine, called every time a new frame   #
#       comes. It reads the common header of the new frame to know  #
#       if it is coming from the up-layer (Interface Queue) or from #
#       the down-layer (Error Model)                                #
#                                                                   #
#         p: the new incoming packet                                #
#                                                                   #
###################################################################*/

void Mac802_11::recv(Packet *p, Handler *h) {

	struct hdr_cmn *hdr = HDR_CMN(p);

	if(hdr->direction() == hdr_cmn::DOWN) {         // Packet coming from the Interface Queue
                int tc_send = hdr->mac_tc;        // Reads the Traffic Category from the common header. 'mac_tc' must be set by
                assert(initialized(tc_send));     // the LLC layer

                switch (hdr->ptype()) {         // Statistics: frames received from the Interface Queue
                        case PT_TCP:
                                (tc[tc_send]->rcv_ifq_tcp)++;
                                HDR_CMN(p)->tmac1 = NOW;
                                break;
                        case PT_ACK:
                                (tc[tc_send]->rcv_ifq_ack)++;
                                break;
                        case PT_MESSAGE:
                                (tc[tc_send]->rcv_ifq_message)++;
                                break;
                        case PT_ARP:
                                (tc[tc_send]->rcv_ifq_ARP)++;
                                break;
                        case PT_VBR:
                        case PT_CBR:
                                (tc[tc_send]->rcv_ifq_udp)++;
                                HDR_CMN(p)->tmac1 = NOW;
                                break;
                        case PT_VOIP:
                                (tc[tc_send]->rcv_ifq_voip)++;
                                HDR_CMN(p)->tmac1 = NOW;
                                break;	
                        case PT_MAC_RTS:
                                (tc[tc_send]->rcv_ifq_mac_rts)++;
                                break;
                        case PT_MAC_CTS:
                                (tc[tc_send]->rcv_ifq_mac_cts)++;
                                break;
                        case PT_MAC_ACK:
                                (tc[tc_send]->rcv_ifq_mac_ack)++;
                                break;
                }

                /*
                * drop the packet if the node is in sleep mode
                * sleep mode can't stop node from sending packets
                */
                EnergyModel *em = netif_->node()->energy_model();
                if (em && em->sleep()) {
                        em->set_node_sleep(0);
                        em->set_node_state(EnergyModel::INROUTE);
                }

                tc[tc_send]->send(p);     // Forwards the frame to the corresponding virtual queue
                return;
        }

        // Packets coming from the Error Model, destined to Transport Layers

	hdr_mac802_11 *mh = HDR_MAC802_11(p);
	u_int8_t type = mh->dh_fc.fc_type;
	u_int8_t subtype = mh->dh_fc.fc_subtype;

        int nt;
        if (num_tc == 1) {
                nt = 0;
        } else {
                nt = HDR_CMN(p)->mac_tc;    // Extracts the Traffic Category from the common header
        }

        switch (hdr->ptype()) {         // Statistics: packets received by the Error Model
                case PT_TCP:
                        (tc[nt]->rcv_chn_tcp)++;
                        //fprintf(stderr, "TCP (%d) --- SRC:%d    DST:%d\n", HDR_CMN(p)->size(), ETHER_ADDR(mh->dh_sa), ETHER_ADDR(mh->dh_da));
                        break;
                case PT_ACK:
                        (tc[nt]->rcv_chn_ack)++;
                        //fprintf(stderr, "ACK_TCP (%d) --- SRC:%d    DST:%d\n", HDR_CMN(p)->size(), ETHER_ADDR(mh->dh_sa), ETHER_ADDR(mh->dh_da));
                        break;
                case PT_MESSAGE:
                        (tc[nt]->rcv_chn_message)++;
                        break;
                case PT_ARP:
                        (tc[nt]->rcv_chn_ARP)++;
                        break;
                case PT_VBR:
                case PT_CBR:
                        (tc[nt]->rcv_chn_udp)++;
                        break;
                case PT_VOIP:
                        (tc[nt]->rcv_chn_voip)++;
                        break;	
                case PT_MAC_RTS:
                        (tc[nt]->rcv_chn_mac_rts)++;
                        break;
                case PT_MAC_CTS:
                        (tc[nt]->rcv_chn_mac_cts)++;
                        break;
                case PT_MAC_ACK:
                        (tc[nt]->rcv_chn_mac_ack)++;
                        break;
        }

	/*
	 *  Handle incoming packets.
	 *
	 *  We just received the 1st bit of a packet on the network
	 *  interface.
	 *
	 */

        /*
	 *  If the interface is currently in transmit mode, then
	 *  it probably won't even see this packet.  However, the
	 *  "air" around me is BUSY so I need to let the packet
	 *  proceed.  Just set the error flag in the common header
	 *  to that the packet gets thrown away.
	 */

	if(tx_active_ != -1 && hdr->error() == 0) {
		hdr->error() = 1;
		// ---------- HCFstart
		// If a CC frame collides, the QSTAs
		// will not receive it, so skip the CAP
		// stopping the CAP timer and returning
		// to normal EDCF access procedure
		if(tc[tx_active_]->tx_state_ == MAC_CC) {
			mhCap_.stop();
			cap_ = -1;
			// Reset the NAV
			if(mhNav_.busy()) {
				mhNav_.stop();
			}
			nav_ = NOW;
			for(int i=0; i<num_tc; i++) {
				if (tc[i]->edcaTXenabled()) {
					tc[i]->set_tx_state(MAC_IDLE);
					tc[i]->resumeTC();
				}
			}
		}
		// ---------- HCFend
	}

	if(rx_state_ == MAC_IDLE) {             // No collision

		set_rx_state(MAC_RECV);
		pktRx_ = p;

		/*
		 * Schedule the reception of this packet, in
		 * txtime seconds.
		 */
		mhRecv_.start(txtime(p));       // This timer simulates the time necessary for receiving all bits of the
                                                // incoming packet

	} else {        // Collision

                if (tx_active_ == -1)  {        // The following check is usefull to correctly count the number of collisions

                        if (detect_coll == 0) {         // This is the first colliding frame, so two stations are colliding and
                                                        //  (N - sta_coll) stations will detect the collision
                                sta_coll = 2;
                                (tc[0]->col) += (1.0 / (double)(cache_node_count_ - sta_coll));
                        } else {
                                                // There are more colliding frames, so 'sta_coll' is incremented and
                                                // (N - sta_coll) stations will detect this collision
                                sta_coll++;
                                (tc[0]->col) -= (1.0 / (double)(cache_node_count_ - sta_coll + 1));
                                (tc[0]->col) += (1.0 / (double)(cache_node_count_ - sta_coll));
                        }
                        detect_coll = 1;
                }
		 /*
                 *  If the power of the incoming packet is smaller than the
                 *  power of the packet currently being received by at least
                 *  the capture threshold, then we ignore the new packet.
                 */
                if((capture_enabled) && (pktRx_->txinfo_.RxPr / p->txinfo_.RxPr >= p->txinfo_.CPThresh)) {     // The power of the incoming packet is low 
                        switch (hdr->ptype()) {         // Statistics: these frames are considered as destined to others nodes
                                case PT_TCP:
                                        (tc[nt]->rcv_chn_other_tcp)++;
                                        break;
                                case PT_ACK:
                                        (tc[nt]->rcv_chn_other_ack)++;
                                        break;
                                case PT_MESSAGE:
                                        (tc[nt]->rcv_chn_other_message)++;
                                        break;
                                case PT_ARP:
                                        (tc[nt]->rcv_chn_other_ARP)++;
                                        break;
                                case PT_VBR:
                                case PT_CBR:
                                        (tc[nt]->rcv_chn_other_udp)++;
                                        break;
                                case PT_VOIP:
                                        (tc[nt]->rcv_chn_other_voip)++;
                                        break;
                                case PT_MAC_RTS:
                                        (tc[nt]->rcv_chn_other_mac_rts)++;
                                        break;
                                case PT_MAC_CTS:
                                        (tc[nt]->rcv_chn_other_mac_cts)++;
                                        break;
                                case PT_MAC_ACK:
                                        (tc[nt]->rcv_chn_other_mac_ack)++;
                                        break;
                        }
                        capture(p);
                } else {
                        collision(p,nt);
                }
	}
}

/*###################################################################
#                                                                   #
# initialized: checks if the 'n' virtual queue is correctly         #
#              configured                                           #
#                                                                   #
#         n: the index of the virtual queue                         #
#                                                                   #
#    returns 1 if all is OK, 0 otherwise                            #
#                                                                   #
###################################################################*/

int Mac802_11::initialized(int n) {
        return (tc[n]->phymib_ && tc[n]->macmib_ && cache_ && logtarget_ && Mac::initialized());
}

/*###################################################################
#                                                                   #
# set_rx_state: changes the receiving state                         #
#                                                                   #
#         x: the new receiving state value (MAC_RECV, MAC_COLL or   #
#            MAC_IDLE)                                              #
#                                                                   #
###################################################################*/

void Mac802_11::set_rx_state(MacState x) {
        rx_state_ = x;
        for (int i=0 ; i<num_tc ; i++) {
                check_backoff_timer(i);         // The receiving state is changed, so we must check whether it is
        }                                       // necessary to stop/start the backoff timers of all virtual queues
}

/*###################################################################
#                                                                   #
# check_backoff_timer: starts/stops the backoff timer of the 'n'    #
#                      virtual queue                                #
#                                                                   #
#         n: the index of the virtual queue                         #
#                                                                   #
###################################################################*/

void Mac802_11::check_backoff_timer(int n) {

        // ---------- HCFstart
	// Not necessary during
	// CFP, CAP or CFB
	if(cfp_ ||
	   cap_ != -1 ||
	   cfb_ != -1)
	   	return;
	// ---------- HCFend
	
	if(tc[n]->is_idle() && tc[n]->mhBackoff_.busy() && tc[n]->mhBackoff_.paused()) {        // The channel is now idle, so we can
                                                                                                // start again the backoff timer
                if (err_frame == 0) {
                        tc[n]->mhBackoff_.resume(tc[n]->difs_);         // Last frame was correctly received  ---> wait AIFS
                                                                        // before decrementing backoff
                } else {
                        tc[n]->mhBackoff_.resume(eifs_);        // Last frame was not correctly received  ---> wait EIFS
                }                                               // before decrementing backoff
        }

        if(!tc[n]->is_idle() && tc[n]->find_idle) {     // In this case, the channel was idle when we received the frame from the
                                                        // Interface Queue, but now it is busy, so we must start the backoff procedure

                tc[n]->find_idle = 0;
                tc[n]->mhDefer_.stop();                                 // Stop deferring and starts a new backoff. The contention window
                tc[n]->mhBackoff_.start(tc[n]->cw_,tc[n]->is_idle());   // is not incremented because it is the first time we do backoff
        }

        if(!tc[n]->is_idle() && tc[n]->mhBackoff_.busy() && !tc[n]->mhBackoff_.paused() &&
                        tc[n]->virtual_coll == 0 && txop != n) {
                                                                  // The virtual queue is doing backoff, but the channel state has
                tc[n]->mhBackoff_.pause();                        // become busy, so we must stop the backoff timer
        }
}

/*###################################################################
#                                                                   #
# set_qos: sets the EDCF parameters of the 'n' virtual queue        #
#                                                                   #
#           n: the index of the virtual queue                       #
#           f: AIFS value                                           #
#           w: CWmin value                                          #
#         srl: SRL value                                            #
#         lrl: LRL value                                            #
#        wmax: CWmax value                                          #
#                                                                   #
###################################################################*/

void Mac802_11::set_qos(int n, double f, int w, int srl, int lrl, int wmax) {
        tc[n]->difs_ = f;
        tc[n]->cw_ = w;
        tc[n]->CWmin = (u_int32_t)w;
        tc[n]->CWmax = (u_int32_t)wmax;
        tc[n]->ShortRL = (u_int32_t)srl;
        tc[n]->LongRL = (u_int32_t)lrl;                 // The following call informs Tcl about the EDCF parameters used during
        tc[n]->nq = n;                                  // the simulation

        Tcl::instance().evalf("catch \"MAC_set %s %d %f %f %f %f %d %d %d %d\"", name(),(int)index_,basicRate_,dataRate_,
                tc[n]->difs_,eifs_,tc[n]->CWmin,tc[n]->CWmax,tc[n]->ShortRL,tc[n]->LongRL);
}

/*###################################################################
#                                                                   #
# set_queue: creates the EDCF Access Categories objects             #
#                                                                   #
#           n: the number of virtual queues                         #
#                                                                   #
###################################################################*/

void Mac802_11::set_queue(int n) {
        for (int i=0; i<n; i++) {
                tc[i] = new Mac802_11e(this,&PMIB, &MMIB);
        }
        num_tc = n;
}

/*###################################################################
#                                                                   #
# set_callback: called by drop-tail.cc in order to set the Handler  #
#               between Mac802_11 and BS_DropTail                   #
#                                                                   #
#           h: a pointer to the Handler (BS_QueueHandler object)    #
#                                                                   #
###################################################################*/

void Mac802_11::set_callback(Handler *h) {
        callback_ = h;
}

/*###################################################################
#                                                                   #
# set_burst: called by drop-tail.cc in order to inform the MAC      #
#               about the length of the next burst                  #
#                                                                   #
#           n: the index of the corresponding virtual queue         #
#         len: the number of frames forming the next burst          #
#                                                                   #
###################################################################*/

void Mac802_11::set_burst(int n, int len) {
        tc[n]->bursting = len;
}

void Mac802_11::decr_ptosend(int n) {
        (tc[n]->ptosend)--;
}

void Mac802_11::reset_txop(int n) { 
        txop = -1; 
        tc[n]->ptosend = 0;
}

/*###################################################################
#                                                                   #
# txtime: plucks the precomputed transmission time from the         #
#         packet header of the frame 'p'                            #
#                                                                   #
#           p: the new packet                                       #
#                                                                   #
#   Returns the transmission time                                   #
#                                                                   #
###################################################################*/

double Mac802_11::txtime(Packet *p) {
	 struct hdr_cmn *ch = HDR_CMN(p);
	 double t = ch->txtime();
	 if (t < 0.0) {
		 drop(p, "XXX");
 		exit(1);
	 }
	 return t;
}

/*###################################################################
#                                                                   #
# txtime: computes the transmission time for a new packet           #
#                                                                   #
#         psz: the size (in bytes) of the packet                    #
#         drt: the transmission rate of the packet (bps)            #
#                                                                   #
#   Returns the transmission time                                   #
#                                                                   #
###################################################################*/

double Mac802_11::txtime(double psz, double drt) {
	double dsz = psz - PLCP_HDR_LEN2;
	int plcp_hdr = PLCP_HDR_LEN2 << 3;
	int datalen = (int)dsz << 3;
	double t = (((double)plcp_hdr)/DSSS_PLCPDataRate) + (((double)datalen)/drt);
	return(t);
}

/*###################################################################
#                                                                   #
# resolve_coll: when a virtual queue ends its backoff procedure,    #
#               this method checks if there are others queues that  #
#               have finished their backoff. If a virtual collision #
#               occurs, the contention is won by the queue with the #
#               greater priority (i.e. the one with the greater     #
#               numeric index)                                      #
#                                                                   #
#           n: The index of the virtual queue that has ended its    #
#              backoff procedure                                    #
#                                                                   #
#   Returns the index of the queue that has won the contention      #
#    ('n' if there is not a virtual collision)                      #
#                                                                   #
###################################################################*/

int Mac802_11::resolve_coll(int n) {
        int how = 1;
        int i = 0;
        tc[n]->virtual_coll = 1;         // The 'n' queue can potentially collide

        for (i=0 ; i<num_tc ; i++) {
                if (i == n)
                        continue;
                tc[i]->virtual_coll = 0;

                // In the following if statement, remaining() returns the number of slots remaining before the end of the backoff
                // phase. Since this is not an integer value, we consider that the backoff is finished when it is less than 0.5, in
                // order to compensate for approximations

                if (tc[i]->mhBackoff_.busy() && !tc[i]->mhBackoff_.paused() && tc[i]->mhBackoff_.remaining() < 0.5) {
                        tc[i]->mhBackoff_.pause();
                        tc[i]->mhBackoff_.stop();        // Whit this expedient, the backoff timer is "removed" from the ns scheduler
                        tc[i]->mhBackoff_.start(0,1);    // and the backoff phase of the queue 'i' is immediately ended

                        tc[i]->virtual_coll = 1;        // The 'i' queue is virtually colliding

                        how++;          // We increment the counter of how many stations are virtually colliding
                }
        }
        int c = how - 1;
        for (i=0 ; i<num_tc && c>=0 ; i++) {
                if(tc[i]->virtual_coll == 1) {          // These instructions give priority to the queue
                        c--;                            // with the greater numeric index
                }
        }
        if (how > 1) {
                virtcoll++;                    // Statistcs: number of virtual collisions
        }
        i--;

        tc[i]->virtual_coll = 0;        // Remove the collision flag for the queue that has gained access to the medium

        return i;
}

/*###################################################################
#                                                                   #
# transmit: Called before the transmitter activation, this function #
#           gives the txop to the appropriate queue, then starts    #
#           the transmission on the physical interface              #
#                                                                   #
#           p: the packet to be transmitted                         #
#           t: the transmission timeout for this packet             #
#           n: the index of the virtual queue at which the frame    #
#              belongs                                              #
#                                                                   #
###################################################################*/

void Mac802_11::transmit(Packet* p, double t, int n) {
        hdr_cmn *ch = HDR_CMN(p);
        hdr_mac802_11* mh = HDR_MAC802_11(p);

        if (!cfp_ && tc[n]->edcaTX_status==false)
            return;

        //if ((tc[n]->pktRTS_ == 0) && (tc[n]->pktTx_ != 0) && )
        //    fprintf(stderr, "STA %d   ---- sending TCP segment at %f\n", index, NOW);

        if (txop == -1) {
	        txop = resolve_coll(n);        // Solves possible virtual collision, giving priority to the appropriate queue

                if (tc[txop]->bursting && tc[txop]->pktCTRL_ == 0) {

                        if (tc[txop]->pktRTS_ != 0 ||
                                tc[txop]->pktRTS_ == 0 && (u_int32_t) ch->size() < tc[txop]->macmib_->RTSThreshold) {
                                txopLimit_ = NOW;       // Here, the time at which the first frame is transmitted is stored
                        }
                }

                if (tc[txop]->bursting && tc[txop]->pktCTRL_ == 0 && tc[txop]->pktRTS_ == 0 && tc[txop]->pktTx_ != 0 &&
                                (u_int32_t)ETHER_ADDR((HDR_MAC802_11(tc[txop]->pktTx_))->dh_da) != MAC_BROADCAST) {

                        tc[txop]->ptosend = tc[txop]->bursting;         // In bursting mode, this sets the number of frame
                                                                        // to be sent during the next burst
                }
        }
	
        if (txop != n) {
                tc[txop]->mhBackoff_.pause();   // The queue that has obtained the txop is not 'n', consequently we must force
                tc[txop]->mhBackoff_.stop();    // the expiration of the backoff for the 'txop' queue
                tc[txop]->backoffHandler();

                hdr_cmn *ch2;           // Since the backoff of the 'n' queue ia already expired, we must immediately consider
                                        // its virtual collision here

                if (tc[n]->pktRTS_ != 0) {
                        ch2 = HDR_CMN(tc[n]->pktRTS_);
                        tc[n]->set_tx_state(MAC_RTS);

                } else if (tc[n]->pktTx_ != 0) {
                        ch2 = HDR_CMN(tc[n]->pktTx_);
                        tc[n]->set_tx_state(MAC_SEND);

                } else {                           // If the backoff is expired, but the queue has not packets to send, then
                        tc[n]->virtual_coll = 0;   // there is not a virtual collision
                        return;
                }

                tc[n]->virt_col++;      // Statistics: per-queue virtual collision counter

                switch (ch2->ptype()) {         // Statistics: there is a virtual collision, so the MAC is busy by another virtual queue
                        case PT_TCP:
                                ++(tc[n]->busy_tcp);
                                break;
                        case PT_ACK:
                                ++(tc[n]->busy_ack);
                                break;
                        case PT_MESSAGE:
                                ++(tc[n]->busy_message);
                                break;
                        case PT_ARP:
                                ++(tc[n]->busy_ARP);
                                break;
                        case PT_VBR:
                        case PT_CBR:
                                ++(tc[n]->busy_udp);
                                break;
                        case PT_VOIP:
                                ++(tc[n]->busy_voip);
                                break;	
                        case PT_MAC_RTS:
                                ++(tc[n]->busy_mac_rts);
                                break;
                        case PT_MAC_CTS:
                                ++(tc[n]->busy_mac_cts);
                                break;
                        case PT_MAC_ACK:
                                ++(tc[n]->busy_mac_ack);
                                break;
                }

                tc[n]->send_timer();    // 'send_timer()' doubles the contenetion window and reschedules the frame transmission by
                                        // starting a new backoff phase, as if there has been a real collision on the channel

                tc[n]->virtual_coll = 0;        // The 'n' virtual collision has been solved

                return;
        }

        // Here the transmission on the channel can start

      for (int i=0 ; i<num_tc ; i++) {
         if (i == n)
            continue;

         // The internal NAV of the others virtual queues is updated. We must
         // take into account for the transmission time of the frame, because
         // it is leaving the node, not coming
         tc[i]->set_int_nav(mh->dh_duration + usec(txtime(p)));
         // It is necessary to stop the backoff timers of the others virtual queues
         check_backoff_timer(i);
      }

        /*
         * If I'm transmitting without doing Carrier Sense, such as when
         * sending an ACK, any incoming packet will be "missed"
         * and hence, must be discarded.
         */

	if(rx_state_ != MAC_IDLE) {
                struct hdr_mac802_11 *dh = HDR_MAC802_11(p);
                assert(dh->dh_fc.fc_type == MAC_Type_Control);
                assert(dh->dh_fc.fc_subtype == MAC_Subtype_ACK);
                assert(pktRx_);
                struct hdr_cmn *ch = HDR_CMN(pktRx_);
		ch->error() = 1;                          // this forces packet discard
		// ---------- HCFstart
		// If a CC frame collides, the QSTAs
		// will not receive it, so skip the CAP by
		// stopping the CAP timer and returning
		// to normal EDCF access procedure
		if(dh->dh_fc.fc_type == MAC_Type_Control &&
		   dh->dh_fc.fc_subtype == MAC_Subtype_CC) {
			mhCap_.stop();
			cap_ = -1;
			// Reset the NAV
			if(mhNav_.busy()) {
				mhNav_.stop();
			}
			nav_ = NOW;
			for(int i=0; i<num_tc; i++) {
				if (tc[i]->edcaTXenabled()) {
					tc[i]->set_tx_state(MAC_IDLE);
					tc[i]->resumeTC();
				}
			}
		}
		// ---------- HCFend
        }

        switch (HDR_CMN(p)->ptype()) {          // Statistics: frames sent to the Network Interface
                case PT_TCP:
                        (tc[n]->send_erm_tcp)++;
                        HDR_CMN(p)->tmac2 = NOW;
                        break;
                case PT_ACK:
                        (tc[n]->send_erm_ack)++;
                        break;
                case PT_MESSAGE:
                        (tc[n]->send_erm_message)++;
                        break;
                case PT_ARP:
                        (tc[n]->send_erm_ARP)++;
                        break;
                case PT_VBR:
                case PT_CBR:
                        (tc[n]->send_erm_udp)++;
                        HDR_CMN(p)->tmac2 = NOW;
                        break;
                case PT_VOIP:
                        (tc[n]->send_erm_voip)++;
                        HDR_CMN(p)->tmac2 = NOW;
                        break;	
                case PT_MAC_RTS:
                        (tc[n]->send_erm_mac_rts)++;
                        break;
                case PT_MAC_CTS:
                        (tc[n]->send_erm_mac_cts)++;
                        break;
                case PT_MAC_ACK:
                        (tc[n]->send_erm_mac_ack)++;
                        break;
	}

        tx_active_ = n;         // Now the 'n' queue can transmit on the channel

	// ---------- HCFstart
	// Next check is skipped in CFP, CAP and CFB,
	// during which bursting mode is not used
	if(!cfp_ && (cfb_ == -1) && (cap_ == -1)) {
		if (tc[n]->bursting &&
		    tc[n]->pktCTRL_ == 0 &&
		    tc[n]->pktRTS_ == 0 &&
		    (u_int32_t)ETHER_ADDR((HDR_MAC802_11(tc[n]->pktTx_))->dh_da) != MAC_BROADCAST) {

                	tc[n]->ptosend --;      // In bursting modality, it is necessary to decrement the counter of how many
                         	               // packets must still be transmitted during this burst
        	}
	}
	// ---------- HCFend

        /*
         * pass the packet on the "interface" which will in turn
         * place the packet on the channel.
         *
         * NOTE: a handler is passed along so that the Network
         *       Interface can distinguish between incoming and
         *       outgoing packets.
         */

	downtarget_->recv(p->copy(), this);
   
 
   //if (HDR_CMN(p)->ptype() == PT_TCP)
   //   fprintf(stderr, "Transmitting...  (DST: %d)  (SRC: %d)\n", ETHER_ADDR(HDR_MAC802_11(p)->dh_da), ETHER_ADDR(HDR_MAC802_11(p)->dh_sa));
	mhIF_.start(txtime(p));  // Network Interface timer, corresponding to the transmitter activation

	tc[n]->mhSend_.start(t); // Timer used to wait for the transmission timeout

}

/*###################################################################
#                                                                   #
# discard: Called when a packet must be discarded. Its main goal    #
#          is to avoid that different nodes drop the same frame     #
#                                                                   #
#           p: the packet to be discarded                           #
#         why: the reason for dropping the frame                    #
#                                                                   #
###################################################################*/

void Mac802_11::discard(Packet *p, const char* why)
{
	hdr_mac802_11* mh = HDR_MAC802_11(p);
	hdr_cmn *ch = HDR_CMN(p);

	/* if the rcvd pkt contains errors, a real MAC layer couldn't
	   necessarily read any data from it, so we just toss it now */

	if(ch->error() != 0) {
		Packet::free(p);
		return;
	}

	switch(mh->dh_fc.fc_type) {
	case MAC_Type_Management:
		drop(p, why);
		return;
	case MAC_Type_Control:
		switch(mh->dh_fc.fc_subtype) {
		case MAC_Subtype_RTS:
			if((u_int32_t)ETHER_ADDR(mh->dh_sa) ==
			   (u_int32_t)index_) {
				drop(p, why);
				return;
			}

			/* fall through - if necessary */

		case MAC_Subtype_CTS:
		case MAC_Subtype_ACK:
		// ---------- HCFstart
		case MAC_Subtype_CFEnd:
		case MAC_Subtype_CC:
		case MAC_Subtype_RR:
		// ---------- HCFend
			if((u_int32_t)ETHER_ADDR(mh->dh_da) ==
			   (u_int32_t)index_) {
				drop(p, why);
				return;
			}
			break;
		default:
			fprintf(stderr, "invalid MAC Control subtype\n");
			exit(1);
		}
		break;
	case MAC_Type_Data:
		switch(mh->dh_fc.fc_subtype) {
		case MAC_Subtype_Data:
		// ---------- HCFstart
		case MAC_Subtype_QoS_Data:
		case MAC_Subtype_QoS_CFAck:
		case MAC_Subtype_QoS_CFPoll:
		case MAC_Subtype_QoS_Data_CFAck:
		case MAC_Subtype_QoS_Data_CFPoll:
		case MAC_Subtype_QoS_Data_CFAck_CFPoll:
		case MAC_Subtype_QoS_Null:
		// ---------- HCFend
			if((u_int32_t)ETHER_ADDR(mh->dh_da) ==
			   (u_int32_t)index_ ||
			   (u_int32_t)ETHER_ADDR(mh->dh_sa) ==
			   (u_int32_t)index_ ||
			   (u_int32_t)ETHER_ADDR(mh->dh_da) == MAC_BROADCAST) {
				drop(p);
				return;
			}
			break;
		default:
			fprintf(stderr, "invalid MAC Data subtype\n");
			exit(1);
		}
		break;
	default:
		fprintf(stderr, "invalid MAC type (%x)\n", mh->dh_fc.fc_type);
		trace_pkt(p);
		exit(1);
	}
	Packet::free(p);
}

/*###################################################################
#                                                                   #
# trace_pkt: used for debug, prints some statistics about           #
#            the frame 'p'                                          #
#                                                                   #
#           p: the packet to be examinated                          #
#                                                                   #
###################################################################*/

void Mac802_11::trace_pkt(Packet *p) {
	struct hdr_cmn *ch = HDR_CMN(p);
	struct hdr_mac802_11* dh = HDR_MAC802_11(p);
	u_int16_t *t = (u_int16_t*) &dh->dh_fc;

	fprintf(stderr, "\t[ %2x %2x %2x %2x ] %x %s %d\n",
		*t, dh->dh_duration,
		ETHER_ADDR(dh->dh_da), ETHER_ADDR(dh->dh_sa),
		index_, packet_info.name(ch->ptype()), ch->size());
}

/*###################################################################
#                                                                   #
# txHandler: called by mac-timers.cc when the transmission on       #
#            the medium is finished                                 #
#                                                                   #
###################################################################*/

void Mac802_11::txHandler() {
	tx_active_ = -1;        // The transmitter is now turned off
        txop = -1;              // The transmission opportunity can be released
}

/*###################################################################
#                                                                   #
# navHandler: called by mac-timers.cc when the NAV timer has        #
#             expired                                               #
#                                                                   #
###################################################################*/

void Mac802_11::navHandler() {
        for (int i=0 ; i<num_tc ; i++) {
                if(tc[i]->is_idle() && tc[i]->mhBackoff_.paused())      // Since the NAV is expired, if the channel is idle it
                                                                        // is necessary to restart the backoff
                        if (err_frame == 0) {
                                tc[i]->mhBackoff_.resume(tc[i]->difs_);         // Last frame was correctly received  ---> wait AIFS
                                                                                // before decrementing backoff
                        } else {
                                tc[i]->mhBackoff_.resume(eifs_);        // Last frame was not correctly received  ---> wait EIFS
                        }                                               // before decrementing backoff
        }
}

/*###################################################################
#                                                                   #
# recvHandler: called by mac-timers.cc when the last bit of a frame #
#              has been received                                    #
#                                                                   #
###################################################################*/

void Mac802_11::recvHandler() {
	recv_timer();
}

// ---------- HCFstart
/*###################################################################
#                                                                   #
# capHandler: transmits the Reservation Request frame during CCIs,  #
#             then waits for the end of the CCI before resuming     #
#             the transmission                                      #
#                                                                   #
###################################################################*/

void Mac802_11::capHandler() {

	// If CAP flag is set to -1, the CCI has ended
	// and all TCs must resume transmission now:
	// the resumeTC() method is used for that
	if(cap_ == num_tc){
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
			fprintf(macLog, "[%d] end of CCI %f\n", index_, NOW);
		}
#endif //=========================================================================
		txopLimit_ = 0;
		// Reset the CAP flag to -1
		// because the CCI is ended
		cap_ = -1;
		for(int i=0; i<num_tc; i++)
			if (tc[i]->edcaTXenabled())
				tc[i]->resumeTC();
		return;
	}
	
	struct hdr_mac802_11* dh = HDR_MAC802_11(tc[cap_]->pktCTRL_);

	if(dh->dh_fc.fc_type != MAC_Type_Control ||
	   dh->dh_fc.fc_subtype != MAC_Subtype_RR) {
	   	fprintf(stderr, "Sending a non-RR frame with capHandler\n");
		exit(1);
	}  	
	
	tc[cap_]->set_tx_state(MAC_RR);
	txop = cap_;
	double timeout = txtime(tc[cap_]->pktCTRL_);
	transmit(tc[cap_]->pktCTRL_, timeout, cap_);

#ifdef DEBUGACTIVE //=============================================================
	if(NOW > DEBUGTIME) { 
		fprintf(macLog, "[%d.%d] send RR %f\n", index_, cap_, NOW);
	}
#endif //=========================================================================

	// CAP flag can be reset to the same
	// unused TC number we set it to when
	// the CC frame was received now
	cap_ = num_tc;

	// Restart the CAP timer, it will exipre
	// at the end of the CCI, resuming transmission
	mhCap_.start(txopLimit_ - NOW);

}

/*###################################################################
#                                                                   #
# cfbHandler: handles QSTA's frames exchange during the CFB.        #
#             Called before and after every frame is sent during    #
#             the CFB, it transmits the frame and then checks for   #
#             the expected answer: if no response is received,      #
#             it retransmits the frame.                             #
#             It also controls if there's enough time left          #
#             in the CFB to send the frame and receive the answer   #
#                                                                   #
###################################################################*/

void Mac802_11::cfbHandler() {
	double timeout, wait;

	// debug
	if(cfb_ == -1) {
		fprintf(stderr, "[%d] CFB Timer active outside CFB\n", index_);
		exit(1);
	}

	// Sent Data, Data+CF-Ack, Cf-Ack, Null.
	// Waiting for answer from HC,
	// check if we're receiving it
	// and wait for end of reception
	if(!cfbFailed_ &&
	   tc[cfb_]->tx_state_ != MAC_IDLE) {
	   	if(rx_state_ != MAC_IDLE) {
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) { 
				fprintf(macLog, "[%d] receiving expected CF-Ack/Data/CF-Poll %f, waiting for end of reception\n", index_, NOW);
			}
#endif //=========================================================================
			return;
		}
		else {
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) { 
				fprintf(macLog, "[%d] not receiving expected CF-Ack/Data/CF-Poll %f\n     Attempting retransmission ", index_, NOW);
			}
#endif //=========================================================================
			// If Data was sent, set the retry bit
			// and update the statistics
			if(tc[cfb_]->tx_state_ == MAC_QOS_DATA ||
			   tc[cfb_]->tx_state_ == MAC_QOS_DATA_CFACK) {
				HDR_MAC802_11(tc[cfb_]->pktTx_)->dh_fc.fc_retry = 1;

				switch (HDR_CMN(tc[cfb_]->pktTx_)->ptype()) {     // Statistics: retransmitted frames
					case PT_TCP:
						++(tc[cfb_]->rtx_tcp);
						break;
					case PT_ACK:
						++(tc[cfb_]->rtx_ack);
						break;
					case PT_MESSAGE:
						++(tc[cfb_]->rtx_message);
						break;
					case PT_ARP:
						++(tc[cfb_]->rtx_ARP);
						break;
               case PT_VBR:
					case PT_CBR:
						++(tc[cfb_]->rtx_udp);
						break;
					case PT_VOIP:
						++(tc[cfb_]->rtx_voip);
						break;
					case PT_MAC_RTS:
						++(tc[cfb_]->rtx_mac_rts);
						break;
					case PT_MAC_CTS:
						++(tc[cfb_]->rtx_mac_cts);
						break;
					case PT_MAC_ACK:
						++(tc[cfb_]->rtx_mac_ack);
						break;
				}
			}

			// Stop the send timer
			if(tc[cfb_]->mhSend_.busy())
				tc[cfb_]->mhSend_.stop();
			// Then wait another PIFS-SIFS so that
			// transmission is resumed PIFS after
			// the the end of the last transmission
			// (802.11e - 9.10.1.2)
			cfbFailed_ = 1;
			mhCfb_.start(pifs_-sifs_);
			return;
		}
	}

	// Just waited PIFS after failure of
	// expected response, resume transmission
	if(cfbFailed_) {
		cfbFailed_ = 0;
	}
	
	// If the TXOP has expired without QoS CF-Poll reception,
	// errors on the channel have occurred and the expected
	// QoS CF-Poll wasn't received, so end the Contention Free Burst
	double remaining = txopLimit_ + txopDuration_ - NOW;
	if(remaining < 0) {
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) { 
			fprintf(macLog, "[%d.%d] ...TXOP expired! End TXOP without reception of CF-Poll (rem: %f)\n", index_, cfb_, remaining);
		}
#endif //=========================================================================
		// Update HCF statistics
		tc[cfb_]->cfb_time += (NOW - txopLimit_);
		tc[cfb_]->cfb_num++;
		tc[cfb_]->cfb_num_bad_end++;
		// Abnormal TXOP termination requires
		// some additional controls over timers
		if(tc[cfb_]->mhSend_.busy())
			tc[cfb_]->mhSend_.stop();
		if(mhCfb_.busy())
			mhCfb_.stop();
		if(tc[cfb_]->pktCTRL_) {
			Packet::free(tc[cfb_]->pktCTRL_);
			tc[cfb_]->pktCTRL_ = 0;
		}
		if(tc[cfb_]->pktTx_) {
			HDR_MAC802_11(tc[cfb_]->pktTx_)->dh_fc.fc_subtype = MAC_Subtype_Data;
			HDR_CMN(tc[cfb_]->pktTx_)->mac_tc_aux = (u_int8_t)num_tc;
			// XXX v2.0Start
			HDR_MAC802_11(tc[cfb_]->pktTx_)->dh_duration = usec(txtime(ETHER_ACK_LEN2, dataRate_) + sifs_);
			// XXX v2.0End
		}
		tc[cfb_]->tx_resume();
		cfb_ = -1;
		txopLimit_ = 0;
		txopDuration_ = 0;
		if(!cfp_) {
			for(int i=0; i<num_tc; i++)
				if (tc[i]->edcaTXenabled())
					tc[i]->resumeTC();
			if(mhNav_.busy())
				mhNav_.stop();
			nav_ = NOW;
		}
		return;
	}
	
	// Check if there's enough time left in the
	// Polled TXOP to send another Data packet
	if(tc[cfb_]->pktTx_ && !(tc[cfb_]->pktCTRL_)) {
		// XXX v2.0Start
		double needed = txtime(tc[cfb_]->pktTx_)
		         + DSSS_MaxPropagationDelay
			 + sifs_
		         + txtime(ETHER_QOS_CFPOLL_LEN2, dataRate_);
		// XXX v2.0End
		struct hdr_cmn* ch = HDR_CMN(tc[cfb_]->pktTx_);
		struct hdr_mac802_11* dh = HDR_MAC802_11(tc[cfb_]->pktTx_);
		u_int32_t dst = (u_int32_t)ETHER_ADDR(dh->dh_da);
		
		// Not enough time: reset pktTx_ subtype to
		// MAC_Subtype_Data and send CF-Ack or Null
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) { 
			fprintf(macLog, "[rem: %f, ned: %f] ", remaining, needed);
		}
#endif //=========================================================================
		if(remaining < needed) {
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) { 
				fprintf(macLog, "...not enough time!\n");
			}
#endif //=========================================================================
			if(dh->dh_fc.fc_subtype == MAC_Subtype_QoS_Data ||
			   dh->dh_fc.fc_subtype == MAC_Subtype_Data) {
				tc[cfb_]->cfbSendNULL(dst);
			}
			if(dh->dh_fc.fc_subtype == MAC_Subtype_QoS_Data_CFAck) {
				tc[cfb_]->cfbSendCFACK(dst, ch->mac_tc_aux);
			}
			dh->dh_fc.fc_subtype = MAC_Subtype_Data;
			ch->mac_tc_aux = (u_int8_t)num_tc;
			// XXX v2.0Start
			dh->dh_duration = usec(txtime(ETHER_ACK_LEN2, dataRate_)
			                       + sifs_);
			// XXX v2.0End
		}
	}
	
	// Send QoS CF-Ack or QoS Null
	if(tc[cfb_]->pktCTRL_) {
		struct hdr_mac802_11* dh = HDR_MAC802_11(tc[cfb_]->pktCTRL_);
		switch(dh->dh_fc.fc_subtype) {
		case MAC_Subtype_QoS_CFAck:
		case MAC_Subtype_QoS_Null:
			tc[cfb_]->set_tx_state(MAC_QOS_CFACK);
			timeout = txtime(tc[cfb_]->pktCTRL_);
			wait = txtime(tc[cfb_]->pktCTRL_)
			       + sifs_
			       + DSSS_MaxPropagationDelay;
			txop = cfb_;
			transmit(tc[cfb_]->pktCTRL_, timeout, cfb_);
			mhCfb_.start(wait);
			break;
		default:
			fprintf(stderr, "Invalid CFB pktCTRL_ subtype\n");
			exit(1);
		}
	}

	// Send QoS Data or QoS Data+CF-Ack	
	if(!(tc[cfb_]->pktCTRL_) && tc[cfb_]->pktTx_) {
		struct hdr_mac802_11* dh = HDR_MAC802_11(tc[cfb_]->pktTx_);
		switch(dh->dh_fc.fc_subtype) {
		case MAC_Subtype_QoS_Data:
			tc[cfb_]->set_tx_state(MAC_QOS_DATA);
			break;
		case MAC_Subtype_QoS_Data_CFAck:
			tc[cfb_]->set_tx_state(MAC_QOS_DATA_CFACK);
			break;
		default:
			fprintf(stderr, "Invalid CFB pktTx_ subtype\n");
			exit(1);
		}
		// XXX v2.0Start
		timeout = txtime(tc[cfb_]->pktTx_)
		          + DSSS_MaxPropagationDelay
		          + sifs_
			  + txtime(MAC_MaxMPDUSize, dataRate_)
			  + DSSS_MaxPropagationDelay;
		// XXX v2.0End
		wait = txtime(tc[cfb_]->pktTx_)
		       + sifs_
		       + DSSS_MaxPropagationDelay;
		// The timeout value must not
		// exceed the CFB remaining time
		if(timeout > remaining)
			timeout = remaining;
		txop = cfb_;
		transmit(tc[cfb_]->pktTx_, timeout, cfb_);
		mhCfb_.start(wait);
	}

#ifdef DEBUGACTIVE //=============================================================
	if(NOW > DEBUGTIME) { 
		fprintf(macLog, "%f (%d)\n", NOW, index_);
	}
#endif //=========================================================================

}
// ---------- HCFend

/*###################################################################
#                                                                   #
# capture: called when an overlapped frame with low power has come. #
#          Since the overlapped frame has low power, it is not      #
#          possible to take any decision                            #
#                                                                   #
#        p: the captured frame                                      #
#                                                                   #
###################################################################*/

void Mac802_11::capture(Packet *p) {
	Packet::free(p);
}

/*###################################################################
#                                                                   #
# collision: a collision has occurred on the channel between 'p'    #
#            and 'pktRx_'. This method discards the frame whose     #
#            receiving timer will expire for first                  #
#                                                                   #
#        p: the colliding frame                                     #
#       nt: the Access Category index at which 'p' belongs          #
#                                                                   #
###################################################################*/

void Mac802_11::collision(Packet *p, int nt) {

        hdr_cmn* ch = hdr_cmn::access(p);
        hdr_cmn* ch2 = hdr_cmn::access(pktRx_);

	switch(rx_state_) {
	case MAC_RECV:
                set_rx_state(MAC_COLL);

		/* fall through */

	case MAC_COLL:
		assert(pktRx_);
		assert(mhRecv_.busy());

		/*
		 *  Since a collision has occurred, figure out
		 *  which packet that caused the collision will
		 *  "last" the longest.  Make this packet,
		 *  pktRx_ and reset the Recv Timer if necessary.
		 */

		if(txtime(p) > mhRecv_.expire()) {                              //   pktRx_
                                                                                // ---------|
                        hdr_mac802_11 *mh = HDR_MAC802_11(pktRx_);              //
                        u_int8_t type = mh->dh_fc.fc_type;                      //        p
                        u_int8_t subtype = mh->dh_fc.fc_subtype;                // ---------------|

                        int nt2;
                        if (num_tc == 1) {
                                nt2 = 0;
                        } else {
                                nt2 = HDR_CMN(pktRx_)->mac_tc;    // Extracts the Traffic Category from the common header
                        }

                        switch (ch2->ptype()) {                         // Statistics: colliding frames
                                case PT_TCP:
                                        (tc[nt2]->rcv_chn_col_tcp)++;
                                        break;
                                case PT_ACK:
                                        (tc[nt2]->rcv_chn_col_ack)++;
                                        break;
                                case PT_MESSAGE:
                                        (tc[nt2]->rcv_chn_col_message)++;
                                        break;
                                case PT_ARP:
                                        (tc[nt2]->rcv_chn_col_ARP)++;
                                        break;
                                case PT_VBR:
                                case PT_CBR:
                                        (tc[nt2]->rcv_chn_col_udp)++;
                                        break;
                                case PT_VOIP:
                                        (tc[nt2]->rcv_chn_col_voip)++;
                                        break;	
                                case PT_MAC_RTS:
                                        (tc[nt2]->rcv_chn_col_mac_rts)++;
                                        break;
                                case PT_MAC_CTS:
                                        (tc[nt2]->rcv_chn_col_mac_cts)++;
                                        break;
                                case PT_MAC_ACK:
                                        (tc[nt2]->rcv_chn_col_mac_ack)++;
                                        break;
                        }
			mhRecv_.stop();
			discard(pktRx_, DROP_MAC_COLLISION);    // Discard 'pktRx_'
			pktRx_ = p;
			mhRecv_.start(txtime(pktRx_));          // Now it's necessary to wait for the complete reception of 'p'


                } else {                                                        //      p
                                                                                // ---------|
                                switch (ch->ptype()) {                          //
                                case PT_TCP:                                    //      pktRx_
                                        (tc[nt]->rcv_chn_col_tcp)++;            // ---------------|
                                        break;
                                case PT_ACK:
                                        (tc[nt]->rcv_chn_col_ack)++;
                                        break;                                  // Statistics: colliding frames
                                case PT_MESSAGE:
                                        (tc[nt]->rcv_chn_col_message)++;
                                        break;
                                case PT_ARP:
                                        (tc[nt]->rcv_chn_col_ARP)++;
                                        break;
                                case PT_VBR:
                                case PT_CBR:
                                        (tc[nt]->rcv_chn_col_udp)++;
                                        break;
                                case PT_VOIP:
                                        (tc[nt]->rcv_chn_col_voip)++;
                                        break;	
                                case PT_MAC_RTS:
                                        (tc[nt]->rcv_chn_col_mac_rts)++;
                                        break;
                                case PT_MAC_CTS:
                                        (tc[nt]->rcv_chn_col_mac_cts)++;
                                        break;
                                case PT_MAC_ACK:
                                        (tc[nt]->rcv_chn_col_mac_ack)++;
                                        break;
                        }
			discard(p, DROP_MAC_COLLISION);         // Discard 'p'
		}
		break;
	default:
		assert(0);
	}
}

/*###################################################################
#                                                                   #
# recv_timer: this function is invoked when the last bit of a frame #
#             is received from the channel. This frame is processed #
#             only if it is error-free and if there is no collision #
#                                                                   #
###################################################################*/

void Mac802_11::recv_timer() {
	u_int32_t src;
	hdr_cmn *ch = HDR_CMN(pktRx_);
	hdr_mac802_11 *mh = HDR_MAC802_11(pktRx_);
	u_int32_t dst = ETHER_ADDR(mh->dh_da);

	u_int8_t  type = mh->dh_fc.fc_type;
	u_int8_t  subtype = mh->dh_fc.fc_subtype;

	assert(pktRx_);
	assert(rx_state_ == MAC_RECV || rx_state_ == MAC_COLL);

        int nt;
        if (num_tc == 1) {
                nt = 0;
        } else {
                nt = nt = HDR_CMN(pktRx_)->mac_tc;          // Extracts the identifier of the queue at which this frame belongs
        }

        /*
         *  If the interface is in TRANSMIT mode when this packet
         *  "arrives", then I would never have seen it and should
         *  do a silent discard without adjusting the NAV.
         */

	if(tx_active_ != -1) {
                switch (ch->ptype()) { // Statistics: these frames are considered as destined to others nodes
                        case PT_TCP:
                                (tc[nt]->rcv_chn_other_tcp)++;
                                break;
                        case PT_ACK:
                                (tc[nt]->rcv_chn_other_ack)++;
                                break;
                        case PT_MESSAGE:
                                (tc[nt]->rcv_chn_other_message)++;
                                break;
                        case PT_ARP:
                                (tc[nt]->rcv_chn_other_ARP)++;
                                break;
                        case PT_VBR:
                        case PT_CBR:
                                (tc[nt]->rcv_chn_other_udp)++;
                                break;
                        case PT_VOIP:
                                (tc[nt]->rcv_chn_other_voip)++;
                                break;	
                        case PT_MAC_RTS:
                                (tc[nt]->rcv_chn_other_mac_rts)++;
                                break;
                        case PT_MAC_CTS:
                                (tc[nt]->rcv_chn_other_mac_cts)++;
                                break;
                        case PT_MAC_ACK:
                                (tc[nt]->rcv_chn_other_mac_ack)++;
                                break;
                }
		// ---------- HCFstart
		// If the response packet is discarded
		// during a CFB while waiting for a response,
		// recovery by retransmitting after SIFS
		// (802.11e - 9.10.1.2)
		if((cfb_ != -1) &&
		   !mhCfb_.busy()) {
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) { 
				fprintf(macLog, "[%d] incoming packet discarded (transmitting now), retransmit ", index_);
			}
#endif //=========================================================================
			// Stop the send timer
			if(tc[cfb_]->mhSend_.busy())
				tc[cfb_]->mhSend_.stop();
			// If Data was sent, set the retry bit
			// and update the statistics
			if(tc[cfb_]->tx_state_ == MAC_QOS_DATA ||
			   tc[cfb_]->tx_state_ == MAC_QOS_DATA_CFACK) {
				HDR_MAC802_11(tc[cfb_]->pktTx_)->dh_fc.fc_retry = 1;

				switch (HDR_CMN(tc[cfb_]->pktTx_)->ptype()) {     // Statistics: retransmitted frames
					case PT_TCP:
						++(tc[cfb_]->rtx_tcp);
						break;
					case PT_ACK:
						++(tc[cfb_]->rtx_ack);
						break;
					case PT_MESSAGE:
						++(tc[cfb_]->rtx_message);
						break;
					case PT_ARP:
						++(tc[cfb_]->rtx_ARP);
						break;
               case PT_VBR:
					case PT_CBR:
						++(tc[cfb_]->rtx_udp);
						break;
					case PT_VOIP:
						++(tc[cfb_]->rtx_voip);
						break;	
					case PT_MAC_RTS:
						++(tc[cfb_]->rtx_mac_rts);
						break;
					case PT_MAC_CTS:
						++(tc[cfb_]->rtx_mac_cts);
						break;
					case PT_MAC_ACK:
						++(tc[cfb_]->rtx_mac_ack);
						break;
				}
			}
			tc[cfb_]->set_tx_state(MAC_IDLE);
			mhCfb_.start(sifs_);
		}
		// ---------- HCFend
      Packet::free(pktRx_);
      goto done;
   }

	/*
	 * Handle collisions.
	 */

	if(rx_state_ == MAC_COLL) {
                switch (ch->ptype()) {          // Statistics: colliding frames
                        case PT_TCP:
                                (tc[nt]->rcv_chn_col_tcp)++;
                                break;
                        case PT_ACK:
                                (tc[nt]->rcv_chn_col_ack)++;
                                break;
                        case PT_MESSAGE:
                                (tc[nt]->rcv_chn_col_message)++;
                                break;
                        case PT_ARP:
                                (tc[nt]->rcv_chn_col_ARP)++;
                                break;
                        case PT_VBR:
                        case PT_CBR:
                                (tc[nt]->rcv_chn_col_udp)++;
                                break;
                        case PT_VOIP:
                                (tc[nt]->rcv_chn_col_voip)++;
                                break;	
                        case PT_MAC_RTS:
                                (tc[nt]->rcv_chn_col_mac_rts)++;
                                break;
                        case PT_MAC_CTS:
                                (tc[nt]->rcv_chn_col_mac_cts)++;
                                break;
                        case PT_MAC_ACK:
                                (tc[nt]->rcv_chn_col_mac_ack)++;
                                break;
                }
		// ---------- HCFstart
		// If the response packet is discarded
		// during a CFB while waiting for a response,
		// recovery by retransmitting after SIFS
		// (802.11e - 9.10.1.2)
		if((cfb_ != -1) &&
		   !mhCfb_.busy()) {
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) { 
				fprintf(macLog, "[%d] incoming packet discarded (reception collision), retransmit ", index_);
			}
#endif //=========================================================================
			// Stop the send timer
			if(tc[cfb_]->mhSend_.busy())
				tc[cfb_]->mhSend_.stop();
			// If Data was sent, set the retry bit
			// and update the statistics
			if(tc[cfb_]->tx_state_ == MAC_QOS_DATA ||
			   tc[cfb_]->tx_state_ == MAC_QOS_DATA_CFACK) {
				HDR_MAC802_11(tc[cfb_]->pktTx_)->dh_fc.fc_retry = 1;

				switch (HDR_CMN(tc[cfb_]->pktTx_)->ptype()) {     // Statistics: retransmitted frames
					case PT_TCP:
						++(tc[cfb_]->rtx_tcp);
						break;
					case PT_ACK:
						++(tc[cfb_]->rtx_ack);
						break;
					case PT_MESSAGE:
						++(tc[cfb_]->rtx_message);
						break;
					case PT_ARP:
						++(tc[cfb_]->rtx_ARP);
						break;
               case PT_VBR:
					case PT_CBR:
						++(tc[cfb_]->rtx_udp);
						break;
					case PT_VOIP:
						++(tc[cfb_]->rtx_voip);
						break;	
					case PT_MAC_RTS:
						++(tc[cfb_]->rtx_mac_rts);
						break;
					case PT_MAC_CTS:
						++(tc[cfb_]->rtx_mac_cts);
						break;
					case PT_MAC_ACK:
						++(tc[cfb_]->rtx_mac_ack);
						break;
				}
			}
			tc[cfb_]->set_tx_state(MAC_IDLE);
			mhCfb_.start(sifs_);
		}
		// ---------- HCFend
		discard(pktRx_, DROP_MAC_COLLISION);
		err_frame = 1;                          // The last received frame is not error-free
		goto done;
	}

	/*
	 * Check to see if this packet was received with enough
	 * bit errors that the current level of FEC still could not
	 * fix all of the problems - ie; after FEC, the checksum still
	 * failed.
	 */

	if( ch->error() ) {
                switch (ch->ptype()) {                          // Statistics: corrupted packets
                        case PT_TCP:
                                (tc[nt]->rcv_chn_err_tcp)++;
                                break;
                        case PT_ACK:
                                (tc[nt]->rcv_chn_err_ack)++;
                                break;
                        case PT_MESSAGE:
                                (tc[nt]->rcv_chn_err_message)++;
                                break;
                        case PT_ARP:
                                (tc[nt]->rcv_chn_err_ARP)++;
                                break;
                        case PT_VBR:
                        case PT_CBR:
                                (tc[nt]->rcv_chn_err_udp)++;
                                break;
                        case PT_VOIP:
                                (tc[nt]->rcv_chn_err_voip)++;
                                break;	
                        case PT_MAC_RTS:
                                (tc[nt]->rcv_chn_err_mac_rts)++;
                                break;
                        case PT_MAC_CTS:
                                (tc[nt]->rcv_chn_err_mac_cts)++;
                                break;
                        case PT_MAC_ACK:
                                (tc[nt]->rcv_chn_err_mac_ack)++;
                                break;
                }
		// ---------- HCFstart
		// If the response packet is discarded
		// during a CFB while waiting for a response,
		// recovery by retransmitting after SIFS
		// (802.11e - 9.10.1.2)
		if((cfb_ != -1) &&
		   !mhCfb_.busy()) {
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) {
				fprintf(macLog, "[%d] incoming packet discarded (error), retransmit ", index_);
			}
#endif //=========================================================================
			// Stop the send timer
			if(tc[cfb_]->mhSend_.busy())
				tc[cfb_]->mhSend_.stop();
			// If Data was sent, set the retry bit
			// and update the statistics
			if(tc[cfb_]->tx_state_ == MAC_QOS_DATA ||
			   tc[cfb_]->tx_state_ == MAC_QOS_DATA_CFACK) {
				HDR_MAC802_11(tc[cfb_]->pktTx_)->dh_fc.fc_retry = 1;

				switch (HDR_CMN(tc[cfb_]->pktTx_)->ptype()) {     // Statistics: retransmitted frames
					case PT_TCP:
						++(tc[cfb_]->rtx_tcp);
						break;
					case PT_ACK:
						++(tc[cfb_]->rtx_ack);
						break;
					case PT_MESSAGE:
						++(tc[cfb_]->rtx_message);
						break;
					case PT_ARP:
						++(tc[cfb_]->rtx_ARP);
						break;
               case PT_VBR:
					case PT_CBR:
						++(tc[cfb_]->rtx_udp);
						break;
					case PT_VOIP:
						++(tc[cfb_]->rtx_voip);
						break;	
					case PT_MAC_RTS:
						++(tc[cfb_]->rtx_mac_rts);
						break;
					case PT_MAC_CTS:
						++(tc[cfb_]->rtx_mac_cts);
						break;
					case PT_MAC_ACK:
						++(tc[cfb_]->rtx_mac_ack);
						break;
				}
			}
			tc[cfb_]->set_tx_state(MAC_IDLE);
			mhCfb_.start(sifs_);
		}
		// ---------- HCFend
		Packet::free(pktRx_);
		err_frame = 1;                  // The last received frame is not error-free
		goto done;
	}

        err_frame = 0;          // The last received frame is error-free

	/*
	 * IEEE 802.11 specs, section 9.2.5.6
	 *	- update the NAV (Network Allocation Vector)
	 */

	if(dst != (u_int32_t)index_) {          // If this frame is not destined to this node, the NAV must be updated
		set_nav(mh->dh_duration);
	}

   /* tap out - */
   if (tap_ && type == MAC_Type_Data && MAC_Subtype_Data == subtype) {
      tap_->tap(pktRx_);
   }

	/*
	 * Adaptive Fidelity Algorithm Support - neighborhood infomation
	 * collection
	 *
	 * Hacking: Before filter the packet, log the neighbor node
	 * I can hear the packet, the src is my neighbor
	 */

	if (netif_->node()->energy_model() &&
	    netif_->node()->energy_model()->adaptivefidelity()) {
		src = ETHER_ADDR(mh->dh_sa);
		netif_->node()->energy_model()->add_neighbor(src);
	}

	/*
	 * Address Filtering
	 */

	// ---------- HCFstart
	// If receiving a QoS CF-Poll that ends
	// a Polled TXOP, reset the CFB flag,
	// free the pktTx_ and stop the send timer.
	// If not in CFP, reset the NAV, too.
	// If this is a legacy station it can't
	// understand QoS CF-Poll frames, so
	// this control is skipped
	if(!legacy_ &&
	   dst == HCmacAddress() &&
	   type == MAC_Type_Data &&
	   subtype == MAC_Subtype_QoS_CFPoll &&
	   mh->dh_duration == 0) {
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) { 
			fprintf(macLog, "[%d] CF-Poll received, TXOP end ", index_);
		}
#endif //=========================================================================
		if(cfb_ != -1) {
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) { 
				fprintf(macLog, "- reset CFB flag ");
			}
#endif //=========================================================================
			// Update HCF statistics
			tc[cfb_]->cfb_time += (NOW - txopLimit_);
			tc[cfb_]->cfb_num++;
			// Possibly free a
			// CF-Ack or Null frame
			if(tc[cfb_]->pktCTRL_) {
				Packet::free(tc[cfb_]->pktCTRL_);
				tc[cfb_]->pktCTRL_ = 0;
			}
			// The QoS CF-Poll that ends the TXOP
			// is considered as an ACK by the QSTA,
			// so free the transmitted Data packet
			if(tc[cfb_]->mhSend_.busy()) {
				Packet::free(tc[cfb_]->pktTx_);
				tc[cfb_]->pktTx_ = 0;
            tc[cfb_]->mhSend_.stop();
            //fprintf(macLog, "\n%f --- Receiving CF-POLL as ACK. (Queue[%d]: %d)", NOW, cfb_, queueLength(cfb_));
				// Still in CFB, this tx_resume() invocation
				// just asks for another packet and resets
				// the tx_state_ to MAC_IDLE
				tc[cfb_]->tx_resume();
            tc[cfb_]->tx_state_ = MAC_IDLE;
            //fprintf(macLog, "\n%f --- Receiving CF-POLL as ACK. (Queue[%d]: %d)", NOW, cfb_, queueLength(cfb_));
			}
			cfb_ = -1;
			txopLimit_ = 0;
			txopDuration_ = 0;
			if(!cfp_) {
				for(int i=0; i<num_tc; i++)
					if (tc[i]->edcaTXenabled())
						tc[i]->resumeTC();
			}
		}
		if(!cfp_) {
			if(mhNav_.busy()) {
#ifdef DEBUGACTIVE //=============================================================
				if(NOW > DEBUGTIME) { 
					fprintf(macLog, "- NAV was %f ", nav_);
				}
#endif //=========================================================================
				mhNav_.stop();
			}
			nav_ = NOW;
		}
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) { 
			fprintf(macLog, "- NAV set to %f\n", nav_);
		}
#endif //=========================================================================
		// Free the received frame
		Packet::free(pktRx_);
		goto done;
	}
	// ---------- HCFend
	
	if(dst != (u_int32_t)index_ && dst != MAC_BROADCAST) {          // The frame belongs to another node, so we can discard it

		/*
		 *  We don't want to log this event, so we just free
		 *  the packet instead of calling the drop routine.
		 */

                switch (ch->ptype()) {          // Statistics: frames destined to others nodes
                        case PT_TCP:
                                (tc[nt]->rcv_chn_other_tcp)++;
                                break;
                        case PT_ACK:
                                (tc[nt]->rcv_chn_other_ack)++;
                                break;
                        case PT_MESSAGE:
                                (tc[nt]->rcv_chn_other_message)++;
                                break;
                        case PT_ARP:
                                (tc[nt]->rcv_chn_other_ARP)++;
                                break;
                        case PT_VBR:
                        case PT_CBR:
                                (tc[nt]->rcv_chn_other_udp)++;
                                break;
                        case PT_VOIP:
                                (tc[nt]->rcv_chn_other_voip)++;
                                break;	
                        case PT_MAC_RTS:
                                (tc[nt]->rcv_chn_other_mac_rts)++;
                                break;
                        case PT_MAC_CTS:
                                (tc[nt]->rcv_chn_other_mac_cts)++;
                                break;
                        case PT_MAC_ACK:
                                (tc[nt]->rcv_chn_other_mac_ack)++;
                                break;
                }

		// ---------- HCFstart
		// If a frame directed to another station is
		// received during the CFB, while waiting for
		// the QoS CF-Poll that ends the Polled TXOP,
		// it means that the QoS CF-Poll we are still
		// waiting for has been missed, so end the
		// Contention Free Burst now
		if(cfb_ != -1 &&
		   !mhCfb_.busy() &&
		   tc[cfb_]->tx_state_ == MAC_QOS_CFACK) {
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) { 
				fprintf(macLog, "[%d.%d] ...QoS Cf-Poll missed! End TXOP without reception of CF-Poll (pktRx_->dst: %d)\n", index_, cfb_, dst);
			}
#endif //=========================================================================
			// Update HCF statistics
			tc[cfb_]->cfb_time += (NOW - txopLimit_);
			tc[cfb_]->cfb_num++;
			tc[cfb_]->cfb_num_bad_end++;
			// Free the CF-Ack or Null frame
			if(tc[cfb_]->pktCTRL_) {
				Packet::free(tc[cfb_]->pktCTRL_);
				tc[cfb_]->pktCTRL_ = 0;
			}
			// Possibly restore Data packet
			// Contention Period original parameters
			if(tc[cfb_]->pktTx_) {
				HDR_MAC802_11(tc[cfb_]->pktTx_)->dh_fc.fc_subtype = MAC_Subtype_Data;
				HDR_CMN(tc[cfb_]->pktTx_)->mac_tc_aux = (u_int8_t)num_tc;
				// XXX v2.0Start
				HDR_MAC802_11(tc[cfb_]->pktTx_)->dh_duration = usec(txtime(ETHER_ACK_LEN2, dataRate_) + sifs_);
				// XXX v2.0End
			}
			// Still in CFB, this tx_resume() invocation
			// just asks for another packet and resets
			// the tx_state_ to MAC_IDLE
			tc[cfb_]->tx_resume();
			// Reset CFB configuration
			cfb_ = -1;
			txopLimit_ = 0;
			txopDuration_ = 0;
			if(!cfp_) {
				for(int i=0; i<num_tc; i++)
					if (tc[i]->edcaTXenabled())
						tc[i]->resumeTC();
				if(mhNav_.busy())
					mhNav_.stop();
				nav_ = NOW;
			}
		}
		// ---------- HCF end
		discard(pktRx_, "---");
		goto done;
	}

	switch(type) {
	case MAC_Type_Management:
		// ---------- HCFstart
		switch(subtype) {
		case MAC_Subtype_Beacon:
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) { 
				fprintf(macLog, "[%d] Beacon received %f\n", index_, NOW);
			}
#endif //=========================================================================
			recvBCN(pktRx_);
			break;
		// ---------- HCFend
		default:
			fprintf(stderr, "recv_timer0:Invalid MAC Management Subtype %x\n", subtype);
			exit(1);
		}
		break;

	case MAC_Type_Control:
		switch(subtype) {
		case MAC_Subtype_RTS:
         (tc[nt]->rcv_chn_my_mac_rts)++;        // Receiving a RTS
			tc[nt]->recvRTS(pktRx_);
			break;
		case MAC_Subtype_CTS:
                        (tc[nt]->rcv_chn_my_mac_cts)++;        // Receiving a CTS
			tc[nt]->recvCTS(pktRx_);
			break;
      case MAC_Subtype_ACK:
         (tc[nt]->rcv_chn_my_mac_ack)++;        // Receiving an ACK
         tc[nt]->recvACK(pktRx_);
         break;
      // ---------- HCFstart
		case MAC_Subtype_CC:
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) { 
				fprintf(macLog, "[%d] Contention Control received %f\n", index_, NOW);
			}
#endif //=========================================================================
			recvCC(pktRx_);
			break;
		case MAC_Subtype_CFEnd:
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) { 
				fprintf(macLog, "[%d] CF-End received %f\n", index_, NOW);
			}
#endif //=========================================================================
			recvCFEND(pktRx_);
			break;
		// ---------- HCFend
		default:
			fprintf(stderr,"recvTimer1:Invalid MAC Control Subtype %x\n",
				subtype);
			exit(1);
		}
		break;

	case MAC_Type_Data:
		switch(subtype) {                       // Receiving a DATA frame
		case MAC_Subtype_Data:
                       switch (ch->ptype()) {          // Statistics: frames correctly received
                                case PT_TCP:
                                        (tc[nt]->rcv_chn_my_tcp)++;
                                        ch->tmac3 = NOW;
                                        break;
                                case PT_ACK:
                                        (tc[nt]->rcv_chn_my_ack)++;
                                        break;
                                case PT_MESSAGE:
                                        (tc[nt]->rcv_chn_my_message)++;
                                        break;
                                case PT_ARP:
                                        (tc[nt]->rcv_chn_my_ARP)++;
                                        break;
                                case PT_VBR:
                                case PT_CBR:
                                        (tc[nt]->rcv_chn_my_udp)++;
                                        ch->tmac3 = NOW;
                                        break;
                                case PT_VOIP:
                                        (tc[nt]->rcv_chn_my_voip)++;
                                        ch->tmac3 = NOW;
                                        break;	
                        }
			tc[nt]->recvDATA(pktRx_);
			break;
		// ---------- HCFstart
		case MAC_Subtype_QoS_Data_CFPoll:
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) { 
				fprintf(macLog, "[%d.%d] Data+CF-Poll received %f\n", index_, nt, NOW);
			}
#endif //=========================================================================
			switch (ch->ptype()) {          // Statistics: frames correctly received
                                case PT_TCP:
                                        (tc[nt]->rcv_chn_my_tcp)++;
                                        ch->tmac3 = NOW;
                                        break;
                                case PT_ACK:
                                        (tc[nt]->rcv_chn_my_ack)++;
                                        break;
                                case PT_MESSAGE:
                                        (tc[nt]->rcv_chn_my_message)++;
                                        break;
                                case PT_ARP:
                                        (tc[nt]->rcv_chn_my_ARP)++;
                                        break;
                                case PT_VBR:
                                case PT_CBR:
                                        (tc[nt]->rcv_chn_my_udp)++;
                                        ch->tmac3 = NOW;
                                        break;
                                case PT_VOIP:
                                        (tc[nt]->rcv_chn_my_voip)++;
                                        ch->tmac3 = NOW;
                                        break;	
                        }
			tc[nt]->recvDATA_CFPOLL(pktRx_);
			break; 
		case MAC_Subtype_QoS_CFPoll:
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) { 
				fprintf(macLog, "[%d.%d] CF-Poll received %f\n", index_, nt, NOW);
			}
#endif //=========================================================================
			tc[nt]->recvCFPOLL(pktRx_);
			break;
		case MAC_Subtype_QoS_CFAck:
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) { 
				fprintf(macLog, "[%d.%d] CF-Ack received %f\n", index_, nt, NOW);
			}
#endif //=========================================================================
			tc[nt]->recvCFACK(pktRx_);
			break;
		case MAC_Subtype_QoS_Data_CFAck:
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) { 
				fprintf(macLog, "[%d.%d] Data+CF-Ack received %f\n", index_, nt, NOW);
			}
#endif //=========================================================================
			switch (ch->ptype()) {          // Statistics: frames correctly received
                                case PT_TCP:
                                        (tc[nt]->rcv_chn_my_tcp)++;
                                        ch->tmac3 = NOW;
                                        break;
                                case PT_ACK:
                                        (tc[nt]->rcv_chn_my_ack)++;
                                        break;
                                case PT_MESSAGE:
                                        (tc[nt]->rcv_chn_my_message)++;
                                        break;
                                case PT_ARP:
                                        (tc[nt]->rcv_chn_my_ARP)++;
                                        break;
                                case PT_VBR:
                                case PT_CBR:
                                        (tc[nt]->rcv_chn_my_udp)++;
                                        ch->tmac3 = NOW;
                                        break;
                                case PT_VOIP:
                                        (tc[nt]->rcv_chn_my_voip)++;
                                        ch->tmac3 = NOW;
                                        break;	
                        }
			tc[nt]->recvDATA_CFACK(pktRx_);
			break;
		case MAC_Subtype_QoS_Data_CFAck_CFPoll:
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) { 
				fprintf(macLog, "[%d.%d] Data+CF-Ack+CF-Poll received %f\n", index_, nt, NOW);
			}
#endif //=========================================================================
			switch (ch->ptype()) {          // Statistics: frames correctly received
                                case PT_TCP:
                                        (tc[nt]->rcv_chn_my_tcp)++;
                                        ch->tmac3 = NOW;
                                        break;
                                case PT_ACK:
                                        (tc[nt]->rcv_chn_my_ack)++;
                                        break;
                                case PT_MESSAGE:
                                        (tc[nt]->rcv_chn_my_message)++;
                                        break;
                                case PT_ARP:
                                        (tc[nt]->rcv_chn_my_ARP)++;
                                        break;
                                case PT_VBR:
                                case PT_CBR:
                                        (tc[nt]->rcv_chn_my_udp)++;
                                        ch->tmac3 = NOW;
                                        break;
                                case PT_VOIP:
                                        (tc[nt]->rcv_chn_my_voip)++;
                                        ch->tmac3 = NOW;
                                        break;	
                        }
			tc[nt]->recvDATA_CFACK_CFPOLL(pktRx_);
			break;
		// ---------- HCFend
		default:
			fprintf(stderr, "recv_timer2:Invalid MAC Data Subtype %x\n",
				subtype);
			exit(1);
		}
		break;
	default:
		fprintf(stderr, "recv_timer3:Invalid MAC Type %x\n", subtype);
		exit(1);
	}
done:
	pktRx_ = 0;
	rx_resume();    // The frame reception is ended
//   if (type == MAC_Type_Control && subtype == MAC_Subtype_ACK)
//      fprintf(macLog, "ACK received  ");
//   fprintf(macLog, "NOW: %f   QUEUE: %d\n", NOW, queueLength(2));
}

/*###################################################################
#                                                                   #
# rx_resume: called at the end of a frame reception, this method    #
#            switches the receiving state to MAC_IDLE               #
#                                                                   #
###################################################################*/

void Mac802_11::rx_resume() {
	assert(pktRx_ == 0);
	assert(mhRecv_.busy() == 0);
	set_rx_state(MAC_IDLE);
        detect_coll = 0;
}

/*###################################################################
#                                                                   #
# upsend: inspects the frames destinated to the LLC layer in order  #
#         to remove possible duplicates. If all goes right, the     #
#         packet is sent to the up-layer                            #
#                                                                   #
#         p: the new packet                                         #
#         n: the Access Category at which this frame belongs        #
#                                                                   #
###################################################################*/

void Mac802_11::upsend(Packet* p, int n) {

        struct hdr_mac802_11 *dh = HDR_MAC802_11(p);
	u_int32_t dst, src, size;
        dst = ETHER_ADDR(dh->dh_da);
	src = ETHER_ADDR(dh->dh_sa);

        /* ============================================================
	    Make/update an entry in our sequence number cache.
	   ============================================================ */

	/* Changed by Debojyoti Dutta. This upper loop of if{}else was
	   suggested by Joerg Diederich <dieder@ibr.cs.tu-bs.de>.
	   Changed on 19th Oct'2000 */

        if(dst != MAC_BROADCAST) {
                if (src < (u_int32_t) cache_node_count_) {
                        Host *h = &cache_[src];

                        if(h->seqno && h->seqno == dh->dh_scontrol) {   // A duplicated frame is detected when its sequence number
                                                                        // doesn't advance the previous (the MAC protocol has a
                                                                        // stop & wait behavior)


                                switch (HDR_CMN(p)->ptype()) {          // Statistics: duplicated frames
                                        case PT_TCP:
                                                (tc[n]->dupl_tcp)++;
                                                break;
                                        case PT_ACK:
                                                (tc[n]->dupl_ack)++;
                                                break;
                                        case PT_MESSAGE:
                                                (tc[n]->dupl_message)++;
                                                break;
                                        case PT_ARP:
                                                (tc[n]->dupl_ARP)++;
                                                break;
                                        case PT_VBR:
                                        case PT_CBR:
                                                (tc[n]->dupl_udp)++;
                                                break;
                                        case PT_VOIP:
                                                (tc[n]->dupl_voip)++;
                                                break;	
                                        case PT_MAC_RTS:
                                                (tc[n]->dupl_mac_rts)++;
                                                break;
                                        case PT_MAC_CTS:
                                                (tc[n]->dupl_mac_cts)++;
                                                break;
                                        case PT_MAC_ACK:
                                                (tc[n]->dupl_mac_ack)++;
                                                break;
                                }
                                discard(p, DROP_MAC_DUPLICATE);
                                return;
                        }
                        h->seqno = dh->dh_scontrol;     // Update the last sequence number that has been seen
                } else {
			static int count = 0;
			if (++count <= 10) {
				printf ("MAC_802_11: accessing MAC cache_ array out of range (src %u, dst %u, size %d)!\n", src, dst, cache_node_count_);
				if (count == 10)
					printf ("[suppressing additional MAC cache_ warnings]\n");
			};
		};
	}

	/*

	 *  Pass the packet up to the link-layer.
	 *  We could schedule an event to account
	 *  for this processing delay.
	 */

        switch (HDR_CMN(p)->ptype()) {          // Statistics: frames sent to the LLC layer
	        case PT_TCP:
                        (tc[n]->send_ll_tcp)++;
                        break;
	        case PT_ACK:
                        (tc[n]->send_ll_ack)++;
                        break;
	        case PT_MESSAGE:
                        (tc[n]->send_ll_message)++;
                        break;
                case PT_ARP:
                        (tc[n]->send_ll_ARP)++;
                        break;
                case PT_VBR:
                case PT_CBR:
                        (tc[n]->send_ll_udp)++;
                        break;
                case PT_VOIP:
                        (tc[n]->send_ll_voip)++;
                        break;	
                case PT_MAC_RTS:
                        (tc[n]->send_ll_mac_rts)++;
                        break;
                case PT_MAC_CTS:
                        (tc[n]->send_ll_mac_cts)++;
                        break;
                case PT_MAC_ACK:
                        (tc[n]->send_ll_mac_ack)++;
                        break;
	}
	uptarget_->recv(p, (Handler*) 0);       // Pass the packet to LLC
}

/*###################################################################
#                                                                   #
# need_pkt: called by the corresponding virtual queue when it has   #
#           the possibility of transmitting a new frame             #
#                                                                   #
#         n: the index of the Access Category that needs a frame    #
#                                                                   #
###################################################################*/

void Mac802_11::need_pkt(int n) {
        if(callback_) {
		Handler *h = callback_;         // We call the BS_QueueHandler in order to inform the BS_DropTail object
		h->handle((Event*)n);           // that the 'n' virtual queue needs a new packet
	}
}

// ---------- HCFstart
/*###################################################################
#                                                                   #
# queueLength: called by the Mac802_11e virtual queues in order     #
#              to know the number of packets waiting to be sent.    #
#              This function actually only forwards the request     #
#              to the IF queue handler (pointed by callback_)       #
#                                                                   #
#              ac: the AC requesting the information                #
#                                                                   #
###################################################################*/

int Mac802_11::queueLength(int ac) {
	return(((BS_QueueHandler*)callback_)->queueLength(ac));
}

/*###################################################################
#                                                                   #
# queueLength: used to communicate the HC how much data has not     #
#              been sent yet.                                       #
#                                                                   #
#              ac: the AC requesting the information                #
#                                                                   #
###################################################################*/

int Mac802_11::queueByteLength(int ac) {
   return(((BS_QueueHandler*)callback_)->queueByteLength(ac));
}


/*###################################################################
#                                                                   #
# HCmacAddress: called by recv_timer, it returns the                #
#               Hybrid Coordinator's MAC address, that the          #
#               station need to know when a QoS CF-Poll that        #
#               ends a Polled TXOP is received                      #
#                                                                   #
###################################################################*/

u_int32_t Mac802_11::HCmacAddress() {
	Tcl& tcl = Tcl::instance();
	tcl.evalc("$obj(mac,0) mac-addr");
	char* addr = tcl.result();
	return((u_int32_t)atoi(addr));
}

/*###################################################################
#                                                                   #
# dataToSend: called during the CFB, this method looks for a TC     #
#             with data to send to the destination QSTA. Used by    #
#             the HC to send Data+CF-Poll, Data+CF-Ack and          #
#             Data+CF-Ack+CF-Poll frames during CFBs. Returns       #
#             the TC if present, num_tc otherwise.                  #
#                                                                   #
#             dst: the MAC address of the destination QSTA          #
#             ac: the TC currently owning the TXOP                  #
#                                                                   #
###################################################################*/

int Mac802_11::dataToSend(int dst, int ac) {
	for(int i=(num_tc-1); i>=0; i--) {
		if(tc[i]->pktTx_) {
			hdr_mac802_11* dh = HDR_MAC802_11(tc[i]->pktTx_);
		   	if ((u_int32_t)ETHER_ADDR(dh->dh_da) == dst) {
				if(i != ac) {
					if(tc[ac]->pktTx_)
						tc[ac]->pktBackup_ = tc[ac]->pktTx_;
					tc[ac]->pktTx_ = tc[i]->pktTx_;
					hdr_cmn *ch = HDR_CMN(tc[i]->pktTx_);
					ch->mac_tc = (u_int8_t)ac;
				}
				return i;
			}
		}
	}
	return num_tc;
}

/*###################################################################
#                                                                   #
# recvBCN: called every time a Beacon frame is received, this       #
#          routine sets up the station for the CFP, setting the     #
#          NAV to the end of the CFP and silencing the TCs          #
#                                                                   #
#          p: the Beacon frame                                      #
#                                                                   #
###################################################################*/

void Mac802_11::recvBCN(Packet *p) {

	struct beacon_frame *bf;
	bf = (struct beacon_frame*)p->access(hdr_mac::offset_);

	// Receiving a beacon, set the NAV according to it
	set_nav(((double)bf->bf_cf_pset.CFPDurRemaining) * 10);
	cfp_ = 1;

	// Prepare the TCs for the CFP
	for(int i=0; i<num_tc; i++) {
		tc[i]->stopTC();
		tc[i]->tx_resume();
	}

	// Free the received frame
	mac_log(p);

}

/*###################################################################
#                                                                   #
# recvCC: called every time a Contention Control frame is received, #
#         this routine sets up the station for the CCI, setting the #
#         NAV to the end of the CCI and silencing the TCs.          #
#         Then it prompts the TCs for Reservation Requests to send  #
#         during this Controlled Contention Interval. If no RR are  #
#         sent, the CAP timer is started so that the station        #
#         remains quiet until the end of the CCI                    #
#                                                                   #
#          p: the Contention Control frame                          #
#                                                                   #
###################################################################*/

void Mac802_11::recvCC(Packet *p) {

	struct hdr_mac802_11 *dh = HDR_MAC802_11(p);

	// Legacy stations don't
	// understand CC frames
	if(legacy_) {
		discard(p, "---");
		return;
	}

	// The CAP flag is set to a unused TC number
	// so that we know that we are in the CCI, the
	// txopLimit is set to the end of the CCI
	cap_ = num_tc;
	txopLimit_ = NOW + (dh->dh_duration * 1e-6);

	// The stopTC() method is used to stop all the TCs
	for(int i=0; i<num_tc; i++)
		tc[i]->stopTC();

	// Look for a TC with traffic to send
	for(int i=(num_tc-1); (i >= 0) && (cap_ == num_tc); i--) {
		tc[i]->recvCC(p);
	}

	// If no Reservation Request was sent, the QSTA
	// must stay silent until the end of the CCI
	// (802.11e - 9.10.4.2)
	if(cap_ == num_tc) {
		mhCap_.start(dh->dh_duration * 1e-6);
	}

	// Free the received frame
	mac_log(p);

}

/*###################################################################
#                                                                   #
# recvCFEND: called every time a CF-End frame is received, this     #
#            routine restores the station internal state for the    #
#            Contention Period, resetting the NAV and restarting    #
#            the TCs' EDCF beahviour                                #
#                                                                   #
#            p: the CF-End frame                                    #
#                                                                   #
###################################################################*/

void Mac802_11::recvCFEND(Packet *p) {

	// Reset the NAV
	if(mhNav_.busy()) {
		mhNav_.stop();
	}
	nav_ = NOW;

	// Back to Contention Period 
	cfp_ = 0;
	txop  = -1;
	for(int i=0; i<num_tc; i++) {
		if (tc[i]->edcaTXenabled())
			tc[i]->resumeTC();
	}

	// Free the received frame
	mac_log(p);
}
// ---------- HCFend


Mac802_11e::Mac802_11e(Mac802_11* upmac, PHY_MIB *p, MAC_MIB *m) : Connector(), pktTx_(0), mhSend_(this), mhintNav_(this),
                                                                   mhDefer_(this, p->SlotTime), mhBackoff_(this, p->SlotTime) {
	macmib_ = m;
	phymib_ = p;

        mac_target_ = upmac;

	tx_state_ = MAC_IDLE;

	pktRTS_ = 0;
	pktCTRL_ = 0;
	// ---------- HCFstart
	pktPOLL_ = 0;
	pktBackup_ = 0;
	// ---------- HCFend

	ssrc_ = slrc_ = 0;

        virtual_coll = 0;
        int_nav = 0.0;
        find_idle = 0;
        ptosend = 0;
        bursting = 0;

	// ---------- HCFstart
	backoffOn_ = 0;
	edcaTX_status = true;
	// ---------- HCFend

        rcv_ifq_tcp = 0;
        rcv_ifq_ack = 0;
        rcv_ifq_message = 0;
        rcv_ifq_ARP = 0;
	rcv_ifq_udp = 0;
        rcv_ifq_voip = 0;
        rcv_ifq_mac_cts = 0;
        rcv_ifq_mac_rts = 0;
        rcv_ifq_mac_ack = 0;
        send_erm_tcp = 0;
        send_erm_ack = 0;
        send_erm_message = 0;
        send_erm_ARP = 0;
	send_erm_udp = 0;
        send_erm_voip = 0;
        send_erm_mac_cts = 0;
        send_erm_mac_rts = 0;
        send_erm_mac_ack = 0;
        rcv_chn_tcp = 0;
        rcv_chn_ack = 0;
        rcv_chn_message = 0;
        rcv_chn_ARP = 0;
	rcv_chn_udp = 0;
        rcv_chn_voip = 0;
        rcv_chn_mac_cts = 0;
        rcv_chn_mac_rts = 0;
        rcv_chn_mac_ack = 0;
        rcv_chn_my_tcp = 0;
        rcv_chn_my_ack = 0;
        rcv_chn_my_message = 0;
        rcv_chn_my_ARP = 0;
	rcv_chn_my_udp = 0;
        rcv_chn_my_voip = 0;
        rcv_chn_my_mac_cts = 0;
        rcv_chn_my_mac_rts = 0;
        rcv_chn_my_mac_ack = 0;
        rcv_chn_other_tcp = 0;
        rcv_chn_other_ack = 0;
        rcv_chn_other_message = 0;
        rcv_chn_other_ARP = 0;
        rcv_chn_other_udp = 0;
        rcv_chn_other_voip = 0;
        rcv_chn_other_mac_cts = 0;
        rcv_chn_other_mac_rts = 0;
        rcv_chn_other_mac_ack = 0;
        rcv_chn_err_tcp = 0;
        rcv_chn_err_ack = 0;
        rcv_chn_err_message = 0;
        rcv_chn_err_ARP = 0;
	rcv_chn_err_udp = 0;
        rcv_chn_err_voip = 0;
        rcv_chn_err_mac_cts = 0;
        rcv_chn_err_mac_rts = 0;
        rcv_chn_err_mac_ack = 0;
        rcv_chn_col_tcp = 0;
        rcv_chn_col_ack = 0;
        rcv_chn_col_message = 0;
        rcv_chn_col_ARP = 0;
	rcv_chn_col_udp = 0;
        rcv_chn_col_voip = 0;
        rcv_chn_col_mac_cts = 0;
        rcv_chn_col_mac_rts = 0;
        rcv_chn_col_mac_ack = 0;
        col = 0;
        virt_col = 0;
        busy_tcp = 0;
        busy_ack = 0;
        busy_message = 0;
        busy_ARP = 0;
        busy_mac_cts = 0;
        busy_mac_rts = 0;
        busy_mac_ack = 0;
        max_rtx_tcp = 0;
        max_rtx_ack = 0;
        max_rtx_message = 0;
        max_rtx_ARP = 0;
	busy_udp = 0;
        busy_voip = 0;
        max_rtx_mac_cts = 0;
        max_rtx_mac_rts = 0;
        max_rtx_mac_ack = 0;
        rtx_tcp = 0;
        rtx_ack = 0;
        rtx_message = 0;
        rtx_ARP = 0;
        rtx_udp = 0;
        rtx_voip = 0;	
	max_rtx_udp = 0;
        max_rtx_voip = 0;
        rtx_mac_cts = 0;
        rtx_mac_rts = 0;
        rtx_mac_ack = 0;
        send_ll_tcp = 0;
        send_ll_ack = 0;
        send_ll_message = 0;
        send_ll_ARP = 0;
	send_ll_udp = 0;
        send_ll_voip = 0;
        send_ll_mac_cts = 0;
        send_ll_mac_rts = 0;
        send_ll_mac_ack = 0;
        dupl_tcp = 0;
        dupl_ack = 0;
        dupl_message = 0;
        dupl_ARP = 0;
	dupl_udp = 0;
        dupl_voip = 0;
        dupl_mac_cts = 0;
        dupl_mac_rts = 0;
        dupl_mac_ack = 0;
	// ---------- HCFstart
	cfb_time = 0;
	cfb_num = 0;
	cfb_num_bad_end = 0;
	// ---------- HCFend
	// XXX v2.0Start
	cfb_data_fast_recv = 0;
	cfb_data_slow_recv = 0;
	cfb_voice_fast_recv = 0;
	cfb_voice_slow_recv = 0;
	cfb_noDataToSend = 0;
	// XXX v2.0End
}

/*###################################################################
#                                                                   #
# set_tx_state: changes the transmitting state                      #
#                                                                   #
#         x: the new transmitting state value (MAC_RTS, MAC_SEND,   #
#            MAC_CTS, MAC_ACK or MAC_IDLE)                          #
#                                                                   #
###################################################################*/

void Mac802_11e::set_tx_state(MacState x) {
        tx_state_ = x;
        for (int i=0 ; i<mac_target_->num_tc ; i++) {           // The transmitting state is changed, so we must check whether it is
                mac_target_->check_backoff_timer(i);            // necessary to stop/start the backoff timers of all virtual queues
        }
}

/*###################################################################
#                                                                   #
# dump: used for debug, prints some statistics about the actual     #
#       MAC state                                                   #
#                                                                   #
#       fname: the file to be used to write the statistics          #
#                                                                   #
###################################################################*/

void Mac802_11e::dump(char *fname) {
	fprintf(stderr,
		"\n%s --- (INDEX: %d, time: %f, name %s, id %d)\n",
		fname, mac_target_->index_, NOW, mac_target_->name(), nq);

	fprintf(stderr,
		"\ttx_state_: %x, rx_state_: %x, nav: %f, int_nav: %f, idle: %d\n",
		tx_state_, mac_target_->rx_state_, mac_target_->nav_, int_nav, is_idle());

	fprintf(stderr,
		"\tpktTx_: %x, pktRx_: %x, pktRTS_: %x, pktCTRL_: %x, callback: %x\n",
		(int) pktTx_, (int) mac_target_->pktRx_, (int) pktRTS_,
		(int) pktCTRL_, (int) mac_target_->callback_);

	fprintf(stderr,
		"\tDefer: %d, Backoff: %d (%d), Recv: %d, Timer: %d Nav: %d\n",
		mhDefer_.busy(), mhBackoff_.busy(), mhBackoff_.paused(),
		mac_target_->mhRecv_.busy(), mhSend_.busy(), mac_target_->mhNav_.busy());
	fprintf(stderr,
		"\tBackoff Slot Remaining: %f\n",
		mhBackoff_.remaining());
        fprintf(stderr,
		"\tBursting: %d, txop: %d, ptosend : %d\n",
		bursting, mac_target_->txop, ptosend);
}

/*###################################################################
#                                                                   #
# tx_resume: starts a new access procedure, according to the type   #
#            of the frame that will be sent. If a new frame is not  #
#            available, a request to the Interface Queue is         #
#            advanced. During CFB or CFP, it just resets the        #
#            tx_state_ and possibly asks for new data to send       #
#                                                                   #
###################################################################*/

void Mac802_11e::tx_resume() {
	double rTime;
	assert(mhSend_.busy() == 0);
	assert(mhDefer_.busy() == 0);

	// ---------- HCFstart
	// During the CFP the TC doesn't resume the transmission
	// but just resets its tx_state_, frees the txop and
	// possibly asks for a new packet
	if(mac_target_->cfp_ || (mac_target_->cfb_ != -1)) {
		// No outgoing data packet, then
		// ask for a new packet to send
		if(!pktTx_) {
			mac_target_->need_pkt(nq);
		}
		set_tx_state(MAC_IDLE);
		return;
	}
	// ---------- HCFend

	if(pktCTRL_) {
		/*
		 *  Need to send a CTS or ACK.
		 */
      if (mhDefer_.busy()) {
         mhDefer_.stop();
      }
      mhDefer_.start(mac_target_->sifs_);     // Before sending CTS or ACK, we must wait for a SIFS time
	} else if(pktRTS_) {
           if(mhBackoff_.busy() == 0) {
               find_idle = 1;          // The medium seems to be idle, so we try to trasmit after AIFS/EIFS without doing backoff
               if (mac_target_->err_frame == 0) {
                  mhDefer_.start(difs_);                  // Last frame was correctly received  ---> wait AIFS
               } else {
                  mhDefer_.start(mac_target_->eifs_);     // Last frame was not correctly received  ---> wait EIFS
               }
           }
	} else if(pktTx_) {
		if(mhBackoff_.busy() == 0) {
			hdr_cmn *ch = HDR_CMN(pktTx_);
			struct hdr_mac802_11 *mh = HDR_MAC802_11(pktTx_);

			if ((u_int32_t) ch->size() < macmib_->RTSThreshold ||           // RTS/CTS handshake not used
			    (u_int32_t) ETHER_ADDR(mh->dh_da) == MAC_BROADCAST) {

                                find_idle = 1;          // The medium seems to be idle, so we try to trasmit after AIFS/EIFS
                                                        // without doing backoff

                                if (mac_target_->err_frame == 0) {
                                        mhDefer_.start(difs_);                  // Last frame was correctly received  ---> wait AIFS
                                } else {
                                        mhDefer_.start(mac_target_->eifs_);     // Last frame was not correctly received  ---> wait EIFS
                                }
			} else {
				mhDefer_.start(mac_target_->sifs_);     // RTS/CTS is used and we have just received the CTS frame, so
			}                                               // we can transmit the DATA frame after a SIFS time
		}
	} else {
                mac_target_->need_pkt(nq);      // ... Otherwise informs the Interface Queue that we are available to receive
   }                                            //       a new frame to transmit
	set_tx_state(MAC_IDLE);
}

/*###################################################################
#                                                                   #
# sendHandler: called by mac-timers.cc when a transmission timeout  #
#              expires                                              #
#                                                                   #
###################################################################*/

void Mac802_11e::sendHandler() {
	send_timer();
}

/*###################################################################
#                                                                   #
# backoffHandler: called by mac-timers.cc when a backoff phase is   #
#                 ended. Before transmitting, this function check   #
#                 that there is not a virtual collision             #
#                                                                   #
###################################################################*/

void Mac802_11e::backoffHandler() {

	if(virtual_coll == 1) {                  // This queue has virtually collided
                hdr_mac802_11* mh;
                hdr_cmn *ch;

                if (pktRTS_ != 0) {
                        ch = HDR_CMN(pktRTS_);
                        set_tx_state(MAC_RTS);

                } else if (pktTx_ != 0) {
                        ch = HDR_CMN(pktTx_);
                        set_tx_state(MAC_SEND);

                } else {                        // If the backoff is expired, but the queue has not packets to send, then
                        virtual_coll = 0;       // there is not a virtual collision
                        return;
                }

                virt_col++;     // Statistics: per-queue virtual collision counter

                switch (ch->ptype()) {          // Statistics: there is a virtual collision, so the MAC is busy by another virtual queue
                        case PT_TCP:
                                busy_tcp++;
                                break;
                        case PT_ACK:
                                busy_ack++;
                                break;
                        case PT_MESSAGE:
                                busy_message++;
                                break;
                        case PT_ARP:
                                busy_ARP++;
                                break;
                        case PT_VBR:
                        case PT_CBR:
                                busy_udp++;
                                break;
                        case PT_VOIP:
                                busy_voip++;
                                break;	
                        case PT_MAC_RTS:
                                busy_mac_rts++;
                                break;
                        case PT_MAC_CTS:
                                busy_mac_cts++;
                                break;
                        case PT_MAC_ACK:
                                busy_mac_ack++;
                                break;
                }

                // 'send_timer()' doubles the contenetion window and reschedules the frame transmission by
                // starting a new backoff phase, as if there has been a real collision on the channel
                send_timer();  

                virtual_coll = 0;       // This ' virtual collision has been solved

                return;         // Don't transmit on the channel
        }

	if(check_pktRTS() == 0)         // Transmit a RTS frame or ...
		return;

	if(check_pktTx() == 0)          // ... transmit a DATA frame
		return;
}

/*###################################################################
#                                                                   #
# deferHandler: called by mac-timers.cc when the defer timer has    #
#               expired, so after waiting for a fixed time interval #
#                                                                   #
###################################################################*/

void Mac802_11e::deferHandler() {
	assert(pktCTRL_ || pktRTS_ || pktTx_);

        if (find_idle == 1) {      // In this case, the channel was idle when the queue received a new frame
                find_idle = 0;     // from the Interface Queue, and the channel is still idle after AIFS/EIFS,
        }                          // so the frame can be transmitted without doing a backoff

	if(check_pktCTRL() == 0)        // Transmit a CTS or an ACK
		return;

	assert(mhBackoff_.busy() == 0);

	if(check_pktRTS() == 0)         // Transmit a RTS
		return;

	if(check_pktTx() == 0)          // Transmit a DATA frame
		return;
}

/*###################################################################
#                                                                   #
# int_navHandler: called by mac-timers.cc when the internal NAV     #
#                 timer has expired                                 #
#                                                                   #
###################################################################*/

void Mac802_11e::int_navHandler() {
        if(is_idle() && mhBackoff_.paused()) {          // Since the internal NAV is expired, if the channel is idle it
                                                        // is necessary to restart the backoff
                if (mac_target_->err_frame == 0) {
                        mhBackoff_.resume(difs_);       // Last frame was correctly received  ---> wait AIFS
                                                        // before decrementing backoff
                } else {
                        mhBackoff_.resume(mac_target_->eifs_);  // Last frame was not correctly received  ---> wait EIFS
                }                                               // before decrementing backoff
        }
}

/*###################################################################
#                                                                   #
# send_timer: called when a transmission timeout expires, this      #
#             method performs the operations that have to be done   #
#             according to the type of the frame that was sent:     #
#             it can retransmit the frame or discard it             #
#                                                                   #
###################################################################*/

void Mac802_11e::send_timer() {

   if (bursting > 1) {             // In bursting mode, the timeout expiration forces the end of the current burst, and the last
      bursting = 1;           // frame not "acked" must be retransmitted with the normal access procedure
      ptosend = 0;
      ((BS_QueueHandler*)(mac_target_->callback_))->stop_bursting(nq);        // Informs the Interface Queue about the end
   }                                                                               // of this burst

	switch(tx_state_) {

	/*
	 * Sent a RTS, but did not receive a CTS.
	 */

	case MAC_RTS:
		RetransmitRTS();
		break;

	/*
	 * Sent a CTS, but did not receive a DATA packet.
	 */

	case MAC_CTS:
		assert(pktCTRL_);                               // CTS frames are not retransmitted
		Packet::free(pktCTRL_); pktCTRL_ = 0;
		break;

	/*
	 * Sent DATA, but did not receive an ACK packet.
	 */

	case MAC_SEND:
		RetransmitDATA();
		break;

	/*
	 * Sent an ACK, and now ready to resume transmission.
	 */

	case MAC_ACK:
		assert(pktCTRL_);                               // ACK frames are not retransmitted
		Packet::free(pktCTRL_); pktCTRL_ = 0;
		break;

	// ---------- HCFstart
	// Sent a Reservation Request
	
	case MAC_RR:
		Packet::free(pktCTRL_);
		pktCTRL_ = 0;
		set_tx_state(MAC_IDLE);
		return;

	case MAC_QOS_DATA:
	case MAC_QOS_DATA_CFACK:
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) { 
			fprintf(macLog, "[%d.%d] Send timer expired during CFB %f, attempt retransmission ", mac_target_->index_, nq, NOW);
		}
#endif //=========================================================================
		// Set the retry bit
		HDR_MAC802_11(pktTx_)->dh_fc.fc_retry = 1;

		switch (HDR_CMN(pktTx_)->ptype()) {     // Statistics: retransmitted frames
			case PT_TCP:
				++rtx_tcp;
				break;
			case PT_ACK:
				++rtx_ack;
				break;
			case PT_MESSAGE:
				++rtx_message;
				break;
			case PT_ARP:
				++rtx_ARP;
				break;
         case PT_VBR:
			case PT_CBR:
				++rtx_udp;
				break;
			case PT_VOIP:
				++rtx_voip;
				break;	
			case PT_MAC_RTS:
				++rtx_mac_rts;
				break;
			case PT_MAC_CTS:
				++rtx_mac_cts;
				break;
			case PT_MAC_ACK:
				++rtx_mac_ack;
				break;
		}

		set_tx_state(MAC_IDLE);
		mac_target_->cfbHandler();
		return;;

	case MAC_QOS_CFACK:
		// QoS CF-Ack or Null packets
		// are not freed here, but
		// only when the TXOP ends.
		// Return, it is necessary
		// to keep current tx_state_
		return;
	// ---------- HCFend
	
	case MAC_IDLE:
		break;
	default:
		assert(0);
	}
	tx_resume();
}


/*###################################################################
#                                                                   #
# check_pktCTRL: after the queue has waited for SIFS, this method   #
#                checks if the transmission of the ACK or CTS can   #
#                proceed                                            #
#                                                                   #
#    Returns -1 if the frame is missing or the transmission state   #
#    is not right, 0 otherwise                                      #
#                                                                   #
###################################################################*/

int Mac802_11e::check_pktCTRL() {
	struct hdr_mac802_11 *mh;
	double timeout;

	if(pktCTRL_ == 0)
		return -1;
	if(tx_state_ == MAC_CTS || tx_state_ == MAC_ACK)
		return -1;

	mh = HDR_MAC802_11(pktCTRL_);
	// XXX v2.0Start
	// During CP set AP's data transmission rate according to destination
	if(!(mac_target_->cfp_) &&
	   mac_target_->index_ == mac_target_->HCmacAddress()) {
		mac_target_->dataRate_ = mac_target_->bitRates_[(int)ETHER_ADDR(mh->dh_da)];
	}
	// XXX v2.0End

	switch(mh->dh_fc.fc_subtype) {

	/*
	 *  If the medium is not IDLE, don't send the CTS.
	 */

	case MAC_Subtype_CTS:
		if(mac_target_->nav_ > Scheduler::instance().clock()) {                 // The medium is not idle, so the CTS frame
                        ++busy_mac_cts;                                                 // must be discarded
			mac_target_->discard(pktCTRL_, DROP_MAC_BUSY); pktCTRL_ = 0;
			return 0;
		}

                // the CTS can be sent

		set_tx_state(MAC_CTS);

		/*
		 * timeout:  cts + data tx time calculated by
		 *           adding cts tx time to the cts duration
		 *           minus ack tx time -- this timeout is
		 *           a guess since it is unspecified
		 *           (note: mh->dh_duration == cf->cf_duration)
		 */
		
		// XXX v2.0Start
		timeout = mac_target_->txtime(ETHER_CTS_LEN, mac_target_->basicRate_)           // The transmission timeout for a CTS
			+ DSSS_MaxPropagationDelay                                              // frame is equal to:
			+ mac_target_->sec(mh->dh_duration)
			+ DSSS_MaxPropagationDelay                                              //     CTS + 2us + sifs + pkt + 2us
			- mac_target_->sifs_
			- mac_target_->txtime(ETHER_ACK_LEN, mac_target_->dataRate_);
		// XXX V2.0End

		break;

		/*
		 * IEEE 802.11 specs, section 9.2.8
		 * Acknowledments are sent after an SIFS, without regard to
		 * the busy/idle state of the medium.
		 */

	case MAC_Subtype_ACK:                   // No virtual carrier sensing
		set_tx_state(MAC_ACK);

		// XXX v2.0Start
		timeout = mac_target_->txtime(ETHER_ACK_LEN, mac_target_->dataRate_);   // The transmission timeout is simply equal to
                // XXX v2.0End                                                          // the transmission time of the ACK, because
		break;                                                                  // we won't receive a reply for this frame
	default:
		fprintf(stderr, "check_pktCTRL:Invalid MAC Control subtype\n");
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) { 
			fprintf(macLog, "[%d] Invalid CTS/ACK frame (type: %x.%x) %f - EXITING ", mac_target_->index_, mh->dh_fc.fc_type, mh->dh_fc.fc_subtype, NOW);
		}
#endif //=========================================================================
		exit(1);
	}
        mac_target_->transmit(pktCTRL_, timeout, nq);   // Call the transmission routine
	return 0;
}

/*###################################################################
#                                                                   #
# check_pktRTS: after the backoff expiration, this function         #
#               verifies if the RTS frame can be transmitted        #
#                                                                   #
#    Returns -1 if the frame is missing, 0 otherwise                #
#                                                                   #
###################################################################*/

int Mac802_11e::check_pktRTS() {
	struct hdr_mac802_11 *mh;
	double timeout;

	assert(mhBackoff_.busy() == 0);

	if(pktRTS_ == 0)
 		return -1;

	mh = HDR_MAC802_11(pktRTS_);

 	switch(mh->dh_fc.fc_subtype) {
	case MAC_Subtype_RTS:
		if(! is_idle()) {      
			mhBackoff_.start(cw_, is_idle());       // The medium is not idle, so the RTS frame must be transmitted
			return 0;                               // after a new backoff procedure
		}

                // the RTS can be sent

		set_tx_state(MAC_RTS);
		timeout = mac_target_->txtime(ETHER_RTS_LEN, mac_target_->basicRate_)      // The transmission timeout for a RTS
			+ DSSS_MaxPropagationDelay			                   // frame is equal to:
			+ mac_target_->sifs_
			+ mac_target_->txtime(ETHER_CTS_LEN, mac_target_->basicRate_)      //     RTS + 2us + sifs + cts + 2us
			+ DSSS_MaxPropagationDelay;
		break;
	default:
		fprintf(stderr, "check_pktRTS:Invalid MAC Control subtype\n");
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) { 
			fprintf(macLog, "[%d] Invalid RTS frame (type: %x.%x) %f - EXITING ", mac_target_->index_, mh->dh_fc.fc_type, mh->dh_fc.fc_subtype, NOW);
		}
#endif //=========================================================================
		exit(1);
	}
        mac_target_->transmit(pktRTS_, timeout, nq);    // Call the transmission routine
	return 0;
}

/*###################################################################
#                                                                   #
# check_pktTx: after the backoff expiration (or after a SIFS period #
#              if the RTS/CTS handshake is used), this function     #
#              checks if the DATA frame can be transmitted          #
#                                                                   #
#    Returns -1 if the frame is missing, 0 otherwise                #
#                                                                   #
###################################################################*/

int Mac802_11e::check_pktTx() {
	struct hdr_mac802_11 *mh;
	double timeout;

	assert(mhBackoff_.busy() == 0);

	if(pktTx_ == 0) {
                mac_target_->txop = -1;
		return -1;
        }

	mh = HDR_MAC802_11(pktTx_);
	// XXX v2.0Start
	// During CP set AP's data transmission rate according to destination
	if(!(mac_target_->cfp_) &&
	   mac_target_->index_ == mac_target_->HCmacAddress()) {
		mac_target_->dataRate_ = mac_target_->bitRates_[(int)ETHER_ADDR(mh->dh_da)];
	}
	// XXX v2.0End
       	
	switch(mh->dh_fc.fc_subtype) {
	case MAC_Subtype_Data:
		if(! is_idle()) {                               // The medium is not idle, so the DATA frame must be transmitted
                                                                // after a new backoff procedure

			sendRTS(ETHER_ADDR(mh->dh_da));         // Checks if a RTS frame must precede this DATA frame

			mhBackoff_.start(cw_, is_idle());

                        if((u_int32_t) (HDR_CMN(pktTx_))->size() > macmib_->RTSThreshold) {

                                switch ((HDR_CMN(pktTx_))->ptype()) {   // Statistics: the MAC is busy
                                        case PT_TCP:
                                                busy_tcp++;
                                                break;
                                        case PT_ACK:
                                                busy_ack++;
                                                break;
                                        case PT_MESSAGE:
                                                busy_message++;
                                                break;
                                        case PT_ARP:
                                                busy_ARP++;
                                                break;
                                        case PT_VBR:
                                        case PT_CBR:
                                                busy_udp++;
                                                break;
                                        case PT_VOIP:
                                                busy_voip++;
                                                break;	
                                        case PT_MAC_RTS:
                                                busy_mac_rts++;
                                                break;
                                        case PT_MAC_CTS:
                                                busy_mac_cts++;
                                                break;
                                        case PT_MAC_ACK:
                                                busy_mac_ack++;
                                                break;
                                }
                        }
			return 0;
		}

                // the DATA frame can be sent

		set_tx_state(MAC_SEND);
		if((u_int32_t)ETHER_ADDR(mh->dh_da) != MAC_BROADCAST)                           // The transmission timeout for a DATA
			// XXX v2.0Start                                                        // frame is equal to:
			timeout = mac_target_->txtime(pktTx_)
				+ DSSS_MaxPropagationDelay
				+ mac_target_->sifs_                                            //     pkt + 2us + sifs + ack + 2us
				+ mac_target_->txtime(ETHER_ACK_LEN, mac_target_->dataRate_)
				+ DSSS_MaxPropagationDelay;
			// XXX v2.0End
		else
			timeout = mac_target_->txtime(pktTx_);                  // For a broadcast frame, the transmission timeout
                                                                                // is simply equal to the transmission time of the
		break;                                                          // frame, because we won't receive a reply for this frame
	default:
		fprintf(stderr, "check_pktTx:Invalid MAC Control subtype\n");
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) { 
			fprintf(macLog, "[%d] Invalid Data frame (type: %x.%x) %f - EXITING ", mac_target_->index_, mh->dh_fc.fc_type, mh->dh_fc.fc_subtype, NOW);
		}
#endif //=========================================================================
		exit(1);
	}

        mac_target_->transmit(pktTx_, timeout, nq);     // Call the transmission routine
	return 0;
}

/*###################################################################
#                                                                   #
# sendRTS: creates a new RTS frame and prepares it for transmission #
#                                                                   #
#       dst: the MAC destination address for the RTS frame          #
#                                                                   #
###################################################################*/

void Mac802_11e::sendRTS(int dst) {
	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct rts_frame *rf = (struct rts_frame*)p->access(hdr_mac::offset_);

	assert(pktTx_);
	assert(pktRTS_ == 0);

	/*
	 *  The RTS/CTS exchange is not performed when:
         *  - the size of the packet is smaller than RTSThreshold
	 *  - the destination is broadcast
         *  - during bursting, for all frames following the first
	 */

	if( (u_int32_t) HDR_CMN(pktTx_)->size() < macmib_->RTSThreshold ||
	    (u_int32_t) dst == MAC_BROADCAST ||
            mac_target_->txop == nq && pktCTRL_ == 0 && mac_target_->tx_active_ != nq) {
		Packet::free(p);
		return;
	}

	ch->uid() = 0;

        ch->ptype() = PT_MAC_RTS;
	// ---------- HCFstart
	ch->mac_tc = (u_int8_t)nq;
	ch->mac_tc_aux = (u_int8_t)(mac_target_->num_tc);
	// ---------- HCFend

	ch->size() = ETHER_RTS_LEN;
	ch->iface() = -2;
	ch->error() = 0;

	bzero(rf, MAC_HDR_LEN);

	rf->rf_fc.fc_protocol_version = MAC_ProtocolVersion;
 	rf->rf_fc.fc_type	= MAC_Type_Control;
 	rf->rf_fc.fc_subtype	= MAC_Subtype_RTS;
 	rf->rf_fc.fc_to_ds	= 0;
 	rf->rf_fc.fc_from_ds	= 0;
 	rf->rf_fc.fc_more_frag	= 0;
 	rf->rf_fc.fc_retry	= 0;
 	rf->rf_fc.fc_pwr_mgt	= 0;
 	rf->rf_fc.fc_more_data	= 0;
 	rf->rf_fc.fc_wep	= 0;
 	rf->rf_fc.fc_order	= 0;

	STORE4BYTE(&dst, (rf->rf_ra));          // Stores the destination MAC address

	/* store rts tx time */

 	ch->txtime() = mac_target_->txtime(ch->size(), mac_target_->basicRate_);

	STORE4BYTE(&(mac_target_->index_), (rf->rf_ta));        // Stores the source MAC address

	/* calculate rts duration field, used for NAV updating */
	// XXX v2.0Start
	// During CP set AP's data transmission rate according to destination
	if(!(mac_target_->cfp_) &&
	   mac_target_->index_ == mac_target_->HCmacAddress()) {
		mac_target_->dataRate_ = mac_target_->bitRates_[dst];
	}

	rf->rf_duration = mac_target_->usec(mac_target_->sifs_ // sifs + CTS + sifs + pkt +sifs +ack
			       + mac_target_->txtime(ETHER_CTS_LEN, mac_target_->basicRate_)
			       + mac_target_->sifs_
			       + mac_target_->txtime(pktTx_)
			       + mac_target_->sifs_
			       + mac_target_->txtime(ETHER_ACK_LEN, mac_target_->dataRate_));
	// XXX v2.0End

	pktRTS_ = p;
}

/*###################################################################
#                                                                   #
# sendCTS: creates a new CTS frame and prepares it for transmission #
#                                                                   #
#       dst: the MAC destination address for the CTS frame          #
#       rts_duration: the duration field of the received RTS frame  #
#                                                                   #
###################################################################*/

void Mac802_11e::sendCTS(int dst, double rts_duration) {
	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct cts_frame *cf = (struct cts_frame*)p->access(hdr_mac::offset_);

	assert(pktCTRL_ == 0);

	ch->uid() = 0;

        ch->ptype() = PT_MAC_CTS;
	// ---------- HCFstart
        ch->mac_tc = (u_int8_t)(HDR_CMN(mac_target_->pktRx_)->mac_tc);
	ch->mac_tc_aux = (u_int8_t)(mac_target_->num_tc);
	// ---------- HCFend
        
	ch->size() = ETHER_CTS_LEN;
	ch->iface() = -2;
	ch->error() = 0;
	bzero(cf, MAC_HDR_LEN);

	cf->cf_fc.fc_protocol_version = MAC_ProtocolVersion;
	cf->cf_fc.fc_type	= MAC_Type_Control;
	cf->cf_fc.fc_subtype	= MAC_Subtype_CTS;
 	cf->cf_fc.fc_to_ds	= 0;
 	cf->cf_fc.fc_from_ds	= 0;
 	cf->cf_fc.fc_more_frag	= 0;
 	cf->cf_fc.fc_retry	= 0;
 	cf->cf_fc.fc_pwr_mgt	= 0;
 	cf->cf_fc.fc_more_data	= 0;
 	cf->cf_fc.fc_wep	= 0;
 	cf->cf_fc.fc_order	= 0;

	STORE4BYTE(&dst, (cf->cf_ra));  // Stores the destination MAC address

	/* store cts tx time */

	ch->txtime() = mac_target_->txtime(ch->size(), mac_target_->basicRate_);

	/* calculate cts duration, used for NAV updating */

	cf->cf_duration = mac_target_->usec(mac_target_->sec(rts_duration)                    // sifs + pkt + sifs + ACK
			       - mac_target_->sifs_
			       - mac_target_->txtime(ETHER_CTS_LEN, mac_target_->basicRate_));

        STORE4BYTE(&(mac_target_->index_), (HDR_MAC802_11(p)->dh_sa));  // Stores the source MAC address

	pktCTRL_ = p;
}

/*###################################################################
#                                                                   #
# sendACK: creates a new ACK frame and prepares it for transmission #
#                                                                   #
#       dst: the MAC destination address for the ACK frame          #
#                                                                   #
###################################################################*/

void Mac802_11e::sendACK(int dst) {
	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct ack_frame *af = (struct ack_frame*)p->access(hdr_mac::offset_);

	assert(pktCTRL_ == 0);

	ch->uid() = 0;

        ch->ptype() = PT_MAC_ACK;
	// ---------- HCFstart
	// If this method is invoked by
	// Mac802_11e::cfbSendCFACK,
	// pktRx_ may have been deleted
	if(mac_target_->pktRx_)
	        ch->mac_tc = (u_int8_t)(HDR_CMN(mac_target_->pktRx_)->mac_tc);
	else
		ch->mac_tc = (u_int8_t)nq;
	ch->mac_tc_aux = (u_int8_t)(mac_target_->num_tc);
	// ---------- HCFend

	ch->size() = ETHER_ACK_LEN;
	ch->iface() = -2;
	ch->error() = 0;

	bzero(af, MAC_HDR_LEN);

	af->af_fc.fc_protocol_version = MAC_ProtocolVersion;
 	af->af_fc.fc_type	= MAC_Type_Control;
 	af->af_fc.fc_subtype	= MAC_Subtype_ACK;
 	af->af_fc.fc_to_ds	= 0;
 	af->af_fc.fc_from_ds	= 0;
 	af->af_fc.fc_more_frag	= 0;
 	af->af_fc.fc_retry	= 0;
 	af->af_fc.fc_pwr_mgt	= 0;
 	af->af_fc.fc_more_data	= 0;
 	af->af_fc.fc_wep	= 0;
 	af->af_fc.fc_order	= 0;

	STORE4BYTE(&dst, (af->af_ra));  // Stores the destination MAC address

	/* store ack tx time */
	
	// XXX v2.0Start
	// During CP set AP's data transmission rate according to destination
	if(!(mac_target_->cfp_) &&
	   mac_target_->index_ == mac_target_->HCmacAddress()) {
		mac_target_->dataRate_ = mac_target_->bitRates_[dst];
	}
 	ch->txtime() = mac_target_->txtime(ch->size(), mac_target_->dataRate_);
	// XXX v2.0End

	/* calculate ack duration */

 	af->af_duration = 0;    // The duration is null, because the ACK is always the last frame in each exchange

        STORE4BYTE(&(mac_target_->index_), (HDR_MAC802_11(p)->dh_sa));          // Stores the source MAC address

	pktCTRL_ = p;
}

/*###################################################################
#                                                                   #
# sendDATA: Prepares the DATA frame 'p' for transmission            #
#                                                                   #
#      p: the DATA frame to be transmitted                          #
#                                                                   #
###################################################################*/

void Mac802_11e::sendDATA(Packet *p) {
	hdr_cmn* ch = HDR_CMN(p);
	struct hdr_mac802_11* dh = HDR_MAC802_11(p);

	assert(pktTx_ == 0);

	/*
	 * Update the MAC header
	 */

	ch->size() += ETHER_HDR_LEN11;

	dh->dh_fc.fc_protocol_version = MAC_ProtocolVersion;
	dh->dh_fc.fc_type       = MAC_Type_Data;
	dh->dh_fc.fc_subtype    = MAC_Subtype_Data;

	dh->dh_fc.fc_to_ds      = 0;
	dh->dh_fc.fc_from_ds    = 0;
	dh->dh_fc.fc_more_frag  = 0;
	dh->dh_fc.fc_retry      = 0;
	dh->dh_fc.fc_pwr_mgt    = 0;
	dh->dh_fc.fc_more_data  = 0;
	dh->dh_fc.fc_wep        = 0;
	dh->dh_fc.fc_order      = 0;
	
        if (mac_target_->txop == nq && pktCTRL_ == 0 && mac_target_->tx_active_ != nq) {

                ch->burst_no_rts = 1;                   // In bursting mode, this flag allows the destination node
        } else {                                        // to know if this frame is the first in a burst, or if it
                ch->burst_no_rts = 0;                   // is one of the following
        }

	if((u_int32_t)ETHER_ADDR(dh->dh_da) != MAC_BROADCAST) {

		/* store data tx time for unicast packets */
		// XXX v2.0Start
		// During CP set AP's data transmission rate according to destination
		if(!(mac_target_->cfp_) &&
		   mac_target_->index_ == mac_target_->HCmacAddress()) {
			mac_target_->dataRate_ = mac_target_->bitRates_[(int)ETHER_ADDR(dh->dh_da)];
		}

		ch->txtime() = mac_target_->txtime(ch->size(), mac_target_->dataRate_);
		dh->dh_duration = mac_target_->usec(mac_target_->txtime(ETHER_ACK_LEN, mac_target_->dataRate_)   // The duration is simply equal to
 				       + mac_target_->sifs_);                                                    // sifs + ack
		// XXX v2.0End
	} else {

		/* store data tx time for broadcast packets (see 9.6) */
      ch->txtime() = mac_target_->txtime(ch->size(), mac_target_->basicRate_);

		dh->dh_duration = 0;    //In this case, the duration is null, because broadcast packets are never "acked"
	}

	pktTx_ = p;
}

// ---------- HCFstart
/*###################################################################
#                                                                   #
# capSendRR: creates a new Reservation Request frame and prepares   #
#            it for transmission                                    #
#                                                                   #
#            dst: the MAC destination address for the RR frame      #
#                                                                   #
###################################################################*/

void Mac802_11e::capSendRR(int dst) {

	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct rr_frame *rr = (struct rr_frame*)p->access(hdr_mac::offset_);

	ch->uid() = 0;
	ch->ptype() = PT_MAC_RR;
	// ---------- HCFstart
        ch->mac_tc = (u_int8_t)nq;
	ch->mac_tc_aux = (u_int8_t)(mac_target_->num_tc);
	// ---------- HCFend
	ch->size() = ETHER_RR_LEN;
	ch->iface() = -2;
	ch->error() = 0;

	bzero(rr, MAC_HDR_LEN);

	rr->rr_fc.fc_protocol_version = MAC_ProtocolVersion;
 	rr->rr_fc.fc_type	= MAC_Type_Control;
 	rr->rr_fc.fc_subtype	= MAC_Subtype_RR;
 	rr->rr_fc.fc_to_ds	= 0;
 	rr->rr_fc.fc_from_ds	= 0;
 	rr->rr_fc.fc_more_frag	= 0;
 	rr->rr_fc.fc_retry	= 0;
 	rr->rr_fc.fc_pwr_mgt	= 0;
 	rr->rr_fc.fc_more_data	= 0;
 	rr->rr_fc.fc_wep	= 0;
 	rr->rr_fc.fc_order	= 0;

	// Store destination and source addresses
	STORE4BYTE(&dst, (rr->rr_ra));
	STORE4BYTE(&(mac_target_->index_), (HDR_MAC802_11(p)->dh_sa));

	// Store RR frame tx time
 	ch->txtime() = mac_target_->txtime(ch->size(), mac_target_->basicRate_);

	// Store RR frame duration
 	rr->rr_duration = 0;

	// Store requested TXOP duration
	rr->rr_txop_duration = mac_target_->usec(mac_target_->TXOPLmt);

	if(pktCTRL_) {
		Packet::free(pktCTRL_);
		pktCTRL_ = 0;
	}

	pktCTRL_ = p;

}

/*###################################################################
#                                                                   #
# cfbSendCFACK: prepares the CF-Ack frame for transmission          #
#               during the Contention Free Burst                    #
#                                                                   #
#               dst: the destination address                        #
#               tc: the TC of the Data frames that is acknowledged  #
#                   by this CF-Ack frame                            #
#                                                                   #
###################################################################*/

void Mac802_11e::cfbSendCFACK(int dst, int tc) {
	
	// Prepare the ACK packet
	sendACK(dst);
	struct hdr_cmn* ch = HDR_CMN(pktCTRL_);
	struct hdr_mac802_11* dh = HDR_MAC802_11(pktCTRL_);
	ch->mac_tc_aux = (u_int8_t)tc;
	dh->dh_fc.fc_type = MAC_Type_Data;
	dh->dh_fc.fc_subtype = MAC_Subtype_QoS_CFAck;

	// Duration value during CFB
	// is the remaining TXOP time
	// (802.11e - 9.10.2.2)
	dh->dh_duration = mac_target_->usec(mac_target_->txopLimit_
	                                    + mac_target_->txopDuration_
					    - NOW);

#ifdef DEBUGACTIVE //=============================================================
	if(NOW > DEBUGTIME) {
		fprintf(macLog, "[%d.%d] send CF-Ack (duration: %d) to [%d.%d-%d], queue length: %d ", mac_target_->index_, nq, dh->dh_duration, dst, nq, tc, mac_target_->queueLength(nq));
	}
#endif //=========================================================================

}

/*###################################################################
#                                                                   #
# cfbSendDATA: prepares the Data frame for transmission             #
#              during the Contention Free Burst                     #
#                                                                   #
#              dst: the destination address                         #
#                                                                   #
###################################################################*/

void Mac802_11e::cfbSendDATA(int dst) {

	struct hdr_mac802_11* dh = HDR_MAC802_11(pktTx_);

	u_int32_t dst_pktTx = (u_int32_t)ETHER_ADDR(dh->dh_da);

	// This should be always false because all
	// the packets must pass through the HC
	
   if(dst != dst_pktTx) {
		// CoSmO
		memset (dh->dh_da, 0, 6);
		memset (dh->dh_da, dst, 5);
		dst_pktTx = (u_int32_t)ETHER_ADDR(dh->dh_da);
		//fprintf(stderr, "node: %d -- dst: %d -- dst_pktTx: %d\n", mac_target_->addr(), dst, dst_pktTx);
		//fprintf(stderr, "Packet not directed to BS\n");
		//exit(1);
	}

	dh->dh_fc.fc_subtype = MAC_Subtype_QoS_Data;
	// Only QSTAs set the more data bit
	// if they have queued data to send
	if((mac_target_->queueLength(nq) > 1) && (mac_target_->index_ != mac_target_->HCmacAddress()))
		dh->dh_fc.fc_more_data = 1;

	// Duration value during CFB
	// is the remaining TXOP time
	// (802.11e - 9.10.2.2)
	dh->dh_duration = mac_target_->usec(mac_target_->txopLimit_
	                                    + mac_target_->txopDuration_
					    - NOW);

#ifdef DEBUGACTIVE //=============================================================
	if(NOW > DEBUGTIME) {
		fprintf(macLog, "[%d.%d] send Data { moredata: %d, queue length: %d } to [%d.%d] (duration: %d) ", mac_target_->index_, nq, dh->dh_fc.fc_more_data, mac_target_->queueLength(nq), dst, nq, dh->dh_duration);
	}
#endif //=========================================================================

}

/*###################################################################
#                                                                   #
# cfbSendDATA_CFACK: prepares the Data+CF-Ack frame for             #
#                    transmission during the Contention Free Burst  #
#                                                                   #
#                    dst: the destination address                   #
#                    tc: the Data TC if the HC sends the frame      #
#                        the CF-Ack TC if a QSTA sends the frame    #
#                                                                   #
###################################################################*/

void Mac802_11e::cfbSendDATA_CFACK(int dst, int tc) {

	struct hdr_cmn* ch = HDR_CMN(pktTx_);
	struct hdr_mac802_11* dh = HDR_MAC802_11(pktTx_);

	u_int32_t dst_pktTx = (u_int32_t)ETHER_ADDR(dh->dh_da);

	// This should be always false because:
	// - in QSTAs all the packets must pass through the HC
	// - in HC this check has already been performed
	if(dst != dst_pktTx) {
		memset (dh->dh_da, 0, 6);
		memset (dh->dh_da, dst, 5);	
		dst_pktTx = (u_int32_t)ETHER_ADDR(dh->dh_da);
		//fprintf(stderr, "node: %d -- dst: %d -- dst_pktTx: %d\n", mac_target_->addr(), dst, dst_pktTx);
		//fprintf(stderr, "Packet not directed to BS\n");
		//exit(1);
	}

	ch->mac_tc_aux = (u_int8_t)tc;
	dh->dh_fc.fc_subtype = MAC_Subtype_QoS_Data_CFAck;
	if((mac_target_->queueLength(nq) > 1) && (mac_target_->index_ != mac_target_->HCmacAddress()))
		dh->dh_fc.fc_more_data = 1;

	// Duration value during CFB
	// is the remaining TXOP time
	// (802.11e - 9.10.2.2)
	dh->dh_duration = mac_target_->usec(mac_target_->txopLimit_
	                                    + mac_target_->txopDuration_
					    - NOW);

#ifdef DEBUGACTIVE //=============================================================
	if(NOW > DEBUGTIME) {
		fprintf(macLog, "[%d.%d] send Data+CF-Ack { moredata: %d, queue length: %d } to [%d.%d-%d] (duration: %d) ", mac_target_->index_, nq, dh->dh_fc.fc_more_data, mac_target_->queueLength(nq), dst, nq, tc, dh->dh_duration);
	}
#endif //=========================================================================

}

/*###################################################################
#                                                                   #
# cfbSendNULL: prepares the Null frame for transmission             #
#              during the Contention Free Burst                     #
#                                                                   #
#              dst: the destination address                         #
#                                                                   #
###################################################################*/

void Mac802_11e::cfbSendNULL(int dst) {

	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct null_frame *nf = (struct null_frame*)p->access(hdr_mac::offset_);

	ch->uid() = 0;
	ch->ptype() = PT_MAC_NULL;
        ch->mac_tc = (u_int8_t)nq;
	ch->mac_tc_aux = (u_int8_t)(mac_target_->num_tc);
	ch->size() = ETHER_NULL_LEN;
	ch->iface() = -2;
	ch->error() = 0;

	bzero(nf, MAC_HDR_LEN);

	nf->nf_fc.fc_protocol_version = MAC_ProtocolVersion;
 	nf->nf_fc.fc_type	= MAC_Type_Data;
 	nf->nf_fc.fc_subtype	= MAC_Subtype_QoS_Null;
 	nf->nf_fc.fc_to_ds	= 0;
 	nf->nf_fc.fc_from_ds	= 0;
 	nf->nf_fc.fc_more_frag	= 0;
 	nf->nf_fc.fc_retry	= 0;
 	nf->nf_fc.fc_pwr_mgt	= 0;
 	nf->nf_fc.fc_more_data	= 0;
	nf->nf_fc.fc_wep	= 0;
 	nf->nf_fc.fc_order	= 0;

	// Store destination and source addresses
	STORE4BYTE(&dst, (nf->nf_ra));
	STORE4BYTE(&(mac_target_->index_), (HDR_MAC802_11(p)->dh_sa));

	// Store Null frame transmission time
	// XXX v2.0Start
 	ch->txtime() = mac_target_->txtime(ch->size(), mac_target_->dataRate_);
	// XXX v2.0End

	// Duration value during CFB
	// is the remaining TXOP time
	// (802.11e - 9.10.2.2)
	nf->nf_duration = mac_target_->usec(mac_target_->txopLimit_
	                                    + mac_target_->txopDuration_
					    - NOW);
	pktCTRL_ = p;

#ifdef DEBUGACTIVE //=============================================================
	if(NOW > DEBUGTIME) {
		fprintf(macLog, "[%d.%d] send Null to [%d.%d], queue length: %d (duration: %d) ", mac_target_->index_, nq, dst, nq, mac_target_->queueLength(nq), nf->nf_duration);
	}
#endif //=========================================================================

}
// ---------- HCFend

/*###################################################################
#                                                                   #
# RetransmitRTS: called when a transmission timeout expires, this   #
#            method checks if the maximum number of retransmissions #
#            has been reached, and consequently it decides to       #
#            retransmit the RTS or to discard it                    #
#                                                                   #
###################################################################*/

void Mac802_11e::RetransmitRTS() {
	assert(pktTx_);
	assert(pktRTS_);
	assert(mhBackoff_.busy() == 0);

	macmib_->RTSFailureCount++;

	ssrc_ += 1;			// Increments the STA Short Retry Count

        if(ssrc_ >= ShortRL) {          // The maximum number of retransmissions has been reached
                ++max_rtx_mac_rts;

		mac_target_->discard(pktRTS_, DROP_MAC_RETRY_COUNT_EXCEEDED); pktRTS_ = 0;      // RTS frame is discarded

		/* tell the callback the send operation failed
		   before discarding the packet */

		hdr_cmn *ch = HDR_CMN(pktTx_);
		if (ch->xmit_failure_) {

                        /*
                         *  Need to remove the MAC header so that
                         *  re-cycled packets don't keep getting
                         *  bigger.
                         */

                        ch->size() -= ETHER_HDR_LEN11;
                        ch->xmit_reason_ = XMIT_REASON_RTS;
                        ch->xmit_failure_(pktTx_->copy(),
                                          ch->xmit_failure_data_);
                }
		mac_target_->discard(pktTx_, DROP_MAC_RETRY_COUNT_EXCEEDED);
		pktTx_ = 0;        // The DATA frame is also dropped

		ssrc_ = 0;      // Resetting the STA Short Retry Count
		rst_cw();       // Resetting the Contention window to CWmin

	} else {                   // We can still retransmit the RTS
                ++rtx_mac_rts;
		struct rts_frame *rf;
		rf = (struct rts_frame*)pktRTS_->access(hdr_mac::offset_);
		rf->rf_fc.fc_retry = 1;

		inc_cw();       // Doubling the Contention Window

		mhBackoff_.start(cw_, is_idle());       // New backoff procedure
	}
}

/*###################################################################
#                                                                   #
# RetransmitDATA: called when a transmission timeout expires, this  #
#            method checks if the maximum number of retransmissions #
#            has been reached, and consequently it decides to       #
#            retransmit the DATA frame or to discard it             #
#                                                                   #
###################################################################*/

void Mac802_11e::RetransmitDATA() {
	struct hdr_cmn *ch;
	struct hdr_mac802_11 *mh;
	u_int32_t *rcount, *thresh;

	assert(mhBackoff_.busy() == 0);

	assert(pktTx_);
	assert(pktRTS_ == 0);

	ch = HDR_CMN(pktTx_);
	mh = HDR_MAC802_11(pktTx_);

	if((u_int32_t)ETHER_ADDR(mh->dh_da) == MAC_BROADCAST) {         // Broadcast packets don't get "acked" and therefore
                                                                        // are never retransmitted
		Packet::free(pktTx_);
		pktTx_ = 0;

		rst_cw();       // Resetting the Contention Window to CWmin

		mhBackoff_.start(cw_, is_idle());       // New backoff

		return;
	}

	macmib_->ACKFailureCount++;

	if((u_int32_t) ch->size() <= macmib_->RTSThreshold) {   // If the packet size is under the RTS_Threshold, use the STA Short
		rcount = &ssrc_;                                // Retry Count and the Short Retry Limit
		thresh = &ShortRL;
	}
	else {                          // ... Otherwise use the STA Long
		rcount = &slrc_;        // Retry Count and the Long Retry Limit
		thresh = &LongRL;
	}

	(*rcount)++;    // Increments the corresponding STA Retry Count

	if(*rcount > *thresh) {         // The maximum number of retransmissions has been reached

		macmib_->FailedCount++;

		/* tell the callback the send operation failed
		   before discarding the packet */

		hdr_cmn *ch = HDR_CMN(pktTx_);
		if (ch->xmit_failure_) {
                        ch->size() -= ETHER_HDR_LEN11;
                        ch->xmit_reason_ = XMIT_REASON_ACK;
                        ch->xmit_failure_(pktTx_->copy(),
                                          ch->xmit_failure_data_);
                }
                switch (ch->ptype()) {                  // Statistics: frames discarded due to maximum retry limit reached
                        case PT_TCP:
                                ++max_rtx_tcp;
                                break;
                        case PT_ACK:
                                ++max_rtx_ack;
                                break;
                        case PT_MESSAGE:
                                ++max_rtx_message;
                                break;
                        case PT_ARP:
                                ++max_rtx_ARP;
                                break;
                        case PT_VBR:
                        case PT_CBR:
                                ++max_rtx_udp;
                                break;
                        case PT_VOIP:
                                ++max_rtx_voip;
                                break;	
                        case PT_MAC_RTS:
                                ++max_rtx_mac_rts;
                                break;
                        case PT_MAC_CTS:
                                ++max_rtx_mac_cts;
                                break;
                        case PT_MAC_ACK:
                                ++max_rtx_mac_ack;
                                break;
                }

		mac_target_->discard(pktTx_, DROP_MAC_RETRY_COUNT_EXCEEDED);
		pktTx_ = 0;        // The DATA frame is dropped

		*rcount = 0;    // Resetting the corresponding STA Retry Count
		rst_cw();       // Resetting the Contention Window to CWmin

	} else {        // We can still retransmit the DATA frame

		struct hdr_mac802_11 *dh;
		dh = HDR_MAC802_11(pktTx_);
		dh->dh_fc.fc_retry = 1;

                if( (u_int32_t) HDR_CMN(pktTx_)->size() < macmib_->RTSThreshold) {

                        switch (HDR_CMN(pktTx_)->ptype()) {     // Statistics: retransmitted frames
                                case PT_TCP:
                                        ++rtx_tcp;
                                        break;
                                case PT_ACK:
                                        ++rtx_ack;
                                        break;
                                case PT_MESSAGE:
                                        ++rtx_message;
                                        break;
                                case PT_ARP:
                                        ++rtx_ARP;
                                        break;
                                case PT_VBR:
                                case PT_CBR:
                                        ++rtx_udp;
                                        break;
                                case PT_VOIP:
                                        ++rtx_voip;
                                        break;	
                                case PT_MAC_RTS:
                                        ++rtx_mac_rts;
                                        break;
                                case PT_MAC_CTS:
                                        ++rtx_mac_cts;
                                        break;
                                case PT_MAC_ACK:
                                        ++rtx_mac_ack;
                                        break;
                        }

                } else {                       // If the DATA frame size is bigger than RTS_Threshold, the station
                        ++rtx_mac_rts;         // must first retransmit the RTS frame
                }
		sendRTS(ETHER_ADDR(mh->dh_da));

		inc_cw();       // Doubling the Contention Window

		mhBackoff_.start(cw_, is_idle());  // New backoff procedure
	}
}

/*###################################################################
#                                                                   #
# send: invoked by the 'recv' method every time a new frame is      #
#       coming from the Interface Queue                             #
#                                                                   #
#       p: the incoming frame                                       #
#                                                                   #
###################################################################*/

void Mac802_11e::send(Packet *p) {

	double rTime;
	struct hdr_mac802_11* dh = HDR_MAC802_11(p);

	// XXX v2.0Start
	// During CP set AP's data transmission rate according to destination
	if(!(mac_target_->cfp_) &&
	   mac_target_->index_ == mac_target_->HCmacAddress()) {
		mac_target_->dataRate_ = mac_target_->bitRates_[(int)ETHER_ADDR(dh->dh_da)];
	}
	// XXX v2.0End

	// ---------- HCFstart
	// No bursting during CFP, CAP or CFB
	if(!mac_target_->cfp_ &&
	   mac_target_->cap_ == -1 &&
	   mac_target_->cfb_ == -1) {
		if (mac_target_->txop == nq && pktCTRL_ == 0 && mac_target_->tx_active_ != nq) {        // Bursting modality: the node is sending
        	                                                                                        // a frame following the first

      	          	double txt = mac_target_->txtime(HDR_CMN(p)->size(), mac_target_->dataRate_);   // Transmission time

			// XXX v2.0Start
			double dur = mac_target_->txtime(ETHER_ACK_LEN, mac_target_->dataRate_) + mac_target_->sifs_;  // Duration
			// XXX v2.0End

        	        if ( (NOW + mac_target_->sifs_ + txt + dur) - mac_target_->txopLimit_ > mac_target_->TXOPLmt) {    // In bursting modality, we must stop
        	                bursting = 1;                                                           		   // the burst if the duration of the next
        	                ptosend = 0;                                                            		   // frame will exceed the TXOPLimit

        	                ((BS_QueueHandler*)(mac_target_->callback_))->stop_bursting(nq);        // Informs the Interface Queue that the
        	                mac_target_->txop = -1;                                                 // burst has been ended

        	                mhBackoff_.start(cw_, is_idle());       // A normal backoff access is needed for this frame
        	        } else {
        	                mhDefer_.start(mac_target_->sifs_);    // In bursting mode, the station must wait only a SIFS period
        	        }                                              // before transmitting this frame
        	}
	}
	// ---------- HCFend

	sendDATA(p);
	// ---------- HCFstart
	// No RTS during CFP, CAP or CFB
	if(!mac_target_->cfp_ &&
	   mac_target_->cap_ == -1 &&
	   mac_target_->cfb_ == -1)
	// ---------- HCFend
		sendRTS(ETHER_ADDR(dh->dh_da));

        /*
	 * Assign the data packet a sequence number.
	 */

	dh->dh_scontrol = mac_target_->sta_seqno_++;

	// ---------- HCFstart
	// No backoff operations
	// during CFP, CAP or CFB
	if(mac_target_->cfp_ ||
	   mac_target_->cap_ != -1 ||
	   mac_target_->cfb_ != -1)
	   	return;
	// ---------- HCFend
	
	/*
	 *  If the medium is IDLE, we must wait for a DIFS
	 *  Space before transmitting.
	 */

	if(mhBackoff_.busy() == 0 && (mac_target_->txop != nq && pktCTRL_ != 0 ||
                                      mac_target_->txop != nq && mac_target_->tx_active_ != nq ||
                                      pktCTRL_ != 0 && mac_target_->tx_active_ == nq))                 {

        // The virtual queue is not doing a backoff

		if(is_idle() && mac_target_->txop != nq && pktCTRL_ == 0 && mac_target_->tx_active_ != nq) {

			if(mhDefer_.busy() == 0) {
				find_idle = 1;          // The medium is idle, therefore it is possible to try to transmit
                                                        // immediately after a AIFS/EIFS period

                                if (mac_target_->err_frame == 0) {
                                        mhDefer_.start(difs_);                  // Last frame was correctly received  ---> wait AIFS
                                } else {
                                        mhDefer_.start(mac_target_->eifs_);     // Last frame was not correctly received  ---> wait EIFS
                                }
			}

		}

	/*
	 * If the medium is NOT IDLE, then we start
	 * the backoff timer.
	 */

		else {
                        mhBackoff_.start(cw_,0);
		}
	}
}

/*###################################################################
#                                                                   #
# recvRTS: called every time a RTS frame is received, this routine  #
#          checks if it is possible to reply with a CTS frame       #
#                                                                   #
#     p: the RTS frame                                              #
#                                                                   #
###################################################################*/

void Mac802_11e::recvRTS(Packet *p) {
	struct rts_frame *rf = (struct rts_frame*)p->access(hdr_mac::offset_);

        if(tx_state_ != MAC_IDLE) {                             // If the medium is not idle, the RTS is discarded and
                ++busy_mac_cts;                                 // the CTS is not sent
		mac_target_->discard(p, DROP_MAC_BUSY);
		return;
	}

	if(pktCTRL_) {                                          // If I'm responding to someone else, discard this RTS
		mac_target_->discard(p, DROP_MAC_BUSY);         // and don't reply
		return;
	}

	sendCTS(ETHER_ADDR(rf->rf_ta), rf->rf_duration);        // Now it's possible to send the CTS frame

	/*
	 *  Stop deferring - will be reset in tx_resume()
	 */

	if(mhDefer_.busy()) mhDefer_.stop();

	tx_resume();

	mac_target_->mac_log(p);
}

/*###################################################################
#                                                                   #
# recvCTS: called every time a CTS frame is received, this routine  #
#          checks if it is possible to send the DATA frame          #
#                                                                   #
#     p: the CTS frame                                              #
#                                                                   #
###################################################################*/

void Mac802_11e::recvCTS(Packet *p) {

	if(tx_state_ != MAC_RTS) {                                      // If we didn't send a RTS, the CTS is discarded
		mac_target_->discard(p, DROP_MAC_INVALID_STATE);
		return;
	}

	assert(pktRTS_);                        // Our RTS has been "acked" by a CTS, so we can discard it
	Packet::free(pktRTS_); pktRTS_ = 0;

	assert(pktTx_);

	mhSend_.stop();

	/*
	 * The successful reception of this CTS packet implies
	 * that our RTS was successful.  Hence, we can reset
	 * the Short Retry Count and the CW.
	 */

	ssrc_ = 0;
	rst_cw();

	tx_resume();

	mac_target_->mac_log(p);
}

/*###################################################################
#                                                                   #
# recvDATA: called every time a DATA frame is received, this method #
#           checks if it is possible to send the ACK frame, then it #
#           passes the DATA frame to the LLC layer                  #
#                                                                   #
#     p: the DATA frame                                             #
#                                                                   #
###################################################################*/

void Mac802_11e::recvDATA(Packet *p) {
	struct hdr_mac802_11 *dh = HDR_MAC802_11(p);
	u_int32_t dst, src, size;

        struct hdr_cmn *ch = HDR_CMN(p);

        dst = ETHER_ADDR(dh->dh_da);
        src = ETHER_ADDR(dh->dh_sa);
        size = ch->size();

        /*
        * Adjust the MAC packet size - ie; strip
        * off the mac header
        */

        ch->size() -= ETHER_HDR_LEN11;
        ch->num_forwards() += 1;

	/*
	 *  If we sent a CTS, clean up...
	 */

	if(dst != MAC_BROADCAST) {
		if(size >= macmib_->RTSThreshold && ch->burst_no_rts == 0) {
			if (tx_state_ == MAC_CTS) {
				assert(pktCTRL_);
				Packet::free(pktCTRL_); pktCTRL_ = 0;
				mhSend_.stop();

				/*
				 * Our CTS got through.
				 */

			} else {                                                // If the size of the received DATA frame is smaller
                                ++busy_mac_ack;                                 // than the RTS_Threshold (or in bursting modality it
				mac_target_->discard(p, DROP_MAC_BUSY);         // is a frame following the first) and we didn't send
				return;                                         // a CTS, discard this frame without sanding an ACK
			}

			sendACK(src);           // Send the ACK frame
			tx_resume();
		}

		/*
		 *  We did not send a CTS and there's no
		 *  room to buffer an ACK.
		 */

		else {
			if(pktCTRL_ || tx_state_ != MAC_IDLE) {          // The MAC is busy, discard the frame
                                ++busy_mac_ack;
				mac_target_->discard(p, DROP_MAC_BUSY);
				return;
			}
			sendACK(src);   // Send the ACK frame

			if(mhSend_.busy() == 0) {
			        tx_resume();
                        }
		}
	}

        mac_target_->upsend(p,nq);      // Pass the DATA frame to the LLC layer
}

/*###################################################################
#                                                                   #
# recvACK: called every time an ACK frame is received, this method  #
#          is responsable for resuming the transmission             #
#                                                                   #
#     p: the ACK frame                                              #
#                                                                   #
###################################################################*/

void Mac802_11e::recvACK(Packet *p) {

	struct hdr_cmn *ch = HDR_CMN(p);

	if((tx_state_ != MAC_SEND) || (!mac_target_->cfp_ && edcaTX_status==false)) {// If we didn't send a DATA frame, the ACK is discarded
	        mac_target_->discard(p, DROP_MAC_INVALID_STATE);
	        return;
	}
	
	// printf("Time %.12f : 802_11(%d) : ricevuto ACK, di durata %.12f\n",Scheduler::instance().clock(),mac_target_->index_,mac_target_->txtime(p));
	
        /*
	 * The successful reception of this ACK packet implies
	 * that our DATA transmission was successful.  Hence,
	 * we can reset the Short/Long Retry Count and the CW.
	 */

	if( (u_int32_t)HDR_CMN(pktTx_)->size() <= macmib_->RTSThreshold )
		ssrc_ = 0;
	else
		slrc_ = 0;

	rst_cw();

	assert(pktTx_);
	Packet::free(pktTx_);
	pktTx_ = 0;       // The transmitted DATA frame can be discarded

	mhSend_.stop();

	assert(mhBackoff_.busy() == 0);

        if (ptosend == 0) {
	        mhBackoff_.start(cw_, is_idle());       // We start a new backoff before resuming transmission

        } else {        // Bursting modality

                mac_target_->txop = nq;              // This queue maintains the TXOP

                mac_target_->need_pkt(nq);           // We need the next frame from the Interface Queue for continuing the burst
                set_tx_state(MAC_IDLE);
        }

        if (ptosend == 0) {
	        tx_resume();        
        }
	mac_target_->mac_log(p);
}

// ---------- HCFstart
/*###################################################################
#                                                                   #
# stopTC: called every time a contention-free access procedure      #
#         takes the place of the EDCF mechanism (i.e. at the        #
#         beginning of a CAP or CFP), this method stops the timers  #
#         and frees the control frames of the TC, so that the       #
#         contention period state doesn't disrupt the               #
#         HCF mechanism                                             #
#                                                                   #
###################################################################*/

void Mac802_11e::stopTC() {

	// Stop or pause the
	// Contention Period timers
	if(mhDefer_.busy())
		mhDefer_.stop();
	if(mhBackoff_.busy() && !mhBackoff_.paused()) {
		mhBackoff_.pause();
		backoffOn_ = 1;
	}
	// Free CTS/ACK packets
	if(pktCTRL_) {
		Packet::free(pktCTRL_);
		pktCTRL_ = 0;
	}
	// Free RTS packets
	if(pktRTS_) {
		Packet::free(pktRTS_);
		pktRTS_ = 0;
	}

	// Reset TC's internal state
	ssrc_ = slrc_ = 0;
	virtual_coll = 0;
	find_idle = 0;
	ptosend = 0;

}

/*###################################################################
#                                                                   #
# recvCC: called every time an Contention Control frame is received #
#         this method checks if this TC has traffic to send and     #
#         prepares the relative Reservation Request frame, then     #
#         starts the CAP timer with a random time in the CCI        #
#                                                                   #
#         p : the Contention Control frame                          #
#                                                                   #
###################################################################*/

void Mac802_11e::recvCC(Packet* p) {

	struct cc_frame *cc = (struct cc_frame*)p->access(hdr_mac::offset_);
	u_int32_t src = (u_int32_t)ETHER_ADDR(cc->cc_bssid);

	double dccop = mac_target_->txtime(ETHER_RR_LEN, mac_target_->basicRate_)
	               + mac_target_->sifs_;
	int random = Random::random() % (cc->cc_cci_length);

	if(pktTx_ /*&& mac_target_->queueLength(nq) > 1*/) {
		mac_target_->cap_ = nq;
		capSendRR(src);
		// Define random waiting
		// time within the CCI
		// (802.11e - 9.10.4.2)
		double wait = mac_target_->pifs_ + (random * dccop);
		mac_target_->mhCap_.start(wait);
	}

}

/*###################################################################
#                                                                   #
# recvCFACK: called every time a QoS CF-ACK frame is received,      #
#            this method frees the packet that was transmitted      #
#            and stops its send timer, then prepare the response    #
#            packet and restarts the CFB timer                      #
#                                                                   #
#            p: the CF-Ack frame                                    #
#                                                                   #
###################################################################*/

void Mac802_11e::recvCFACK(Packet* p) {

	// Outside CFB, discard this packet
	if(mac_target_->cfb_ != nq) {
		mac_target_->discard(p, "---");
		tx_resume();
		return;
	}

	// If no data was sent discard CF-Ack
	// and resume CFB transmissions
	if(tx_state_ != MAC_QOS_DATA  &&
	   tx_state_ != MAC_QOS_DATA_CFACK) {
	        mac_target_->discard(p, DROP_MAC_INVALID_STATE);
		if(!(mac_target_->mhCfb_.busy())) {
			// Stop the send timer
			if(mhSend_.busy())
				mhSend_.stop();
			set_tx_state(MAC_IDLE);
			mac_target_->mhCfb_.start(mac_target_->sifs_);
		}
	        return;
	}

	// Free pktTx_ and stop the send timer
	Packet::free(pktTx_);
	pktTx_ = 0;
	mhSend_.stop();

	// Ask for a new packet to send
	// and reset tx_state to MAC_IDLE
	tx_resume();

	// QSTA must send a Null packet
	// if it has no more packets to send,
	// a QoS Data packet otherwise
	struct hdr_mac802_11 *dh = HDR_MAC802_11(p);
	u_int32_t src = (u_int32_t)ETHER_ADDR(dh->dh_sa);

	if(!pktTx_)
		cfbSendNULL(src);
	else
		cfbSendDATA(src);

	// Restart the CFB timer
	mac_target_->mhCfb_.start(mac_target_->sifs_);

	// Free the received frame
	mac_target_->mac_log(p);

}

/*###################################################################
#                                                                   #
# recvCFPOLL: called every time a QoS CF-Poll frame is received,    #
#             this method prepares the response Data or Null        #
#             packet and starts up the CFB behaviour                #
#                                                                   #
#             p: the CF-Poll frame                                  #
#                                                                   #
###################################################################*/

void Mac802_11e::recvCFPOLL(Packet* p) {

	// If a CFB for another TC is already
	// going on, discard this packet
	if((mac_target_->cfb_ != -1) && (mac_target_->cfb_ != nq)) {
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
			fprintf(macLog, "[%d.%d] CF-Poll discarded\n");
		}
#endif //=========================================================================
		mac_target_->discard(p, "---");
		tx_resume();
		return;
	}

	struct hdr_mac802_11 *dh = HDR_MAC802_11(p);
	struct qos_cfpoll_frame* pf = (struct qos_cfpoll_frame*)p->access(hdr_mac::offset_);
	u_int32_t src = (u_int32_t)ETHER_ADDR(dh->dh_sa);

	// Free pktCTRL_
	if(pktCTRL_) {
		Packet::free(pktCTRL_);
		pktCTRL_ = 0;
	}

	// Set up the node for the Polled TXOP
	// setting the NAV to the end of the TXOP
	// and silencing all the Traffic Categories.
	// Note: in CFP this has already been done
	// at Beacon reception, but we provide
	// further protection this way
	mac_target_->set_nav(dh->dh_duration);
	for(int i=0; i<(mac_target_->num_tc); i++)
		mac_target_->tc[i]->stopTC();

	// Due to errors on the channel, the QSTA
	// may be trying to answer to previous
	// frames and the Send timer could be active
	if(mhSend_.busy())
		mhSend_.stop();
  
	// Setup CFB
	mac_target_->cfb_ = nq;
	mac_target_->txopLimit_ = NOW - mac_target_->txtime(p);
	mac_target_->txopDuration_ = pf->pf_txop_duration * 1e-6;
	
	// Answer to the CF-Poll
	if(pktTx_)
		cfbSendDATA(src);
	else {
		cfb_noDataToSend++;
		cfbSendNULL(src);
	}

	// Start CFB
	mac_target_->mhCfb_.start(mac_target_->sifs_);

	tx_resume();

	// Free the received frame
	mac_target_->mac_log(p);

}

/*###################################################################
#                                                                   #
# recvDATA_CFPOLL: called every time a QoS Data+CF-Poll frame       #
#                  is received, this method delivers the data to    #
#                  upper layer, prepares the response Data+CF-Ack   #
#                  or CF-Ack packet and starts up the CFB behaviour #
#                                                                   #
#                  p: the Data+CF-Poll frame                        #
#                                                                   #
###################################################################*/

void Mac802_11e::recvDATA_CFPOLL(Packet* p) {

	// If a CFB for this or another TC is
	// already going on, discard this packet
	if((mac_target_->cfb_ != -1) && (mac_target_->cfb_ != nq)) {
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
			fprintf(macLog, "[%d.%d] Data+CF-Poll discarded\n");
		}
#endif //=========================================================================
		mac_target_->discard(p, "---");
		tx_resume();
		return;
	}
	
	struct hdr_cmn *ch = HDR_CMN(p);
	struct hdr_mac802_11 *dh = HDR_MAC802_11(p);
	
	u_int8_t dataTc = ch->mac_tc_aux;
	u_int32_t src = (u_int32_t)ETHER_ADDR(dh->dh_sa);
	u_int32_t size = ch->size();

	// XXX v2.0Start
	// Update CFB statistics
	// NOTE: we don't specify the actual rate
	//       (fast or slow) of the station while
	//       updating the statistics, since it
	//       will be clear in the finish file
	switch(ch->ptype()) {
		case PT_TCP:
			cfb_data_fast_recv++;
			break;
		case PT_VOIP:
			cfb_voice_fast_recv++;
			break;
	}
	// XXX v2.0End

	// Adjust the MAC packet size
	// i.e. strip off the mac header
	ch->size() -= ETHER_HDR_LEN11;
	ch->num_forwards() += 1;
	
	// Free pktCTRL_
	if(pktCTRL_) {
		Packet::free(pktCTRL_);
		pktCTRL_ = 0;
	}

	// Set up the node for the Polled TXOP
	// setting the NAV to the end of the TXOP
	// and silencing all the Traffic Categories.
	// Note: in CFP this has already been done
	// at Beacon reception, but we provide
	// further protection this way
	mac_target_->set_nav(dh->dh_duration);
	for(int i=0; i<(mac_target_->num_tc); i++)
		mac_target_->tc[i]->stopTC();

	// Due to errors on the channel, the QSTA
	// may be trying to answer to previous
	// frames and the Send timer could be active
	if(mhSend_.busy())
		mhSend_.stop();

	// Setup CFB
	mac_target_->cfb_ = nq;
	mac_target_->txopLimit_ = NOW - mac_target_->txtime(p);
	mac_target_->txopDuration_ = (dh->dh_duration * 1e-6) - DSSS_SlotTime;

	// Answer to the Data+CF-Poll
	if(pktTx_)
		cfbSendDATA_CFACK(src, dataTc);
	else {
		cfb_noDataToSend++;
		cfbSendCFACK(src, dataTc);
	}

	// Start CFB
	mac_target_->mhCfb_.start(mac_target_->sifs_);

	// Pass the Data frame to the LLC layer, as if
	// it came from the dataTC Traffic Category
	// NOTE: it must be done here, or it will
	//       disrupt the MAC header fields
	mac_target_->upsend(p, dataTc);
	
	tx_resume();

}

/*###################################################################
#                                                                   #
# recvDATA_CFACK: called every time a QoS Data+CF-Ack frame         #
#                 is received, this method delivers the data to     #
#                 upper layer, and prepares the response            #
#                 Data+CF-Ack or CF-Ack packet                      #
#                                                                   #
#                 p: the Data+CF-Ack frame                          #
#                                                                   #
###################################################################*/

void Mac802_11e::recvDATA_CFACK(Packet* p) {

	// Outside CFB, discard this packet
	if(mac_target_->cfb_ != nq) {
		mac_target_->discard(p, "---");
		tx_resume();
		return;
	}

	// Handle CF-Ack reception
	
	// Free pktTx_ and stop the send timer
	Packet::free(pktTx_);
	pktTx_ = 0;
	mhSend_.stop();

	// Ask for a new packet to send
	// and reset tx_state to MAC_IDLE
	tx_resume();

	// Handler Data reception

	struct hdr_cmn *ch = HDR_CMN(p);
	struct hdr_mac802_11 *dh = HDR_MAC802_11(p);
	
	u_int8_t dataTc = ch->mac_tc_aux;
	u_int32_t src = (u_int32_t)ETHER_ADDR(dh->dh_sa);
	u_int32_t size = ch->size();

	// XXX v2.0Start
	// Update CFB statistics
	// NOTE: we don't specify the actual rate
	//       (fast or slow) of the station while
	//       updating the statistics, since it
	//       will be clear in the finish file
	switch(ch->ptype()) {
		case PT_TCP:
			cfb_data_fast_recv++;
			break;
		case PT_VOIP:
			cfb_voice_fast_recv++;
			break;
	}
	// XXX v2.0End

	// Adjust the MAC packet size
	// i.e. strip off the mac header
	ch->size() -= ETHER_HDR_LEN11;
	ch->num_forwards() += 1;
		
	// Answer to the Data+CF-Ack
	if(pktTx_)
		cfbSendDATA_CFACK(src, dataTc);
	else
		cfbSendCFACK(src, dataTc);

	// Pass the Data frame to the LLC layer
	mac_target_->upsend(p, dataTc);

	// Restart the CFB timer
	mac_target_->mhCfb_.start(mac_target_->sifs_);

}

/*###################################################################
#                                                                   #
# recvDATA_CFACK_CFPOLL: called every time a                        #
#                        QoS Data+CF-Ack+CF-Poll frame is received, #
#                        this method delivers the data to upper     #
#                        layer, prepares the response Data+CF-Ack   #
#                        or CF-Ack packet, and sets the Polled TXOP #
#                        according to the new QoS CF-Poll duration  #
#                                                                   #
#                        p: the Data+CF-Ack+CF-Poll frame           #
#                                                                   #
###################################################################*/

void Mac802_11e::recvDATA_CFACK_CFPOLL(Packet* p) {

	// Outside CFB, discard this packet
	if(mac_target_->cfb_ != nq) {
		mac_target_->discard(p, "---");
		tx_resume();
		return;
	}

	// Update the Polled TXOP duration
	struct hdr_mac802_11 *dh = HDR_MAC802_11(p);

	double dur = (dh->dh_duration) * 1e-6;
	double txt = NOW - (mac_target_->txopLimit_ + mac_target_->txopDuration_ - dur);
	mac_target_->txopLimit_ += txt;

	// Handler Data+CF-Ack reception
	recvDATA_CFACK(p);

}

/*###################################################################
#                                                                   #
# resumeTC: called every time a contention-free access period       #
#           terminates (i.e. at the end of a CAP of CFP), this      #
#           method is responsable for resetting the TC's state,     #
#           by freeing the control frames, restarting the backoff   #
#           timer and resuming the EDCF behavior                    #
#                                                                   #
###################################################################*/

void Mac802_11e::resumeTC() {

	// Free CF-ACK/Null packet
	if(pktCTRL_) {
		Packet::free(pktCTRL_);
		pktCTRL_ = 0;
	}
	// Resume the backoff
	if(backoffOn_) {
		if(is_idle() && mhBackoff_.paused())
			mhBackoff_.resume(mac_target_->pifs_);
		backoffOn_ = 0;
	}
	// Resume contention
	// period transmission
	tx_resume();

}

/*###################################################################
#                                                                   #
# edcaTX:   called in order to enable/disable the transmission      #
#           of data for a given AC during EDCA channel access       #
#                                                                   #
# edcaTXenabled: returns true/false according to the fact that      #
#           data transmission during EDCA channel access is         #
#           enabled/disabled                                        #
#           NOTE: this method is called before each call to         #
#           resumeTC(). If edcaTX is disabled resumeTC() is not     #
#           invoked.                                                #
###################################################################*/

void Mac802_11e::edcaTX(bool enabled) {
	edcaTX_status = enabled;
}
bool Mac802_11e::edcaTXenabled() {
	return edcaTX_status;
}


static class Mac802_11HCClass : public TclClass {
public:
	Mac802_11HCClass() : TclClass("Mac/802_11HC") {}
	TclObject* create(int, const char*const*) {
		return (new Mac802_11HC());
	}
} class_mac802_11HC;

Mac802_11HC::Mac802_11HC() : Mac802_11(), mhTbtt_(this), mhCapRate_(this) {
	
#ifdef DEBUGACTIVE //=============================================================
	setbuf(macLog, 0);
#endif //=========================================================================

	capRate_ = 0;
	// Minimum number of CCOPs
	// (802.11e - 9.10.4.4)
	nccop_ = 2;	
	for(int i=0; i<MAX_RR_POLL; i++) {
		rrPollList_[i].addr = 0;
		rrPollList_[i].tc = 0;
		rrPollList_[i].txop_duration = 0.0;
	}
	rrPollNum_ = 0;
	rrPollIndex_ = 0;
	

	// XXX v2.0Start
	cfpRate_ = 0.0;
	cfpDuration_ = 0.0;
	// XXX v2.0End
	for(int i=0; i<MAX_CF_POLL; i++) {
      cfPollList_[i].mac = NULL;
		cfPollList_[i].addr = 0;
		cfPollList_[i].tc = -1;
	}
   cfPollNum_ = 0;
   cfPollIndex_ = 0;
   cfStartTime_ = 0.0;
   cap_time = 0;
   cap_num = 0;
   cfp_time = 0;
   cfp_num = 0;

   // XXX v2.0Start
   nSTAs_ = 0;
   // XXX v2.0End

   // XXX v3.0Start
   scheduler = SCHED_DRAFT; // the default sched is very similar to that provided by 802.11e DRAFT
   second_schedule = false;
   // XXX v3.0End

}

int Mac802_11HC::command(int argc, const char*const* argv) {
	if(argc == 2) {
		// Start the Point Coordination Function mechanism
		if(strcmp(argv[1], "pcf-start") == 0) {
			mhTbtt_.start(cfpRate_);
			return TCL_OK;
		}
		// Stop the Point Coordination Function mechanism
		else if(strcmp(argv[1], "pcf-stop") == 0) {
			if(mhTbtt_.busy())
				mhTbtt_.stop();
			return TCL_OK;
		}
		// Start the Controlled Access mechanism
		else if(strcmp(argv[1], "cap-start") == 0) {
			if(capRate_ == 0) {
				fprintf(stderr, "\nERROR : Cannot start CAP if capRate is not defined\n\n");
				exit(1); 
			}
			mhCapRate_.start(capRate_);
			return TCL_OK;
		}
		// Stop the Controlled Access mechanism
		else if(strcmp(argv[1], "cap-stop") == 0) {
			if(mhCapRate_.busy())
				mhCapRate_.stop();
			return TCL_OK;
		}
      // Compute TXOPLMTs and CFPrate
      else if(strcmp(argv[1], "schedule") == 0) {
         schedule(true); // the boolean tells that this is the first schedule
         return TCL_OK;
      }
   }
   else if(argc == 3) {
      if(strcmp(argv[1], "use-scheduler") == 0) {
         if (strcmp(argv[2], "draft") == 0)
            scheduler = SCHED_DRAFT;
         else if (strcmp(argv[2], "enhanced") == 0)
            scheduler = SCHED_ENHANCED;
         else if (strcmp(argv[2], "mmfair") == 0)
            scheduler = SCHED_MMFAIR;
         else {
            fprintf(stderr, "ERROR: no HCF scheduler available named \"%s\"... exiting!\n", argv[2]);
            return TCL_ERROR;
         }
         return TCL_OK;
      }
   }
   else if(argc == 4) {
		// XXX v2.0End
		// Define the time interval between two
		// Controlled Access Periods (CAPs) and
		// the number of Controlled Contention
		// Opportunities (CCOPs) during each
		// Controlled Contention Interval (CCI)
		if(strcmp(argv[1], "set-cap") == 0) {
			capRate_ = atof(argv[2]);
			nccop_ = atoi(argv[3]);
			return TCL_OK;
		}
      // Insert requesting node's address
      // and TC into the CFP Polling list
      if(strcmp(argv[1], "cf-poll") == 0) {
         cfPollList_[cfPollNum_].mac = NULL;
         cfPollList_[cfPollNum_].addr = atoi(argv[2]);
         cfPollList_[cfPollNum_].tc = atoi(argv[3]);
         // Update the number of
         // CF-Pollable STA/TC elements
         cfPollNum_++;
         return TCL_OK;
      }
      // XXX v2.0Start
      // Store data transmission rate 
      // of each station in the network
      if (strcmp(argv[1], "STA_bitrate") == 0) {
         int addr = atoi(argv[2]);
         bitRates_[addr] = atof(argv[3]);
         printf("HC %d bitrate: %f\n", addr, bitRates_[addr]);
         // Update the number of
         // stations in the network
         nSTAs_++;
         return TCL_OK;
      }
      // XXX v2.0End
   }
   else if (argc == 5) {
      // Insert requesting node's address
      // and TC into the CFP Polling list
      if(strcmp(argv[1], "cf-poll") == 0) {
         Tcl& tcl = Tcl::instance();
         cfPollList_[cfPollNum_].mac = (Mac802_11*)tcl.lookup(argv[2]);
         cfPollList_[cfPollNum_].addr = atoi(argv[3]);
         cfPollList_[cfPollNum_].tc = atoi(argv[4]);
         // Update the number of
         // CF-Pollable STA/TC elements
         cfPollNum_++;
         return TCL_OK;
      }
      // XXX v3.0Start
      // Sets TSPEC parameters
      else if (strcmp(argv[1], "TSPEC") == 0) {
         int mac_addr = atoi(argv[2]);
         if (mac_addr < 0 || mac_addr >= MAX_STA)
            return TCL_ERROR;
         if (strcmp(argv[3], "NominalMSDUSize") == 0) {
            tspecs_[mac_addr].NominalMSDUSize = atoi(argv[4]);
            return TCL_OK;            
         }
         else if (strcmp(argv[3], "MaximumMSDUSize") == 0) {
            tspecs_[mac_addr].MaximumMSDUSize = atoi(argv[4]);
            return TCL_OK;            
         }
         else if (strcmp(argv[3], "MaximumServiceInterval") == 0) {
            tspecs_[mac_addr].MaximumServiceInterval = atof(argv[4]);
            return TCL_OK;            
         }
         else if (strcmp(argv[3], "DelayBound") == 0) {
            tspecs_[mac_addr].DelayBound = atof(argv[4]);
            return TCL_OK;            
         }
         else if (strcmp(argv[3], "MinimumDataRate") == 0) {
            tspecs_[mac_addr].MinimumDataRate = atoi(argv[4]);
            return TCL_OK;            
         }
         else if (strcmp(argv[3], "MeanDataRate") == 0) {
            tspecs_[mac_addr].MeanDataRate = atoi(argv[4]);
            return TCL_OK;            
         }
         else if (strcmp(argv[3], "PeakDataRate") == 0) {
            tspecs_[mac_addr].PeakDataRate = atoi(argv[4]);
            return TCL_OK;            
         }
         return TCL_ERROR;
      }
      // XXX v3.0End

   }
	return Mac802_11::command(argc, argv);
}

// XXX v3.0Start
/*###################################################################
#                                                                   #
# schedule: schedule per-station TXOPLIMIT and the CFPrate.         #
#     2 schedulers are implemented:                                 #
#        1- reference scheduler proposed in IEEE802.11e draft       #
#        2- an enhanced scheduler that reschedules at each BI       #
#           the TXOPLMTs according to the data each                 #
#           station still has in queue                              #
#                                                                   #
###################################################################*/

void Mac802_11HC::schedule(bool first_time) {

   // If it's the first schedule then get the minimum MaximumServiceInterval
   // Note: shared by DRAFT and ENHANCED schedulers
   if (first_time)
   {
      cfpRate_ = 10e10;  // HACK: one MaximuServiceInterval less than this value shall exist 
      for (int i = 0; i < cfPollNum_; ++i)
      {
         int addr = cfPollList_[i].addr;
         if (tspecs_[addr].MaximumServiceInterval < cfpRate_)
            cfpRate_ = tspecs_[addr].MaximumServiceInterval;
      }
   }

   // 1 - Draft (static) scheduler
   if (scheduler == SCHED_DRAFT)
   {
      for (int i = 0; i < cfPollNum_; ++i)
      {
         int addr = cfPollList_[i].addr;    // station address (used as index later on)
         double QoSPoll_time = txtime(ETHER_QOS_CFPOLL_LEN2, basicRate_);
         double ACK_time = txtime(ETHER_ACK_LEN2, bitRates_[addr]);
         double Null_time = txtime(ETHER_NULL_LEN2, bitRates_[addr]);

         double dataframe_time = dataTxTime(tspecs_[addr].NominalMSDUSize, bitRates_[addr]);

         // x is the number of MPDUs generated during a Service Interval
         printf("%f         %d\n", (tspecs_[addr].MeanDataRate * cfpRate_ / 8) / tspecs_[addr].NominalMSDUSize, (int)ceil((tspecs_[addr].MeanDataRate * cfpRate_ / 8) / tspecs_[addr].NominalMSDUSize));
         int x = (int)ceil((tspecs_[addr].MeanDataRate * cfpRate_ / 8) /
                            tspecs_[addr].NominalMSDUSize);

         //if ((float)x -((tspecs_[addr].MeanDataRate * cfpRate_ / 8)/tspecs_[addr].NominalMSDUSize) > 0.9)
         //   --x;

         tspecs_[addr].CFP_TXOPLmt = QoSPoll_time + sifs_ +
                                     x * (dataframe_time + 2*sifs_ + ACK_time);

         // Comuputing the minimum guaranteed TXOP, equal to the time
         // needed for transmitting one MPDU of maximum size
         double max_dataframe_time = dataTxTime(tspecs_[addr].MaximumMSDUSize, bitRates_[addr]);
         tspecs_[addr].MinimumTXOPGuaranteed = QoSPoll_time + sifs_ +
                                               max_dataframe_time + 2*sifs_ + ACK_time;

         // Check if the TXOP is less than the minimum guaranteed
         if (tspecs_[addr].CFP_TXOPLmt < tspecs_[addr].MinimumTXOPGuaranteed)
         {
            fprintf(stderr, "WARNING: allocating MinimumTXOPGuaranteed as TXOP!\n");
            tspecs_[addr].CFP_TXOPLmt = tspecs_[addr].MinimumTXOPGuaranteed;
         }

         printf("MAC %d -> %d frames per-POLL -> TXOP = %f\n", addr, x, tspecs_[addr].CFP_TXOPLmt);
      }
   }
   // 2 - First schedule of Enhanced scheduler or MaxMinFair scheduler
   else if ((scheduler == SCHED_ENHANCED  || scheduler == SCHED_MMFAIR) && first_time)
      // if it's the first_schedule then compute the minimum TXOP guarateed values
      // and assign the initial TXOP values
         for (int i = 0; i < cfPollNum_; ++i)
         {
            int addr = cfPollList_[i].addr;    // station address (used as index later on)
            double QoSPoll_time = txtime(ETHER_QOS_CFPOLL_LEN2, basicRate_);
            double ACK_time = txtime(ETHER_ACK_LEN2, bitRates_[addr]);
            double Null_time = txtime(ETHER_NULL_LEN2, bitRates_[addr]);

            double min_TXOP1, min_TXOP2;

            // 1 - Computing the minimum TXOP duration
            int x = (int)ceil((tspecs_[addr].MinimumDataRate * cfpRate_ / 8) /
                               tspecs_[addr].NominalMSDUSize);
            double dataframe_time = dataTxTime(tspecs_[addr].NominalMSDUSize, bitRates_[addr]);
            double max_dataframe_time = dataTxTime(tspecs_[addr].MaximumMSDUSize, bitRates_[addr]);

            min_TXOP1 = QoSPoll_time + sifs_ + x * (dataframe_time + 2*sifs_ + ACK_time);
            min_TXOP2 = QoSPoll_time + sifs_ + max_dataframe_time + 2*sifs_ + ACK_time;

            // Choose the maximum value as guaranteed TXOP limit
            if (min_TXOP2 > min_TXOP1)
               tspecs_[addr].MinimumTXOPGuaranteed = min_TXOP2;
            else
               tspecs_[addr].MinimumTXOPGuaranteed = min_TXOP1;

            // 2 - Computing the maximum TXOP duration
            x = (int)ceil((tspecs_[addr].PeakDataRate * cfpRate_ / 8) /
                           tspecs_[addr].NominalMSDUSize);
            tspecs_[addr].MaximumTXOPGuaranteed = QoSPoll_time + x * (dataframe_time + 2*sifs_ + ACK_time);

            fprintf(stderr, "MAX_TXOP: %f    MIN_TXOP: %f\n", tspecs_[addr].MaximumTXOPGuaranteed, tspecs_[addr].MinimumTXOPGuaranteed);

            // Assign TXOP limit using MeanDataRate and NominalMSDUSize
            x = (int)ceil((tspecs_[addr].MeanDataRate * cfpRate_ / 8) /
                           tspecs_[addr].NominalMSDUSize);
            tspecs_[addr].MeanDataRate_TXOP = tspecs_[addr].CFP_TXOPLmt = QoSPoll_time +
                                        x * (dataframe_time + 2*sifs_ + ACK_time);

            if (tspecs_[addr].CFP_TXOPLmt < tspecs_[addr].MinimumTXOPGuaranteed)
               tspecs_[addr].CFP_TXOPLmt = tspecs_[addr].MinimumTXOPGuaranteed;
         }
   else if (scheduler == SCHED_ENHANCED  && !first_time)
      // adjust each TXOP according to RemainingData and TimeNotUsed values
      for (int i = 0; i < cfPollNum_; ++i)
      {
         int addr = cfPollList_[i].addr;    // station address (used as index later on)
         double QoSPoll_time = txtime(ETHER_QOS_CFPOLL_LEN2, basicRate_);
         double ACK_time = txtime(ETHER_ACK_LEN2, bitRates_[addr]);
         double Null_time = txtime(ETHER_NULL_LEN2, bitRates_[addr]);

         // Eventually add to TXOP the time needed to transmit the remaining data...
         if (tspecs_[addr].RemainingData > 0)
         {
            double dataframe_time = dataTxTime(tspecs_[addr].NominalMSDUSize, bitRates_[addr]);
            double time_to_add = ceil(tspecs_[addr].RemainingData/tspecs_[addr].NominalMSDUSize) * (dataframe_time+2*sifs_+ACK_time);
            tspecs_[addr].CFP_TXOPLmt += time_to_add;
            tspecs_[addr].RemainingData = 0;
            tspecs_[addr].TimeNotUsed = 0.0;
         }
         // ... and eventually remove the time not used at last poll
         if (tspecs_[addr].TimeNotUsed > 0.0)
         {
            tspecs_[addr].CFP_TXOPLmt -= tspecs_[addr].TimeNotUsed;
            tspecs_[addr].TimeNotUsed = -1.0;
         }
         // ... as usual the TXOP must be at least the guaranteed value
         // and not more than request by PeakDataRate
         if (tspecs_[addr].CFP_TXOPLmt < tspecs_[addr].MinimumTXOPGuaranteed)
            tspecs_[addr].CFP_TXOPLmt = tspecs_[addr].MinimumTXOPGuaranteed;
         else if (tspecs_[addr].CFP_TXOPLmt > tspecs_[addr].MaximumTXOPGuaranteed)
            tspecs_[addr].CFP_TXOPLmt = tspecs_[addr].MaximumTXOPGuaranteed;
      }
   else if (scheduler == SCHED_MMFAIR  && !first_time)
   {
      // allocate TXOPs according to a MaxMinFair algorithm
      double CFEND_needed = txtime(ETHER_CFEND_LEN2, basicRate_)
                          + DSSS_MaxPropagationDelay
                          + sifs_
                          + DSSS_SlotTime;
      double remaining = mhTbtt_.startTime() + cfpDuration_ - NOW - CFEND_needed;
      int tot_buffer = 0;
      
      // Allocate to all stations the TXOP corresponding to the MeanDataRate, compute the sum
      // of all buffer sizes and the time left to schedule
      for (int i = 0; i < cfPollNum_; ++i)
      {
         int addr = cfPollList_[i].addr;    // station address (used as index later on)

         tspecs_[addr].CFP_TXOPLmt = tspecs_[addr].MeanDataRate_TXOP;
         remaining -= tspecs_[addr].MeanDataRate_TXOP;
         if (tspecs_[addr].RemainingData >= tspecs_[addr].NominalMSDUSize)
            tot_buffer += tspecs_[addr].RemainingData;
      }
      if (remaining > 0)
      {
         // Allocate the remaining time using a MaxMinFair algorithm
         for (int i = 0; i < cfPollNum_; ++i)
         {
            if (tot_buffer <= 0)
               break;
            int addr = cfPollList_[i].addr;    // station address (used as index later on)
            int num_MPDUs;

            if (tspecs_[addr].RemainingData < tspecs_[addr].NominalMSDUSize)
               continue;

            double ACK_time = txtime(ETHER_ACK_LEN2, bitRates_[addr]);
            double dataframe_time = dataTxTime(tspecs_[addr].NominalMSDUSize, bitRates_[addr]) + sifs_ + ACK_time;

            num_MPDUs = (int)(((float)tspecs_[addr].RemainingData/tot_buffer)*(remaining/dataframe_time));
            if (num_MPDUs > 0)
               tspecs_[addr].CFP_TXOPLmt += num_MPDUs * (sifs_ + dataframe_time);
            //fprintf(stderr, "MAC addr %d  --- %d # %d # %f # %f\n", addr, tspecs_[addr].RemainingData, tot_buffer, remaining, dataframe_time);
            //fprintf(stderr, "         %f  --- %f --- %d\n", (float)tspecs_[addr].RemainingData/tot_buffer, (float)remaining/dataframe_time, num_MPDUs);
         }
      }
   }
   else
   {
      fprintf(stderr, "ERROR: trying to use an unknown or not implemented scheduler!\n");
   }

   // comment this line...
   cfpDuration_ = cfpRate_;
   fprintf(stderr, "cfpRate %f   cfpDuration %f\n", cfpRate_, cfpDuration_);
   /*// ... and uncomment the following lines if you
     // want the cfpDuration to be equal to the sum of all TXOP
     // Note: the duration is anyway such a value since, once every
     // station has been polled, the QAP sends a CF-END
   if (first_time || scheduler == SCHED_ENHANCED)
   {
      cfpDuration_ = 0.0;
      for (int i = 0; i < cfPollNum_; ++i)
      {
         int addr = cfPollList_[i].addr;
         cfpDuration_ += tspecs_[addr].CFP_TXOPLmt;
      }
   }*/

   // NOTE: when you call this method the cfpDuration_
   // must already be set.
   normalizeTXOPs();
}

/*###################################################################
#                                                                   #
# normalizeTXOPs: called after the TXOPs scheduling in order to     #
#       normalize their values to the duration of the CF-Period     #
#                                                                   #
###################################################################*/
void Mac802_11HC::normalizeTXOPs()
{
   double CFEND_needed = txtime(ETHER_CFEND_LEN2, basicRate_) + DSSS_MaxPropagationDelay + sifs_ + DSSS_SlotTime;
   double allocated_time = 0.0;
   double allocable_time = cfpDuration_ - CFEND_needed;

   for (int i = 0; i < cfPollNum_; ++i)
      allocated_time += tspecs_[cfPollList_[i].addr].CFP_TXOPLmt;


   //fprintf(stderr, "alloc: %f    remaining: %f\n", allocated_time, cfp_time);
   if (allocated_time < allocable_time)
      return;

   //fprintf(stderr, "-------------------------------------------\n");
   //fprintf(stderr, "Allocable: %f       Allocated: %f\n", allocable_time, allocated_time);
   //fflush(stderr);
   if (scheduler == SCHED_ENHANCED)
         // Using this type of scheduler the normalization is accomplished
         // without accounting the minimum TXOP guaranteed
         for (int i = 0; i < cfPollNum_; ++i)
         {
            tspecs_[cfPollList_[i].addr].CFP_TXOPLmt -= tspecs_[cfPollList_[i].addr].MinimumTXOPGuaranteed;
            allocable_time -= tspecs_[cfPollList_[i].addr].MinimumTXOPGuaranteed;
            allocated_time -= tspecs_[cfPollList_[i].addr].MinimumTXOPGuaranteed;
         }

   if (allocable_time < 0.0)
   {
      // (Only with SCHED_ENHANCED!) This means that the cfp_duration is neither enough
      // for allocate the min TXOP guaranteed
      for (int i = 0; i < cfPollNum_; ++i)
         tspecs_[cfPollList_[i].addr].CFP_TXOPLmt = (cfpDuration_ - CFEND_needed)/cfPollNum_;
      return;
   }
   //fprintf(stderr, "Allocable: %f       Allocated: %f\n", allocable_time, allocated_time);

   //double sum = 0.0;
   for (int i = 0; i < cfPollNum_; ++i)
   {
       //fprintf(stderr, "STA_MAC: %d   TXOP_before: %f   ", cfPollList_[i].addr, tspecs_[cfPollList_[i].addr].CFP_TXOPLmt + tspecs_[cfPollList_[i].addr].MinimumTXOPGuaranteed);
       //fprintf(stderr, "STA_MAC: %d   TXOP_before: %f   ", cfPollList_[i].addr, tspecs_[cfPollList_[i].addr].CFP_TXOPLmt);
       //fflush(stderr);
       tspecs_[cfPollList_[i].addr].CFP_TXOPLmt *= allocable_time/allocated_time;
       if (scheduler == SCHED_ENHANCED)
          tspecs_[cfPollList_[i].addr].CFP_TXOPLmt += tspecs_[cfPollList_[i].addr].MinimumTXOPGuaranteed;
       //fprintf(stderr, "TXOP_after: %f\n", tspecs_[cfPollList_[i].addr].CFP_TXOPLmt);
       //sum += tspecs_[cfPollList_[i].addr].CFP_TXOPLmt;
   }
   //fprintf(stderr, "Allocable: %f       Allocated: %f\n", cfpDuration_ - CFEND_needed, sum);
}

/*###################################################################
#                                                                   #
# dataTxTime: given a size and a tx_bitrate returns the time        #
#       needed for transmitting it (including the headers of        #
#       several protocols - RTP, UDP, IP, ETHER)                    #
#                                                                   #
#   payload:   data size in bytes                                   #
#   bitrate:   tx bitrate (bit/sec)                                 #
#                                                                   #
###################################################################*/

double Mac802_11HC::dataTxTime(int payload, double bitrate)
{
   int dataframe_size = payload +                           // data size
                        12 +                                // RTP header
                        8 +                                 // UDP header
                        IP_HDR_LEN +                        // IP header
                        ETHER_HDR_LEN11_2 + 2;              // MAC header (+ PLCP)
   return txtime(dataframe_size, bitrate);
}


/*###################################################################
#                                                                   #
# getRemainingData: gets from a given stations the size of          #
#     the data (belonging to a given TC) not sent yet               #
#                                                                   #
###################################################################*/

void Mac802_11HC::getRemainingData(int poll_index, int tc, u_int32_t addr)
{
   if ((cfPollList_[poll_index].tc != tc) || (cfPollList_[poll_index].addr != addr))
      return;
   if (cfPollList_[poll_index].mac == NULL)
   {
      fprintf(stderr, "WARNING: no MAC object available while HC is retrieving queue size!\n");
      return;
   }
   int queue_length = cfPollList_[poll_index].mac->queueByteLength(tc);  // measured in bytes
   //int queue_length = cfPollList_[poll_index].mac->queueLength(tc);     // measured in MSDUs
   //   fprintf(stderr, "getRemainingData... address: %d    len: %d \n", addr, queue_length);
   tspecs_[addr].RemainingData = queue_length;
}

// XXX v3.0End

/*###################################################################
#                                                                   #
# bursting: evaluates the $opt(wtburst) value from Tcl script       #
#           in order to understand if the bursting is active.       # 
#           Called by tbttHandler() and cfpStartHandler()           #
#                                                                   #
###################################################################*/

int Mac802_11HC::bursting() {
	Tcl& tcl = Tcl::instance();
	tcl.evalc("set opt(wtburst)");
	char* burst = tcl.result();
	return(atoi(burst));
}

/*###################################################################
#                                                                   #
# pcf: evaluates the $opt(pcf_start) value from Tcl script          #
#      in order to understand if the PCF mechanism is active        #
#      and at which time it starts. Called by capRateHandler()      #
#                                                                   #
###################################################################*/

double Mac802_11HC::pcf() {
	Tcl& tcl = Tcl::instance();
	tcl.evalc("set opt(pcf_start)");
	char* pcf = tcl.result();
	return(atof(pcf));
}

/*###################################################################
#                                                                   #
# simDuration: evaluates the $opt(start) and $opt(stop) values      #
#              from Tcl script and returns the duration of the      #
#              network simulation. Called by dump_tc()              #
#                                                                   #
###################################################################*/

double Mac802_11HC::simDuration() {
	Tcl& tcl = Tcl::instance();
	tcl.evalc("set opt(start)");
	double start = atof(tcl.result());
	tcl.evalc("set opt(stop)");
	double stop = atof(tcl.result());
	return(stop - start);
}

/*###################################################################
#                                                                   #
# set_queue: creates the EDCF Access Categories objects             #
#            for Hybrid Coordinator                                 # 
#                                                                   #
#            n: the number of virtual queues                        #
#                                                                   #
###################################################################*/

void Mac802_11HC::set_queue(int n) {
        for (int i=0; i<n; i++) {
                tc[i] = new Mac802_11eHC(this,&PMIB, &MMIB);
        }
        num_tc = n;
}

/*###################################################################
#                                                                   #
# dump_tc: dumps some variables associated to each virtual queue    #
#          and calls Tcl for statistics                             #
#                                                                   #
###################################################################*/

void Mac802_11HC::dump_tc() {
	// This calculation rounds down
	// the simulation duration to an
	// integer number of CFP+CP time
	// intervals, in order to evaluate
	// the exact percentage of time used
	// by the CAP/CFP access procedures
	double dur = simDuration();
	int n = (int)floor(dur / cfpRate_);
	double cfp_dur = n * cfpRate_;
	// XXX v2.0Start
	double cfp_occ = cfp_time / (n * cfpDuration_);
	// XXX v2.0End
	double cap_avg;
	if(cap_num > 0)
		cap_avg = cap_time / cap_num;
	else
		cap_avg = 0;
	double cfp_avg;
	if(cfp_num > 0)
		cfp_avg = cfp_time / cfp_num;
	else
		cfp_avg = 0;
	Tcl::instance().evalf("catch \"802_11eHC_tc %f %f %d %f %f %f %d\"",
		(double)(cap_time / dur), (double)cap_avg, (int)cap_num,
		(double)(cfp_occ), (double)(cfp_time / cfp_dur), (double)cfp_avg,
		(int)cfp_num);

	Mac802_11::dump_tc();
}
			
/*###################################################################
#                                                                   #
# is_idle: returns true if the channel is idle,                     #
#          false otherwise. Used by tbttHandler()                   #
#                                                                   #
###################################################################*/

int Mac802_11HC::is_idle() {

	if(rx_state_ != MAC_IDLE)
		return 0;
	for(int i=0; i<num_tc; i++) {
		if(tc[i]->tx_state_ != MAC_IDLE)
			return 0;
	}
	if(nav_ > NOW)
		return 0;
	return 1;

}

/*###################################################################
#                                                                   #
# navHandler: called by mac-timers.cc when the NAV timer has        #
#             expired. Control on foreshortened CFP start added     #
#                                                                   #
###################################################################*/

void Mac802_11HC::navHandler() {

        // Pending CFP or CAP [case NAV].
	// The NAV has just expired: frame exchange terminated.
	// If we were waiting to start the CFP or CAP, it can start now
	if(is_idle()) {
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
			if(cfpWaiting_ || capWaiting_) {
	   			fprintf(macLog, "[%d] NAV expired, channel idle %f\n", index_, NOW);
			}
		}
#endif //=========================================================================
		// wait PIFS
		if(cfpWaiting_) {
			mhCfpStart_.start(pifs_);
		}
		else if(capWaiting_)
			mhCapStart_.start(pifs_);
	}

	else Mac802_11::navHandler();

}

/*###################################################################
#                                                                   #
# recvHandler: called by mac-timers.cc when the last bit of a frame #
#              has been received                                    #
#                                                                   #
###################################################################*/

void Mac802_11HC::recvHandler() {
	recv_timer();
}

/*###################################################################
#                                                                   #
# capRateHandler: called by mac-timers.cc when the Controlled       #
#                 Accees Period (CAP) must be started, it starts    #
#                 the CAP in PIFS if the channel is idle, begins    #
#                 the CAP start waiting otherwise                   #
#                                                                   #
###################################################################*/

void Mac802_11HC::capRateHandler() {

#ifdef DEBUGACTIVE //=============================================================
	if(NOW > DEBUGTIME) {
		fprintf(macLog, "CAP at %f (capRate_ : %f)", NOW, capRate_);
	}
#endif //=========================================================================

	// Immediately restart the CAP rate timer
	mhCapRate_.start(capRate_);

	// Reset the CAP poll index
	rrPollIndex_ = 0;
	
	// Skip the CAP:
	// - during CFP or CFB
	// - if the previous CAP
	//   is still going on
	// - if waiting to start a CFB
	// - if already waiting to start
	//   the previous CAP
	// - if there's not enough time
	//   for a CAP before next TBTT
	// - if already waiting PIFS
	//   before starting a CAP or CFP
	double remaining = 0;
	double needed = 0;
	if(pcf() != 0) {
		double dccop = txtime(ETHER_RR_LEN2, basicRate_) + sifs_;
		// Needed time: CCI length + (max CFB duration * nccop_)
		needed = pifs_
		         + txtime(ETHER_CC_LEN2, basicRate_)
			 + DSSS_MaxPropagationDelay
		         + pifs_
		         + (nccop_ * dccop)
		         + (nccop_ * TXOPLmt);
		if(mhTbtt_.busy())
			remaining = mhTbtt_.startTime() + cfpRate_ - NOW;
		else
			remaining = pcf() - NOW;
	}
	if(cfp_ || (cfb_ != -1) || (cap_ != -1) ||
	   cfpWaiting_ || capWaiting_ ||
	   (remaining < needed) || mhCapStart_.busy() || mhCfpStart_.busy()) {
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
			fprintf(macLog, " - CAP skipped:\nCFP(%d) CFB(%d) CFPwait(%d) rem(%f) ned(%f)\n", cfp_, cfb_, cfpWaiting_, remaining, needed);
		}
#endif //=========================================================================
		return;
	}

	// If channel is idle then start the CAP after PIFS
	if(is_idle()) {
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
			fprintf(macLog, " - channel is idle\n");
		}
#endif //=========================================================================
		mhCapStart_.start(pifs_);
	}

	// If channel is busy we need to wait for it to become idle
	else {
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
			fprintf(macLog, " - channel is busy\n");
		}
#endif //=========================================================================
		capWaiting_ = 1;

		// Controls will then be done where channel can become idle: 
		// see 'navHandler()', 'send_timer()', 'recv_timer()', 'recvACK()'
	}
}

/*###################################################################
#                                                                   #
# capStartHandler: called by mac-timers.cc when the CAP must        #
#                  start and PIFS have been waited and the channel  #
#                  is idle, it sets up the HC for the CAP itself    #
#                  by setting the CAP flag to 1 and sending the     #
#                  Controlled Contention (CC) frame                 #
#                                                                   #
###################################################################*/

void Mac802_11HC::capStartHandler() {

	// If the medium is idle a PIFS
	// after CAP timer expired
	if(is_idle()) {
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
			fprintf(macLog, "CAP start at %f\n", NOW);
		}
#endif //=========================================================================

		// Store CAP start time for HCF statistics
		cfStartTime_ = NOW;

		if(capWaiting_)
			capWaiting_ = 0;

		cap_ = 1;
		sendCC();
		
		// Start the CAP timer, it will
		// expire after the CCI has ended
		double dccop = txtime(ETHER_RR_LEN2, basicRate_) + sifs_;
		double wait = txtime(ETHER_CC_LEN2, basicRate_)
		              + DSSS_MaxPropagationDelay
		              + pifs_
		              + (nccop_ * dccop);
		mhCap_.start(wait);
	}
	
	else {
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
			fprintf(macLog, "channel busy at %f, CAP delayed again\n", NOW);
		}
#endif //=========================================================================
		// This can happen when the channel has been sensed
		// idle, but then it has become busy within PIFS:

		// Case I
		// We come from a pending CAP [NAV, ERROR, ACKSENT, ACKRECV].
		// It happens when a burst is going on,
		// just wait for the burst to end

		// Case II
		// We come from the CAP timer expiration.
		// We have to start the waiting period
		capWaiting_ = 1;

		// Controls will then be done where channel can become idle.
		// See 'navHandler()', 'send_timer()', 'recv_timer()', 'recvACK()'
	}

}

/*###################################################################
#                                                                   #
# capHandler: called by mac-timers.cc during the Controlled         #
#             Access Period (CAP), it defines the HC behaviour      #
#             when the CAP mechanism is active. If there are QSTAs  #
#             to poll, it prepares a CF-Poll or Data+CF-Poll frame, #
#             sets the CFB flag to the right TC and invokes the     #
#             cfbHandler() method. If there are no stations left    #
#             in the CAP polling list, it ends the CAP              #
#                                                                   #
###################################################################*/

void Mac802_11HC::capHandler() {

#ifdef DEBUGACTIVE //=============================================================
	if(NOW > DEBUGTIME) {
		if(rrPollNum_ != 0 && rrPollIndex_ == 0) {
			fprintf(macLog, "[%d] RR Polling List (rrPollNum_: %d, rrPollIndex_: %d)\n", index_, rrPollNum_, rrPollIndex_);
			for(int i=0; i<rrPollNum_; i++) {
				fprintf(macLog, "    -%d- addr: %d, tc: %d, txop: %f\n", i, rrPollList_[i].addr, rrPollList_[i].tc, rrPollList_[i].txop_duration);
			}
		}
	}
#endif //=========================================================================

	// If there are no more QSTAs to poll reset the CAP flag,
	// 'awake' all the TCs using the resumeTC() method and
	// return to normal EDCF behaviour
	if(rrPollNum_ == 0) {
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
			fprintf(macLog, "[%d] No QSTAs to poll, CAP terminated %f\n", index_, NOW);
		}
#endif //=========================================================================
		// Update HCF statistics
		cap_time += (NOW - cfStartTime_ - pifs_);
		cap_num++;

		cap_ = -1;
		// Reset the NAV
		if(mhNav_.busy()) {
			mhNav_.stop();
		}
		nav_ = NOW;
		for(int i=0; i<num_tc; i++) {
         if (tc[i]->edcaTXenabled()) {
            tc[i]->set_tx_state(MAC_IDLE);
            tc[i]->resumeTC();
         }
		}
		return;
	}
	
	// If there are QSTAs to poll
	if(rrPollIndex_ < rrPollNum_) {

		int ac = rrPollList_[rrPollIndex_].tc;
		u_int32_t dst = rrPollList_[rrPollIndex_].addr;
		txopDuration_ = rrPollList_[rrPollIndex_].txop_duration;
		txopLimit_ = NOW;
			
		// Prepare the packet
		int dataTc = dataToSend(dst, ac);
		if(dataTc != num_tc)
			tc[ac]->cfbSendDATA_CFPOLL(dst, dataTc);
		else
			tc[ac]->cfbSendCFPOLL(dst);

		// Set the CFB Traffic Category
		cfb_ = ac;

		// Send the packet
		cfbHandler();

	}

	// If that was the last QSTA in the
	// Reservation Request Polling List
	// clear the CAP internal state
	if(rrPollIndex_ == (rrPollNum_-1)) {

		for(int i=0; i<rrPollNum_; i++) {
			rrPollList_[i].addr = 0;
			rrPollList_[i].tc = 0;
			rrPollList_[i].txop_duration = 0.0;
		}
		rrPollNum_ = 0;

	}

}

/*###################################################################
#                                                                   #
# tbttHandler: called by mac-timers.cc when the Target Beacon       #
#              Transmission Time (TBTT) comes. A Contention         #
#              Free Period (CFP) is started in PIFS if the channel  #
#              is idle, a waiting period is begun otherwise         #
#                                                                   #
###################################################################*/

void Mac802_11HC::tbttHandler() {

#ifdef DEBUGACTIVE //=============================================================
	if(NOW > DEBUGTIME) {
		fprintf(macLog, "TBTT at %f (cfpRate_ : %f)", NOW, cfpRate_);
	}
#endif //=========================================================================

	// Immediately restart the TBTT timer
	mhTbtt_.start(cfpRate_);

   cfpDurRemaining_ = 0;
   cfPollIndex_ = 0;
   
	// If the channel is idle then
	// start the CFP after PIFS
	if(is_idle()) {
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
			fprintf(macLog, " - channel is idle\n");
		}
#endif //=========================================================================
		mhCfpStart_.start(pifs_);
	}

	// If the channel is busy we need to wait for it to become idle and
	// then start a foreshortened CFP, so we keep track of the delay
	else {
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
			fprintf(macLog, " - channel is busy\n");
		}
#endif //=========================================================================
		cfpWaiting_ = 1;
		// Max time we have to wait
		double max;
		if(!bursting()) {
			// XXX v2.0Start
			max = txtime(ETHER_RTS_LEN2, dataRate_)
			      + DSSS_MaxPropagationDelay
			      + sifs_
			      + txtime(ETHER_CTS_LEN2, dataRate_)
			      + DSSS_MaxPropagationDelay
			      + sifs_
			      + txtime(MAC_MaxMPDUSize, dataRate_)
			      + DSSS_MaxPropagationDelay
			      + sifs_
			      + txtime(ETHER_ACK_LEN2, dataRate_)
			      + DSSS_MaxPropagationDelay;
			// XXX v2.0End
		}
		else {
			max = TXOPLmt + 0.001;		// 5ms margin;
		}
		mhCfp_.start(max);

		// Controls will then be done where channel can become idle: 
		// see 'navHandler()', 'send_timer()', 'recv_timer()', 'recvACK()'
	}

}

/*###################################################################
#                                                                   #
# cfpStartHandler: called by mac-timers.cc when the CFP must        #
#                  start, PIFS have alredy been waited and the      #
#                  channel is idle, it sets up the HC for the CFP   #
#                  by defining the CFP duration, setting the CFP    #
#                  flag and sending the Beacon frame                #
#                                                                   #
###################################################################*/

void Mac802_11HC::cfpStartHandler() {

	// If the medium is idle
	// a PIFS after TBTT
	if(is_idle()) {
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
			fprintf(macLog, "CFP start at %f", NOW);
		}
#endif //=========================================================================

		// Store CFP start time for HCF statistics
		cfStartTime_ = NOW;
		
		cfp_ = 1;
		
		// If we were waiting, now we
		// must start a foreshortened CFP
		// XXX v2.0Start
		if(cfpWaiting_) {
			cfpWaiting_ = 0;
			cfpDurRemaining_ += cfpDuration_ - mhCfp_.passed() + pifs_;
			mhCfp_.stop();
		}

		// If channel was idle at TBTT,
		// now start a normal CFP
		else {
			cfpDurRemaining_ = cfpDuration_;
		}
		// XXX V2.0End

#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
			fprintf(macLog, " (duration %f)\n", cfpDurRemaining_);
		}
#endif //=========================================================================
		sendBCN();
	}
	
	else {
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
			fprintf(macLog, "channel busy at %f, CFP delayed again\n", NOW);
		}
#endif //=========================================================================
		// This can happen when the channel has been sensed
		// idle, but then it has become busy within PIFS:

		// Case I
		// We come from a pending CFP [NAV, ERROR, ACKSENT, ACKRECV].
		// It happens when a burst is going on, mhCfp_ timer is already
		// running, we just wait for the burst to end

		// Case II
		// We come from the TBTT.
		// We have to start the waiting period and the mhCfp_ timer
		// to keep record of the time that has passed
		if(!mhCfp_.busy()) {
			cfpWaiting_ = 1;
			// Max time we have to wait
			double max;
			if(!bursting()) {
				// XXX v2.0Start
				max = txtime(ETHER_RTS_LEN2, dataRate_)
				      + DSSS_MaxPropagationDelay
				      + sifs_
				      + txtime(ETHER_CTS_LEN2, dataRate_)
				      + DSSS_MaxPropagationDelay
				      + sifs_
				      + txtime(MAC_MaxMPDUSize, dataRate_)
				      + DSSS_MaxPropagationDelay
				      + sifs_
				      + txtime(ETHER_ACK_LEN2, dataRate_)
				      + DSSS_MaxPropagationDelay;
				// XXX v2.0End
			}
			else {
				max = TXOPLmt + 0.005;		// 5ms margin
			}
			mhCfp_.start(max);

			// In this case the timer is started when PIFS have
			// already passed, so we don't have to count these
			// when we calculate the shortened length of the CFP
			cfpDurRemaining_ = 0 - pifs_;
		}

		// Controls will then be done where channel can become idle.
		// See 'navHandler()', 'send_timer()', 'recv_timer()', 'recvACK()'
	}

}

/*###################################################################
#                                                                   #
# cfpHandler: called by mac-timers.cc during the Contention         #
#             Free Period (CFP), it defines the HC behaviour        #
#             when the PCF mechanism is active (i.e. in CFP).       #
#             If there are QSTAs to poll, it prepares a CF-Poll     #
#             or Data+CF-Poll frame, sets the CFB flag to the       #
#             right TC and invokes the cfbHandler() method.         #
#             If there are no stations left in the CFP polling      #
#             list, it ends the CFP by sending a CF-End frame       #
#                                                                   #
###################################################################*/

void Mac802_11HC::cfpHandler() {

	// if channel is busy longer than max time, skip CFP
	if(cfpWaiting_) {
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
			fprintf(macLog, "channel busy longer than maximum time\n");
		}
#endif //=========================================================================
		cfpWaiting_ = 0;
		return;
	}

// XXX v3.0 --- Start
   // if there are QSTAs to poll
   if(cfPollIndex_ < cfPollNum_) {
      //fprintf(stderr, "cfNumPolled_: %d   cfPollNum_:   %d   cfPollIndex_: %d \n", cfNumPolled_, cfPollNum_, cfPollIndex_);
      
      double remaining = mhTbtt_.startTime() + cfpDuration_ - NOW;
      int da = cfPollList_[cfPollIndex_].addr;
      dataRate_ = bitRates_[da];
      double txop_needed = txtime(ETHER_QOS_CFPOLL_LEN2, dataRate_)
                      + DSSS_MaxPropagationDelay
                      + sifs_
                      + tspecs_[da].CFP_TXOPLmt
                      + DSSS_SlotTime;
      double basic_needed = txtime(ETHER_QOS_CFPOLL_LEN2, dataRate_)
                      + DSSS_MaxPropagationDelay
                      + sifs_
                      + DSSS_SlotTime;

#ifdef DEBUGACTIVE //=============================================================
	  if(NOW > DEBUGTIME) {
	    fprintf(macLog, "remaining: %f, txop_needed: %f (dataRate_: %f | TXOPLmt: %f)\n", remaining, txop_needed, dataRate_, tspecs_[da].CFP_TXOPLmt);
	  }
#endif //=========================================================================

      // if there's enough time to send a CF-Poll
      //fprintf(stderr, "TBTT start %f, cfpDuration: %f,   NOW: %f\n", mhTbtt_.startTime(), cfpDuration_, NOW);
      //fprintf(stderr, "Remaining: %f    Neeeded: %f\n", remaining, needed);
      double CFEND_needed = txtime(ETHER_CFEND_LEN2, basicRate_)
                          + DSSS_MaxPropagationDelay
                          + sifs_
                          + DSSS_SlotTime;

      if(remaining > txop_needed) {
         int ac = cfPollList_[cfPollIndex_].tc;
         u_int32_t dst = cfPollList_[cfPollIndex_].addr;
         txopLimit_ = NOW;
         txopDuration_ = tspecs_[dst].CFP_TXOPLmt;

         // This prevent ns2 scheduler errors!!!
         if ((remaining-txop_needed) < CFEND_needed)
            txopDuration_ -= CFEND_needed - (remaining-txop_needed);

         // Prepare the packet
         int dataTc = dataToSend(dst, ac);
         if(dataTc != num_tc)
            tc[ac]->cfbSendDATA_CFPOLL(dst, dataTc);
         else
            tc[ac]->cfbSendCFPOLL(dst);
         // Set the CFB Traffic Category
         cfb_ = ac;
         // Send the packet
         cfbHandler();
      }
      // if there is not abough time for the next txop,
      // if possible, allocate that time anyway!
      else if ((remaining - CFEND_needed) > basic_needed)
      {
         int ac = cfPollList_[cfPollIndex_].tc;
         u_int32_t dst = cfPollList_[cfPollIndex_].addr;
         txopLimit_ = NOW;
         // This allows to allocate a TXOP shorter than the one
         // 
         txopDuration_ = (remaining - CFEND_needed);

         // Prepare the packet
         int dataTc = dataToSend(dst, ac);
         if(dataTc != num_tc)
            tc[ac]->cfbSendDATA_CFPOLL(dst, dataTc);
         else
            tc[ac]->cfbSendCFPOLL(dst);

         // Set the CFB Traffic Category
         cfb_ = ac;

         // Send the packet
         cfbHandler();
      }
      else 
         // not enough time, then end the current cfp
         sendCFEND();
   }
   // In this case all stations have already been polled. The remaining time is
   // allocated for the stations having the busiest buffers.
   else if (scheduler == SCHED_MMFAIR)
   {
      double basic_needed = txtime(ETHER_QOS_CFPOLL_LEN2, dataRate_)
                           + DSSS_MaxPropagationDelay
                           + sifs_
                           + DSSS_SlotTime;
      double CFEND_needed = txtime(ETHER_CFEND_LEN2, basicRate_)
                           + DSSS_MaxPropagationDelay
                           + sifs_
                           + DSSS_SlotTime;
      double remaining = mhTbtt_.startTime() + cfpDuration_ - NOW - CFEND_needed;

      if (remaining > basic_needed)
      {
         second_schedule = true; // It might be already true!

         int addr_to_poll = -1;
         int index_to_poll = -1;
         int max1_buff_size = 0;  // the size of the busiest buffer
         int max2_buff_size = 0;  // the size of the second most filled buffer

         // Figuring out which station has the greatest buffered data size
         for (int i = 0; i < cfPollNum_; ++i)
         {
            int addr = cfPollList_[i].addr;
            if (tspecs_[addr].RemainingData > max1_buff_size)
            {
               max2_buff_size = max1_buff_size;
               addr_to_poll = addr;
               index_to_poll = i;
               max1_buff_size = tspecs_[addr].RemainingData;
            }
         }
         if (addr_to_poll == -1)
            sendCFEND();
         else
         {
            double ACK_time = txtime(ETHER_ACK_LEN2, bitRates_[addr_to_poll]);
            double maxdataframe_time = dataTxTime(tspecs_[addr_to_poll].MaximumMSDUSize, bitRates_[addr_to_poll]);
            double nominaldataframe_time = dataTxTime(tspecs_[addr_to_poll].NominalMSDUSize, bitRates_[addr_to_poll]);
            double maxMPDU_needed = txtime(ETHER_QOS_CFPOLL_LEN2, dataRate_)
                                 + DSSS_MaxPropagationDelay
                                 + 2*sifs_
                                 + maxdataframe_time
                                 + ACK_time
                                 + DSSS_SlotTime;

            int x = (int)ceil(  ((float)(max1_buff_size-max2_buff_size)) / tspecs_[addr_to_poll].NominalMSDUSize);
            txopDuration_ = txtime(ETHER_QOS_CFPOLL_LEN2, basicRate_) + x * (nominaldataframe_time + 2*sifs_ + ACK_time);  

            //fprintf(stderr, "%d   ----   %d\n",max1_buff_size, max2_buff_size );

            // at least a max-size MPDU can be sent
            if (maxMPDU_needed > txopDuration_)
               txopDuration_ = maxMPDU_needed;


            cfPollIndex_ = index_to_poll;
            txopLimit_ = NOW;
            // if the remaining time is less than the txop, then assign only the remaining
            if (remaining < txopDuration_)
               txopDuration_ = remaining;

            int ac = cfPollList_[index_to_poll].tc;
            // Prepare the packet
            int dataTc = dataToSend(addr_to_poll, ac);
            if(dataTc != num_tc)
               tc[ac]->cfbSendDATA_CFPOLL(addr_to_poll, dataTc);
            else
               tc[ac]->cfbSendCFPOLL(addr_to_poll);

            //fprintf(stderr, "%d ", index_to_poll);

            // Set the CFB Traffic Category
            cfb_ = ac;

            // Send the packet
            cfbHandler();
         }
      }
      else sendCFEND();
   }
// HCF v3.0 --- End
   // no more QSTAs to poll
   else sendCFEND();
}

/*###################################################################
#                                                                   #
# cfbHandler: handles HC's frames exchange during the CFB.          #
#             Called before and after every frame is sent during    #
#             the CFB, it transmits the frame and then checks for   #
#             the expected answer: if no response is received,      #
#             it retransmits the frame.                             #
#             It also controls if there's enough time left          #
#             in the CFB to send the frame and receive the answer   #
#                                                                   #
###################################################################*/

void Mac802_11HC::cfbHandler() {
	double timeout, wait;

	// Sent CF-Poll, Data+CF-Poll, CF-Ack.
	// Waiting for data from QSTA,
	// check if we're receiving it
	// and wait for end of reception
	if(!cfbFailed_ &&
	   tc[cfb_]->tx_state_ != MAC_IDLE) {
	   	if(rx_state_ != MAC_IDLE) {
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) {
				fprintf(macLog, "[%d] receiving expected CF-Ack/Data/Null %f, waiting for end of reception\n", index_, NOW);
			}
#endif //=========================================================================
			return;
		}
		else {
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) {
				fprintf(macLog, "[%d] not receiving expected CF-Ack/Data/Null %f\n    Attempting retransmission ", index_, NOW);
			}
#endif //=========================================================================

			// If Data was sent, set the retry bit
			// and update the statistics
			if(tc[cfb_]->tx_state_ == MAC_QOS_DATA_CFPOLL ||
			   tc[cfb_]->tx_state_ == MAC_QOS_DATA_CFACK ||
			   tc[cfb_]->tx_state_ == MAC_QOS_DATA_CFACK_CFPOLL) {
				HDR_MAC802_11(tc[cfb_]->pktTx_)->dh_fc.fc_retry = 1;

				switch (HDR_CMN(tc[cfb_]->pktTx_)->ptype()) {     // Statistics: retransmitted frames
					case PT_TCP:
						++(tc[cfb_]->rtx_tcp);
						break;
					case PT_ACK:
						++(tc[cfb_]->rtx_ack);
						break;
					case PT_MESSAGE:
						++(tc[cfb_]->rtx_message);
						break;
					case PT_ARP:
						++(tc[cfb_]->rtx_ARP);
						break;
               case PT_VBR:
					case PT_CBR:
						++(tc[cfb_]->rtx_udp);
						break;
					case PT_VOIP:
						++(tc[cfb_]->rtx_voip);
						break;	
					case PT_MAC_RTS:
						++(tc[cfb_]->rtx_mac_rts);
						break;
					case PT_MAC_CTS:
						++(tc[cfb_]->rtx_mac_cts);
						break;
					case PT_MAC_ACK:
						++(tc[cfb_]->rtx_mac_ack);
						break;
				}
			}

			// If a CF-Poll or Data+CF-Poll frame
			// was sent, update its Duration/ID
			// field to the remaining TXOP duration
			if(tc[cfb_]->tx_state_ == MAC_QOS_CFPOLL) {
				struct qos_cfpoll_frame* pf = (struct qos_cfpoll_frame*)(tc[cfb_]->pktPOLL_)->access(hdr_mac::offset_);
				
				pf->pf_txop_duration = usec(txopLimit_
				                            + txopDuration_
				                            - NOW);
				pf->pf_duration = usec(txopLimit_
				                       + txopDuration_
				                       - NOW
				                       + DSSS_SlotTime);
			}
			if(tc[cfb_]->tx_state_ == MAC_QOS_DATA_CFPOLL) {
				struct hdr_mac802_11* dh = HDR_MAC802_11(tc[cfb_]->pktTx_);

				dh->dh_duration = usec(txopLimit_
				                       + txopDuration_
				                       - NOW
				                       + DSSS_SlotTime);
			}
			// Stop the send timer
			if(tc[cfb_]->mhSend_.busy())
				tc[cfb_]->mhSend_.stop();
			// Then wait another PIFS-SIFS so that
			// transmission is resumed PIFS after
			// the the end of the last transmission
			// (802.11e - 9.10.1.2)
			cfbFailed_ = 1;
			mhCfb_.start(pifs_-sifs_);
			return;
		}
	}

	// Just waited PIFS after failure of
	// expected response, resume transmission
	if(cfbFailed_) {
		cfbFailed_ = 0;
	}
	
	// Check if there's enough time left in the
	// Polled TXOP to send another packet.
	double remaining = txopLimit_ + txopDuration_ - NOW;
	double needed = 0;
	struct hdr_mac802_11* dh;
	u_int32_t dst;
	if(tc[cfb_]->pktCTRL_) {
		dh = HDR_MAC802_11(tc[cfb_]->pktCTRL_);
		dst = (u_int32_t)ETHER_ADDR(dh->dh_da);
		// XXX v2.0Start
		needed = txtime(tc[cfb_]->pktCTRL_)
		         + DSSS_MaxPropagationDelay
			 + sifs_
			 + txtime(MAC_MinMPDUSize, dataRate_)
			 + DSSS_MaxPropagationDelay
			 + sifs_
		         + txtime(ETHER_QOS_CFPOLL_LEN2, dataRate_);
		// XXX v2.0End
	}
	if(tc[cfb_]->pktPOLL_ && !(tc[cfb_]->pktCTRL_)) {
		dh = HDR_MAC802_11(tc[cfb_]->pktPOLL_);
		if(dh->dh_duration != 0) {
			dst = (u_int32_t)ETHER_ADDR(dh->dh_da);
			// XXX v2.0Start
			needed = txtime(tc[cfb_]->pktPOLL_)
			         + DSSS_MaxPropagationDelay
				 + sifs_
				 + txtime(MAC_MinMPDUSize, dataRate_)
				 + DSSS_MaxPropagationDelay
				 + sifs_
			         + txtime(ETHER_QOS_CFPOLL_LEN2, dataRate_);
			// XXX v2.0End
		}
	}
	if(tc[cfb_]->pktTx_ && !(tc[cfb_]->pktPOLL_) && !(tc[cfb_]->pktCTRL_)) {
		dh = HDR_MAC802_11(tc[cfb_]->pktTx_);
		dst = (u_int32_t)ETHER_ADDR(dh->dh_da);
		// XXX v2.0Start
		needed = txtime(tc[cfb_]->pktTx_)
		         + DSSS_MaxPropagationDelay
			 + sifs_
			 + txtime(MAC_MinMPDUSize, dataRate_)
			 + DSSS_MaxPropagationDelay
			 + sifs_
		         + txtime(ETHER_QOS_CFPOLL_LEN2, dataRate_);
		// XXX v2.0End
	}
	// Not enough time: reset pktCTRL_ and pktPOLL_,
	// reset pktTx_ subtype to MAC_Subtype_Data and
	// send the QoS CF-Poll that ends the Polled TXOP
#ifdef DEBUGACTIVE //=============================================================
	if(NOW > DEBUGTIME) {
		fprintf(macLog, "[rem: %f, ned: %f] ", remaining, needed);
	}
#endif //=========================================================================
	if(remaining < needed) {
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
			fprintf(macLog, "...not enough time!\n");
		}
#endif //=========================================================================
		if(tc[cfb_]->pktCTRL_){
			Packet::free(tc[cfb_]->pktCTRL_);
			tc[cfb_]->pktCTRL_ = 0;
		}
		if(tc[cfb_]->pktPOLL_) {
			Packet::free(tc[cfb_]->pktPOLL_);
			tc[cfb_]->pktPOLL_ = 0;
		}
		if(tc[cfb_]->pktTx_) {
			dh = HDR_MAC802_11(tc[cfb_]->pktTx_);
			dh->dh_fc.fc_subtype = MAC_Subtype_Data;
			hdr_cmn* ch = HDR_CMN(tc[cfb_]->pktTx_);
			if(ch->mac_tc_aux != num_tc &&
			   ch->mac_tc_aux != cfb_) {
				ch->mac_tc = ch->mac_tc_aux;
				ch->mac_tc_aux = (u_int8_t)num_tc;
				if(tc[cfb_]->pktBackup_) {
					tc[cfb_]->pktTx_ = tc[cfb_]->pktBackup_;
					tc[cfb_]->pktBackup_ = 0;
				}
				else
					tc[cfb_]->pktTx_ = 0;
			}
			// XXX v2.0Start
			dh->dh_duration = usec(txtime(ETHER_ACK_LEN2, dataRate_)
			                       + sifs_);
			// XXX v2.0End
		}			
		tc[cfb_]->cfbSendCFPOLL(dst);
	}

	// Send QoS CF-Ack
	if(tc[cfb_]->pktCTRL_) {
		struct hdr_mac802_11* dh = HDR_MAC802_11(tc[cfb_]->pktCTRL_);
		switch(dh->dh_fc.fc_subtype) {
		case MAC_Subtype_QoS_CFAck:
			tc[cfb_]->set_tx_state(MAC_QOS_CFACK);
			timeout = txtime(tc[cfb_]->pktCTRL_);
			wait = txtime(tc[cfb_]->pktCTRL_)
			       + sifs_
			       + DSSS_MaxPropagationDelay;
			txop = cfb_;
			transmit(tc[cfb_]->pktCTRL_, timeout, cfb_);
			mhCfb_.start(wait);
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) {
				fprintf(macLog, "%f\n", NOW);
			}
#endif //=========================================================================
			return;
		default:
			fprintf(stderr, "Invalid CFB pktCTRL_ subtype\n");
			exit(1);
		}
	}

	// Send QoS CF-Poll
	if(tc[cfb_]->pktPOLL_) {
		struct hdr_mac802_11* dh = HDR_MAC802_11(tc[cfb_]->pktPOLL_);
		switch(dh->dh_fc.fc_subtype) {
		case MAC_Subtype_QoS_CFPoll:
			tc[cfb_]->set_tx_state(MAC_QOS_CFPOLL);
			// If sending the QoS CF-Poll
			// that ends the Polled TXOP
			// free cfb_ and txopLimit_, then return to
			// the CFP timer (that is done through send_timer)
			if(dh->dh_duration == 0) {
				// Update HCF statistics
				tc[cfb_]->cfb_time += (NOW
				                       - txopLimit_
				                       + txtime(tc[cfb_]->pktPOLL_));
				tc[cfb_]->cfb_num++;
				// Transmit the QoS CF-Poll frame
		   		timeout = txtime(tc[cfb_]->pktPOLL_);
		   		txop = cfb_;
				transmit(tc[cfb_]->pktPOLL_, timeout, cfb_);
				cfb_ = -1;
				txopLimit_ = 0;
				txopDuration_ = 0;
				// During CFP, increment
				// the Polling List index
            if(cfp_)
               if (!second_schedule)
                  cfPollIndex_++;
               else
                  cfPollIndex_ = cfPollNum_;
				// During CAP, increment
				// the Polling List index
				if(cap_ != -1)
					rrPollIndex_++;
			}
			// If sending a QoS CF-Poll
			// that starts the Polled TXOP
			// maintain cfb_ and restart CFB timer
			else {
				// XXX v2.0Start
				timeout = txtime(tc[cfb_]->pktPOLL_)
				          + DSSS_MaxPropagationDelay
				          + sifs_
					  + txtime(MAC_MaxMPDUSize, dataRate_)
					  + DSSS_MaxPropagationDelay;
				// XXX v2.0End
				wait = txtime(tc[cfb_]->pktPOLL_)
				       + sifs_
				       + DSSS_MaxPropagationDelay;
				// The timeout value must not
				// exceed the CFB remaining time
				if(timeout > remaining)
					timeout = remaining;
				txop = cfb_;
            //fprintf(stderr, "Sending QOS-CF-POLL: %d  %d\n", cfNumPolled_, cfPollIndex_);
				transmit(tc[cfb_]->pktPOLL_, timeout, cfb_);
				mhCfb_.start(wait);
			}
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) {
				fprintf(macLog, "%f\n", NOW);
			}
#endif //=========================================================================
			return;
		default:
			fprintf(stderr, "Invalid CFB pktPOLL_ subtype\n");
			exit(1);
		}
	}

	
	// Send QoS Data+CF-Poll, Data+CF-Ack, Data+CF-Ack+CF-Poll
	if(tc[cfb_]->pktTx_) {
		struct hdr_mac802_11* dh = HDR_MAC802_11(tc[cfb_]->pktTx_);
		switch(dh->dh_fc.fc_subtype) {
		case MAC_Subtype_QoS_Data_CFPoll:
			tc[cfb_]->set_tx_state(MAC_QOS_DATA_CFPOLL);
			break;
		case MAC_Subtype_QoS_Data_CFAck:
			tc[cfb_]->set_tx_state(MAC_QOS_DATA_CFACK);
			break;
		case MAC_Subtype_QoS_Data_CFAck_CFPoll:
			tc[cfb_]->set_tx_state(MAC_QOS_DATA_CFACK_CFPOLL);
			txopLimit_ += txtime(tc[cfb_]->pktTx_);
			break;
		default:
			fprintf(stderr, "Invalid CFB pktTx_ subtype\n");
			exit(1);
		}
		// XXX v2.0Start
		timeout = txtime(tc[cfb_]->pktTx_)
		          + DSSS_MaxPropagationDelay
		          + sifs_
			  + txtime(MAC_MaxMPDUSize, dataRate_)
			  + DSSS_MaxPropagationDelay;
		// XXX v2.0End
		wait = txtime(tc[cfb_]->pktTx_)
		       + sifs_
		       + DSSS_MaxPropagationDelay;
		// The timeout value must not
		// exceed the CFB remaining time
		if(timeout > remaining)
			timeout = remaining;
		txop = cfb_;
		transmit(tc[cfb_]->pktTx_, timeout, cfb_);
		mhCfb_.start(wait);
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
			fprintf(macLog, "%f\n", NOW);
		}
#endif //=========================================================================
	}

}

/*###################################################################
#                                                                   #
# sendCC: prepares and transmits the Contention Control frame       #
#         that starts the Controlled Contention Interval (CCI).     #
#         The Traffic Category 0 is associated to this frame        #
#                                                                   #
###################################################################*/

void Mac802_11HC::sendCC() {

	Packet *p = Packet::alloc();
	hdr_cmn *ch = HDR_CMN(p);
	struct cc_frame *cc = (struct cc_frame*)p->access(hdr_mac::offset_);
	double timeout;

	// Misc packet settings
	ch->uid() = 0;
	ch->ptype() = PT_MAC_CC;
	ch->mac_tc = 0;
	ch->mac_tc_aux = (u_int8_t)num_tc;
	ch->size() = ETHER_CC_LEN2;
	ch->iface() = -2;
	ch->error() = 0;

	bzero(cc, MAC_HDR_LEN);

	cc->cc_fc.fc_protocol_version = MAC_ProtocolVersion;
 	cc->cc_fc.fc_type	= MAC_Type_Control;
 	cc->cc_fc.fc_subtype	= MAC_Subtype_CC;
 	cc->cc_fc.fc_to_ds	= 0;
 	cc->cc_fc.fc_from_ds	= 0;
 	cc->cc_fc.fc_more_frag	= 0;
 	cc->cc_fc.fc_retry	= 0;
 	cc->cc_fc.fc_pwr_mgt	= 0;
 	cc->cc_fc.fc_more_data	= 0;
 	cc->cc_fc.fc_wep	= 0;
 	cc->cc_fc.fc_order	= 0;

	// The CC frame must be broadcasted
	int broadcast = (int) MAC_BROADCAST;
	STORE4BYTE(&broadcast, (cc->cc_ra));

	// Set current node as source
	STORE4BYTE(&index_, (cc->cc_bssid));

	// Store Nccop
	cc->cc_cci_length = nccop_;

	// Store tx time
	ch->txtime() = txtime(ch->size(), basicRate_);

	// Calculate duration
	double dccop = txtime(ETHER_RR_LEN2, basicRate_) + sifs_;
	cc->cc_duration = usec(pifs_ + (nccop_ * dccop) + DSSS_SlotTime);

	pktMNG_ = p;

	// Invoke the packet transmission
	
	tc[0]->set_tx_state(MAC_CC);
	timeout = txtime(ETHER_CC_LEN2, basicRate_);
	// 'txop' is set to 0 in order to
	// avoid controls in transmit()
	txop = 0;
#ifdef DEBUGACTIVE //=============================================================
	if(NOW > DEBUGTIME) {
		fprintf(macLog, "[%d] send CC frame %f\n", index_, NOW);
	}
#endif //=========================================================================
	transmit(pktMNG_, timeout, 0);

	// Set the NAV to the end of the CAP
	double dur = pifs_
	             + txtime(ETHER_CC_LEN2, basicRate_)
	             + DSSS_MaxPropagationDelay
	             + pifs_
	             + (nccop_ * dccop)
	             + (nccop_ * TXOPLmt);
	set_nav(dur * 1e6);
	// After a Contention Control frame has been sent,
	// the CAP starts: the HC's Traffic Categories
	// have to stay idle for all the CAP duration
	// so the stopTC() method is used to force
	// them to be silent
	for(int i=0; i<num_tc; i++)
		tc[i]->stopTC();

}

/*###################################################################
#                                                                   #
# sendBCN: prepares and transmits the Beacon frame that starts      #
#          the CFP. The Traffic Category 0 is associated to this    #
#          frame                                                    #
#                                                                   #
###################################################################*/

void Mac802_11HC::sendBCN() {

	Packet *p = Packet::alloc();
	hdr_cmn *ch = HDR_CMN(p);
	struct beacon_frame *bf = (struct beacon_frame*)p->access(hdr_mac::offset_);
	double timeout;

   //fprintf(stderr, "Sending the Beacon (%f)!\n", NOW);

   // Start HCF v3.0
   // Reschedule TXOPLMTs and CFPrate
   second_schedule = false; // needed only by SCHED_MMFAIR... we reset it anyway!
   if (scheduler == SCHED_ENHANCED || scheduler == SCHED_MMFAIR)
      schedule();
   // End HCF v3.0

	// Misc packet settings
	ch->uid() = 0;
	ch->ptype() = PT_MAC_BCN;
	ch->mac_tc = 0;
	ch->mac_tc_aux = (u_int8_t)num_tc;
	ch->size() = ETHER_BCN_LEN2;
	ch->iface() = -2;
	ch->error() = 0;

	bzero(bf, MAC_HDR_LEN);

	bf->bf_fc.fc_protocol_version = MAC_ProtocolVersion;
 	bf->bf_fc.fc_type	= MAC_Type_Management;
 	bf->bf_fc.fc_subtype	= MAC_Subtype_Beacon;
 	bf->bf_fc.fc_to_ds	= 0;
 	bf->bf_fc.fc_from_ds	= 0;
 	bf->bf_fc.fc_more_frag	= 0;
 	bf->bf_fc.fc_retry	= 0;
 	bf->bf_fc.fc_pwr_mgt	= 0;
 	bf->bf_fc.fc_more_data	= 0;
 	bf->bf_fc.fc_wep	= 0;
 	bf->bf_fc.fc_order	= 0;

	// Set current node as source
	STORE4BYTE(&index_, (bf->bf_sa));

	// The beacon must be broadcasted
	int broadcast = (int) MAC_BROADCAST;
	STORE4BYTE(&broadcast, (bf->bf_da));

	bf->bf_cf_pset.elementID 	= 4;	// beacon ID: 4
	bf->bf_cf_pset.length		= 8;	// 8 octects
	bf->bf_cf_pset.CFPCount		= 0;	// CFP starts now
	bf->bf_cf_pset.CFPPeriod	= (u_int8_t)(cfpRate_ / MAC_TIM);
	// The following duration fields have
	// to be stored in TU (double)
	// but they are 16bit integers:
	// therefore the maximum value is
	// 65536, and this is not enough for
	// a conversion in us. The solution
	// is to store the values in multiples
	// of 10us, so that the maximum
	// value of 41000 will fit
	bf->bf_cf_pset.CFPMaxDuration	= (u_int16_t)(MAC_CFPMaxDuration*1e5);
	bf->bf_cf_pset.CFPDurRemaining	= (u_int16_t)(cfpDurRemaining_*1e5);

	bf->bf_tim.elementID			= 5;	// TIM ID: 5
	bf->bf_tim.length			= 4;	// 4 octects
	bf->bf_tim.DTIMCount			= 0;	// this is a DTIM
	bf->bf_tim.DTIMPeriod			= 1;	// all TIMs are DTIMs

	// Store beacon tx time
	ch->txtime() = txtime(ch->size(), basicRate_);

	// Calculate beacon duration
	bf->bf_duration = 0;

	pktMNG_ = p;

	// Invoke the packet transmission
	
	tc[0]->set_tx_state(MAC_BEACON);
	timeout = txtime(ETHER_BCN_LEN2, basicRate_);
	// 'txop' is set to 0 in order to
	// avoid controls in transmit()
	txop = 0;
#ifdef DEBUGACTIVE //=============================================================
	if(NOW > DEBUGTIME) {
		fprintf(macLog, "[%d] send Beacon %f\n", index_, NOW);
	}
#endif //=========================================================================
	transmit(pktMNG_, timeout, 0);

	// Set the NAV to the end of the CFP
	set_nav(cfpDurRemaining_ * 1e6);
	// After a Beacon frame has been sent,
	// the CFP starts: the HC's Traffic Categories
	// have to stay idle for all the CFP duration
	// so the stopTC() method is used to force
	// them to be silent
	for(int i=0; i<num_tc; i++)
		tc[i]->stopTC();

}

/*###################################################################
#                                                                   #
# sendCFEND: prepares and transmits the CF-End packet that          #
#            terminates the CFP. The Traffic Category 0 is          #
#            associated to this frame                               #
#                                                                   #
###################################################################*/

void Mac802_11HC::sendCFEND() {

	Packet *p = Packet::alloc();
	hdr_cmn *ch = HDR_CMN(p);
	struct cfend_frame *ef = (struct cfend_frame*)p->access(hdr_mac::offset_);
	double timeout;
	
	// Misc packet settings
	ch->uid() = 0;
	ch->ptype() = PT_MAC_CFEND;
	ch->mac_tc = 0;
	ch->mac_tc_aux = (u_int8_t)num_tc;
	ch->size() = ETHER_CFEND_LEN2;
	ch->iface() = -2;
	ch->error() = 0;

	bzero(ef, MAC_HDR_LEN);

	ef->ef_fc.fc_protocol_version = MAC_ProtocolVersion;
 	ef->ef_fc.fc_type	= MAC_Type_Control;
 	ef->ef_fc.fc_subtype	= MAC_Subtype_CFEnd;
 	ef->ef_fc.fc_to_ds	= 0;
 	ef->ef_fc.fc_from_ds	= 0;
 	ef->ef_fc.fc_more_frag	= 0;
 	ef->ef_fc.fc_retry	= 0;
 	ef->ef_fc.fc_pwr_mgt	= 0;
 	ef->ef_fc.fc_more_data	= 0;
 	ef->ef_fc.fc_wep	= 0;
 	ef->ef_fc.fc_order	= 0;

	// The CF-End frame must be broadcasted
	int broadcast = (int) MAC_BROADCAST;
	STORE4BYTE(&broadcast, (ef->ef_ra));

	// Store CF-End frame tx time
	ch->txtime() = txtime(ch->size(), basicRate_);

	// Calculate CF-End frame duration
	ef->ef_duration = 0;

	pktMNG_ = p;

	// Update HCF statistics
	cfp_time += (NOW - cfStartTime_ + txtime(p));
	cfp_num++;

	// Invoke the packet transmission

	tc[0]->set_tx_state(MAC_CFEND);
	timeout = txtime(ETHER_CFEND_LEN2, basicRate_);
	// 'txop' is set to avoid checks in transmit()
	txop = 0;
#ifdef DEBUGACTIVE //=============================================================
	if(NOW > DEBUGTIME) {
		fprintf(macLog, "[%d] send CF-End %f (cfp_time: %f, cfp_num: %d)\n", index_, NOW, (double)cfp_time, (int)cfp_num);
	}
#endif //=========================================================================
	transmit(pktMNG_, timeout, 0);

}

// ---------- HCFstart
/*###################################################################
#                                                                   #
# recv_pktCheck: due to the error model, stations can ignore        #
#                QoS CF-Poll/CF-End frames, therefore transmitting  #
#                QoS frames outside their Polled TXOP. This method  #
#                controls if the received QoS frame is the one we   #
#                actually expected to be, providing protection      #
#                to the medium access procedure                     #
#                                                                   #
#                 src: the source address of the received frame     #
#                  nt: the TC of the received frame                 #
#                type: the type of the frame, as a string           #
#                                                                   #
###################################################################*/

bool Mac802_11HC::recv_pktCheck(int src, int nt, char* type) {

	bool pktOK = 1;

	// During a CFB, if:
	// - frame directed to the wrong TC
	// - frame coming from the wrong station during CFP
	// - frame coming from the wrong station during CAP
	// then the packet must be discarded and the
	// transmission must be resumed restarting mhCfp_
	if(cfb_ != -1) {
		if(nt != cfb_	||
		   (nt == cfb_	&& cfp_		&& src != cfPollList_[cfPollIndex_].addr) ||
		   (nt == cfb_  && cap_ != -1	&& src != rrPollList_[rrPollIndex_].addr)    ) {
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) {
	  				fprintf(macLog, "[%d] incoming %s discarded during CFB (wrong station/TC: %d.%d), retransmit ", index_, type, src, nt);
			}
#endif //=========================================================================
			// Stop the send timer
			if(tc[cfb_]->mhSend_.busy())
				tc[cfb_]->mhSend_.stop();
			// If Data was sent, set the retry bit
			// and update the statistics
			if(tc[cfb_]->tx_state_ == MAC_QOS_DATA_CFPOLL ||
			   tc[cfb_]->tx_state_ == MAC_QOS_DATA_CFACK ||
			   tc[cfb_]->tx_state_ == MAC_QOS_DATA_CFACK_CFPOLL) {
				HDR_MAC802_11(tc[cfb_]->pktTx_)->dh_fc.fc_retry = 1;
				switch (HDR_CMN(tc[cfb_]->pktTx_)->ptype()) {     // Statistics: retransmitted frames
					case PT_TCP:
						++(tc[cfb_]->rtx_tcp);
						break;
					case PT_ACK:
						++(tc[cfb_]->rtx_ack);
						break;
					case PT_MESSAGE:
						++(tc[cfb_]->rtx_message);
						break;
					case PT_ARP:
						++(tc[cfb_]->rtx_ARP);
						break;
               case PT_VBR:
					case PT_CBR:
						++(tc[cfb_]->rtx_udp);
						break;
					case PT_MAC_RTS:
						++(tc[cfb_]->rtx_mac_rts);
						break;
					case PT_MAC_CTS:
						++(tc[cfb_]->rtx_mac_cts);
						break;
					case PT_MAC_ACK:
						++(tc[cfb_]->rtx_mac_ack);
						break;
				}
			}
			tc[cfb_]->set_tx_state(MAC_IDLE);
			mhCfb_.start(sifs_);
			discard(pktRx_, "---");
			pktOK = 0;
		}
	}
	// If the frame is received during CFP
	// or CAP while waiting SIFS between
	// two CFBs, simply discard it
	if((cfp_ && cfb_ == -1) ||
	   (cap_ != -1 && cfb_ == -1)) {
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
   			fprintf(macLog, "[%d] incoming %s discarded during CFP/CAP (sender: %d.%d)\n", index_, type, src, nt);
		}
#endif //=========================================================================
	   	discard(pktRx_, "---");
	   	pktOK = 0;
	}
	// If the frame is received during CP
	// discard it and resume transmission
	if(!cfp_ && cap_ == -1) {
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
			fprintf(macLog, "[%d] incoming %s discarded during CP (sender: %d.%d)\n", index_, type, src, nt);
		}
#endif //=========================================================================
		discard(pktRx_, "---");
		pktOK = 0;
	}

	return(pktOK);

}
// ---------- HCFend

/*###################################################################
#                                                                   #
# recv_timer: this function is invoked when the last bit of a frame #
#             is received from the channel. This frame is processed #
#             only if it is error-free and if there is no collision.#
#                                                                   #
###################################################################*/

void Mac802_11HC::recv_timer() {
	hdr_cmn *ch = HDR_CMN(pktRx_);
	hdr_mac802_11 *mh = HDR_MAC802_11(pktRx_);
	u_int32_t dst = ETHER_ADDR(mh->dh_da);
	u_int32_t src = ETHER_ADDR(mh->dh_sa);
	bool discarded = 0;

	u_int8_t  type = mh->dh_fc.fc_type;
	u_int8_t  subtype = mh->dh_fc.fc_subtype;

	assert(pktRx_);
	assert(rx_state_ == MAC_RECV || rx_state_ == MAC_COLL);

        int nt;
        if (num_tc == 1) {
                nt = 0;
        } else {
                nt = nt = HDR_CMN(pktRx_)->mac_tc;          // Extracts the identifier of the queue at which this frame belongs
        }

        /*
         *  If the interface is in TRANSMIT mode when this packet
         *  "arrives", then I would never have seen it and should
         *  do a silent discard without adjusting the NAV.
         */

        if(tx_active_ != -1) {
                switch (ch->ptype()) {                          // Statistics: these frames are considered as destined to others nodes
                        case PT_TCP:
                                (tc[nt]->rcv_chn_other_tcp)++;
                                break;
                        case PT_ACK:
                                (tc[nt]->rcv_chn_other_ack)++;
                                break;
                        case PT_MESSAGE:
                                (tc[nt]->rcv_chn_other_message)++;
                                break;
                        case PT_ARP:
                                (tc[nt]->rcv_chn_other_ARP)++;
                                break;
                        case PT_VBR:
                        case PT_CBR:
                                (tc[nt]->rcv_chn_other_udp)++;
                                break;
                        case PT_VOIP:
                                (tc[nt]->rcv_chn_other_voip)++;
                                break;	
                        case PT_MAC_RTS:
                                (tc[nt]->rcv_chn_other_mac_rts)++;
                                break;
                        case PT_MAC_CTS:
                                (tc[nt]->rcv_chn_other_mac_cts)++;
                                break;
                        case PT_MAC_ACK:
                                (tc[nt]->rcv_chn_other_mac_ack)++;
                                break;
                }
		// If the response packet is discarded
		// during a CFB while waiting for a response,
		// recovery by retransmitting after SIFS
		// (802.11e - 9.10.1.2)
		if((cfb_ != -1) &&
		   !mhCfb_.busy()) {
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) {
				fprintf(macLog, "[%d] incoming packet discarded (transmitting now), retransmit ", index_);
			}
#endif //=========================================================================
			// Stop the send timer
			if(tc[cfb_]->mhSend_.busy())
				tc[cfb_]->mhSend_.stop();
			// If Data was sent, set the retry bit
			// and update the statistics
			if(tc[cfb_]->tx_state_ == MAC_QOS_DATA_CFPOLL ||
			   tc[cfb_]->tx_state_ == MAC_QOS_DATA_CFACK ||
			   tc[cfb_]->tx_state_ == MAC_QOS_DATA_CFACK_CFPOLL) {
				HDR_MAC802_11(tc[cfb_]->pktTx_)->dh_fc.fc_retry = 1;

				switch (HDR_CMN(tc[cfb_]->pktTx_)->ptype()) {     // Statistics: retransmitted frames
					case PT_TCP:
						++(tc[cfb_]->rtx_tcp);
						break;
					case PT_ACK:
						++(tc[cfb_]->rtx_ack);
						break;
					case PT_MESSAGE:
						++(tc[cfb_]->rtx_message);
						break;
					case PT_ARP:
						++(tc[cfb_]->rtx_ARP);
						break;
               case PT_VBR:
					case PT_CBR:
						++(tc[cfb_]->rtx_udp);
						break;
					case PT_VOIP:
						++(tc[cfb_]->rtx_voip);
						break;	
					case PT_MAC_RTS:
						++(tc[cfb_]->rtx_mac_rts);
						break;
					case PT_MAC_CTS:
						++(tc[cfb_]->rtx_mac_cts);
						break;
					case PT_MAC_ACK:
						++(tc[cfb_]->rtx_mac_ack);
						break;
				}
			}
			tc[cfb_]->set_tx_state(MAC_IDLE);
			mhCfb_.start(sifs_);
		}
                Packet::free(pktRx_);
                goto done;
        }

	/*
	 * Handle collisions.
	 */

	if(rx_state_ == MAC_COLL) {
                switch (ch->ptype()) {          // Statistics: colliding frames
                        case PT_TCP:
                                (tc[nt]->rcv_chn_col_tcp)++;
                                break;
                        case PT_ACK:
                                (tc[nt]->rcv_chn_col_ack)++;
                                break;
                        case PT_MESSAGE:
                                (tc[nt]->rcv_chn_col_message)++;
                                break;
                        case PT_ARP:
                                (tc[nt]->rcv_chn_col_ARP)++;
                                break;
                        case PT_VBR:
                        case PT_CBR:
                                (tc[nt]->rcv_chn_col_udp)++;
                                break;
                        case PT_VOIP:
                                (tc[nt]->rcv_chn_col_voip)++;
                                break;	
                        case PT_MAC_RTS:
                                (tc[nt]->rcv_chn_col_mac_rts)++;
                                break;
                        case PT_MAC_CTS:
                                (tc[nt]->rcv_chn_col_mac_cts)++;
                                break;
                        case PT_MAC_ACK:
                                (tc[nt]->rcv_chn_col_mac_ack)++;
                                break;
                }
		// If the response packet is discarded
		// during a CFB while waiting for a response,
		// recovery by retransmitting after SIFS
		// (802.11e - 9.10.1.2)
		if((cfb_ != -1) &&
		   !mhCfb_.busy()) {
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) {
				fprintf(macLog, "[%d] incoming packet discarded (reception collision), retransmit ", index_);
			}
#endif //=========================================================================
			// Stop the send timer
			if(tc[cfb_]->mhSend_.busy())
				tc[cfb_]->mhSend_.stop();
			// If Data was sent, set the retry bit
			// and update the statistics
			if(tc[cfb_]->tx_state_ == MAC_QOS_DATA_CFPOLL ||
			   tc[cfb_]->tx_state_ == MAC_QOS_DATA_CFACK ||
			   tc[cfb_]->tx_state_ == MAC_QOS_DATA_CFACK_CFPOLL) {
				HDR_MAC802_11(tc[cfb_]->pktTx_)->dh_fc.fc_retry = 1;

				switch (HDR_CMN(tc[cfb_]->pktTx_)->ptype()) {     // Statistics: retransmitted frames
					case PT_TCP:
						++(tc[cfb_]->rtx_tcp);
						break;
					case PT_ACK:
						++(tc[cfb_]->rtx_ack);
						break;
					case PT_MESSAGE:
						++(tc[cfb_]->rtx_message);
						break;
					case PT_ARP:
						++(tc[cfb_]->rtx_ARP);
						break;
               case PT_VBR:
					case PT_CBR:
						++(tc[cfb_]->rtx_udp);
						break;
					case PT_VOIP:
						++(tc[cfb_]->rtx_voip);
						break;	
					case PT_MAC_RTS:
						++(tc[cfb_]->rtx_mac_rts);
						break;
					case PT_MAC_CTS:
						++(tc[cfb_]->rtx_mac_cts);
						break;
					case PT_MAC_ACK:
						++(tc[cfb_]->rtx_mac_ack);
						break;
				}
			}
			tc[cfb_]->set_tx_state(MAC_IDLE);
			mhCfb_.start(sifs_);
		}
		discard(pktRx_, DROP_MAC_COLLISION);
		err_frame = 1;                  // The last received frame is not error-free
		// Set 'discarded' to TRUE,
		// because the frame
		// exchange terminates here
		if(dst == (u_int32_t)index_)
			discarded = 1;
		goto done;
	}

	/*
	 * Check to see if this packet was received with enough
	 * bit errors that the current level of FEC still could not
	 * fix all of the problems - ie; after FEC, the checksum still
	 * failed.
	 */

	if( ch->error() ) {
                switch (ch->ptype()) {                          // Statistics: corrupted packets
                        case PT_TCP:
                                (tc[nt]->rcv_chn_err_tcp)++;
                                break;
                        case PT_ACK:
                                (tc[nt]->rcv_chn_err_ack)++;
                                break;
                        case PT_MESSAGE:
                                (tc[nt]->rcv_chn_err_message)++;
                                break;
                        case PT_ARP:
                                (tc[nt]->rcv_chn_err_ARP)++;
                                break;
                        case PT_VBR:
                        case PT_CBR:
                                (tc[nt]->rcv_chn_err_udp)++;
                                break;
                        case PT_VOIP:
                                (tc[nt]->rcv_chn_err_voip)++;
                                break;	
                        case PT_MAC_RTS:
                                (tc[nt]->rcv_chn_err_mac_rts)++;
                                break;
                        case PT_MAC_CTS:
                                (tc[nt]->rcv_chn_err_mac_cts)++;
                                break;
                        case PT_MAC_ACK:
                                (tc[nt]->rcv_chn_err_mac_ack)++;
                                break;
                }
		// If the response packet is discarded
		// during a CFB while waiting for a response,
		// recovery by retransmitting after SIFS
		// (802.11e - 9.10.1.2)
		if((cfb_ != -1) &&
		   !mhCfb_.busy()) {
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) {
				fprintf(macLog, "[%d] incoming packet discarded (error), retransmit ", index_);
			}
#endif //=========================================================================
			// Stop the send timer
			if(tc[cfb_]->mhSend_.busy())
				tc[cfb_]->mhSend_.stop();
			// If Data was sent, set the retry bit
			// and update the statistics
			if(tc[cfb_]->tx_state_ == MAC_QOS_DATA_CFPOLL ||
			   tc[cfb_]->tx_state_ == MAC_QOS_DATA_CFACK ||
			   tc[cfb_]->tx_state_ == MAC_QOS_DATA_CFACK_CFPOLL) {
				HDR_MAC802_11(tc[cfb_]->pktTx_)->dh_fc.fc_retry = 1;

				switch (HDR_CMN(tc[cfb_]->pktTx_)->ptype()) {     // Statistics: retransmitted frames
					case PT_TCP:
						++(tc[cfb_]->rtx_tcp);
						break;
					case PT_ACK:
						++(tc[cfb_]->rtx_ack);
						break;
					case PT_MESSAGE:
						++(tc[cfb_]->rtx_message);
						break;
					case PT_ARP:
						++(tc[cfb_]->rtx_ARP);
						break;
               case PT_VBR:
					case PT_CBR:
						++(tc[cfb_]->rtx_udp);
						break;
					case PT_VOIP:
						++(tc[cfb_]->rtx_voip);
						break;	
					case PT_MAC_RTS:
						++(tc[cfb_]->rtx_mac_rts);
						break;
					case PT_MAC_CTS:
						++(tc[cfb_]->rtx_mac_cts);
						break;
					case PT_MAC_ACK:
						++(tc[cfb_]->rtx_mac_ack);
						break;
				}
			}
			tc[cfb_]->set_tx_state(MAC_IDLE);
			mhCfb_.start(sifs_);
		}	
		Packet::free(pktRx_);
		err_frame = 1;                  // The last received frame is not error-free
		// Set 'discarded' to TRUE,
		// because the frame
		// exchange terminates here
		if(dst == (u_int32_t)index_)
			discarded = 1;
		goto done;
	}

        err_frame = 0;          // The last received frame is error-free

	/*
	 * IEEE 802.11 specs, section 9.2.5.6
	 *	- update the NAV (Network Allocation Vector)
	 */

	if(dst != (u_int32_t)index_) {          // If this frame is not destined to this node, the NAV must be updated
		set_nav(mh->dh_duration);
	}

        // tap out

        if (tap_ && type == MAC_Type_Data && MAC_Subtype_Data == subtype) {
                tap_->tap(pktRx_);
        }

	/*
	 * Adaptive Fidelity Algorithm Support - neighborhood infomation
	 * collection
	 *
	 * Hacking: Before filter the packet, log the neighbor node
	 * I can hear the packet, the src is my neighbor
	 */

	if (netif_->node()->energy_model() &&
	    netif_->node()->energy_model()->adaptivefidelity()) {
		netif_->node()->energy_model()->add_neighbor(src);
	}

	/*
	 * Address Filtering
	 */

	if(dst != (u_int32_t)index_ && dst != MAC_BROADCAST) {          // The frame belongs to another node, so we can discard it

		/*
		 *  We don't want to log this event, so we just free
		 *  the packet instead of calling the drop routine.
		 */

                switch (ch->ptype()) {          // Statistics: frames destined to others nodes
                        case PT_TCP:
                                (tc[nt]->rcv_chn_other_tcp)++;
                                break;
                        case PT_ACK:
                                (tc[nt]->rcv_chn_other_ack)++;
                                break;
                        case PT_MESSAGE:
                                (tc[nt]->rcv_chn_other_message)++;
                                break;
                        case PT_ARP:
                                (tc[nt]->rcv_chn_other_ARP)++;
                                break;
                        case PT_VBR:
                        case PT_CBR:
                                (tc[nt]->rcv_chn_other_udp)++;
                                break;
                        case PT_VOIP:
                                (tc[nt]->rcv_chn_other_voip)++;
                                break;	
                        case PT_MAC_RTS:
                                (tc[nt]->rcv_chn_other_mac_rts)++;
                                break;
                        case PT_MAC_CTS:
                                (tc[nt]->rcv_chn_other_mac_cts)++;
                                break;
                        case PT_MAC_ACK:
                                (tc[nt]->rcv_chn_other_mac_ack)++;
                                break;
                }
		discard(pktRx_, "---");
		goto done;
	}

	switch(type) {
	case MAC_Type_Management:
		discard(pktRx_, DROP_MAC_PACKET_ERROR);         // Management frames not implemented yet
		goto done;
		break;

	case MAC_Type_Control:
		switch(subtype) {
		case MAC_Subtype_RTS:
                        (tc[nt]->rcv_chn_my_mac_rts)++;        // Receiving a RTS
			tc[nt]->recvRTS(pktRx_);
			break;
		case MAC_Subtype_CTS:
                        (tc[nt]->rcv_chn_my_mac_cts)++;        // Receiving a CTS
			tc[nt]->recvCTS(pktRx_);
			break;
		case MAC_Subtype_ACK:
                        (tc[nt]->rcv_chn_my_mac_ack)++;        // Receiving an ACK
			tc[nt]->recvACK(pktRx_);
			break;
		case MAC_Subtype_RR:
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) {
				fprintf(macLog, "[%d] Reservation Request received %f\n", index_, NOW);
			}
#endif //=========================================================================
			recvRR(pktRx_);
			break;
		default:
			fprintf(stderr,"recvTimer1:Invalid MAC Control Subtype %x\n", subtype);
			exit(1);
		}
		break;
	case MAC_Type_Data:
		switch(subtype) {                       // Receiving a DATA frame
		case MAC_Subtype_QoS_Data:
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) {
				fprintf(macLog, "[%d.%d] Data received %f\n", index_, nt, NOW);
			}
#endif //=========================================================================
			// Check if received frame is
			// the one we are waiting for
			if(!recv_pktCheck(src, nt, "QoS Data"))
				goto done;
			// If yes, fall through
		case MAC_Subtype_Data:
                        switch (ch->ptype()) {          // Statistics: frames correctly received
                                case PT_TCP:
                                        (tc[nt]->rcv_chn_my_tcp)++;
                                        ch->tmac3 = NOW;
                                        break;
                                case PT_ACK:
                                        (tc[nt]->rcv_chn_my_ack)++;
                                        break;
                                case PT_MESSAGE:
                                        (tc[nt]->rcv_chn_my_message)++;
                                        break;
                                case PT_ARP:
                                        (tc[nt]->rcv_chn_my_ARP)++;
                                        break;
                                case PT_VBR:
                                case PT_CBR:
                                        (tc[nt]->rcv_chn_my_udp)++;
                                        ch->tmac3 = NOW;
                                        break;
                                case PT_VOIP:
                                        (tc[nt]->rcv_chn_my_voip)++;
                                        ch->tmac3 = NOW;
                                        break;	
                        }
			tc[nt]->recvDATA(pktRx_);
			break;
		case MAC_Subtype_QoS_Data_CFAck:
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) {
				fprintf(macLog, "[%d.%d] Data+CF-Ack received %f\n", index_, nt, NOW);
			}
#endif //=========================================================================
			// Check if received frame is
			// the one we are waiting for
			if(!recv_pktCheck(src, nt, "Data+CF-Ack"))
				goto done;
			// If yes, continue with reception
			switch (ch->ptype()) {          // Statistics: frames correctly received
                                case PT_TCP:
                                        (tc[nt]->rcv_chn_my_tcp)++;
                                        ch->tmac3 = NOW;
                                        break;
                                case PT_ACK:
                                        (tc[nt]->rcv_chn_my_ack)++;
                                        break;
                                case PT_MESSAGE:
                                        (tc[nt]->rcv_chn_my_message)++;
                                        break;
                                case PT_ARP:
                                        (tc[nt]->rcv_chn_my_ARP)++;
                                        break;
                                case PT_VBR:
                                case PT_CBR:
                                        (tc[nt]->rcv_chn_my_udp)++;
                                        ch->tmac3 = NOW;
                                        break;
                                case PT_VOIP:
                                        (tc[nt]->rcv_chn_my_voip)++;
                                        ch->tmac3 = NOW;
                                        break;	
                        }
			tc[nt]->recvDATA_CFACK(pktRx_);
			break;
		case MAC_Subtype_QoS_CFAck:
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) {
				fprintf(macLog, "[%d.%d] CF-Ack received (end of CFB) %f\n", index_, nt, NOW);
			}
#endif //=========================================================================
			// Check if received frame is
			// the one we are waiting for
			if(!recv_pktCheck(src, nt, "CF-Ack"))
				goto done;
			// If yes, continue with reception
			tc[nt]->recvCFACK(pktRx_);
			break;
		case MAC_Subtype_QoS_Null:
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) {
				fprintf(macLog, "[%d.%d] Null received (end of CFB) %f\n", index_, nt, NOW);
			}
#endif //=========================================================================
			// Check if received frame is
			// the one we are waiting for
			if(!recv_pktCheck(src, nt, "Null"))
				goto done;
			// If yes, fall through
			tc[nt]->recvNULL(pktRx_);
			break;
		default:
			fprintf(stderr, "recv_timer2:Invalid MAC Data Subtype %x\n", subtype);
			exit(1);
		}
		break;
	default:
		fprintf(stderr, "recv_timer3:Invalid MAC Type %x\n", subtype);
		exit(1);
	}
done:
	pktRx_ = 0;
	rx_resume();    // The frame reception is ended

	// Pending CFP or CAP [case ERROR].
	// We just received an error packet: frame exchange terminated.
	// If we were waiting to start the CFP or CAP, it can start now
	if(discarded && is_idle()) {
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
			if(cfpWaiting_ || capWaiting_) {
				fprintf(macLog, "[%d] frame exchange error, channel idle %f\n", index_, NOW);
			}
		}
#endif //=========================================================================
		// wait PIFS
		if(cfpWaiting_) {
			mhCfpStart_.start(pifs_);
		}
		if(capWaiting_)
			mhCapStart_.start(pifs_);
	}
}

/*###################################################################
#                                                                   #
# recvRR: called every time a Reservation Request frame is received #
#         this method updates the Polling List for the following    #
#         Controlled Access Period, inserting the MAC address, the  #
#         Traffic Category and the requested TXOP duration that     #
#         are in the received RR frame                              #
#                                                                   #
#         p : the Reservation Request frame                         #
#                                                                   #
###################################################################*/

void Mac802_11HC::recvRR(Packet *p) {

	// Ignore RR frames outside CAPs
	if(cap_ == -1) {
		discard(p, "---");
		return;
	}

	struct hdr_cmn *ch = HDR_CMN(p);
	struct hdr_mac802_11 *dh = HDR_MAC802_11(p);
	struct rr_frame *rr = (struct rr_frame*)p->access(hdr_mac::offset_);

	// Update Reservation Request List
	rrPollList_[rrPollNum_].addr = (u_int32_t)ETHER_ADDR(dh->dh_sa);
	rrPollList_[rrPollNum_].tc = ch->mac_tc;
	rrPollList_[rrPollNum_].txop_duration = rr->rr_txop_duration * 1e-6;
	rrPollNum_++;
	
	// Free the received frame
	mac_log(p);

}


Mac802_11eHC::Mac802_11eHC(Mac802_11* upmac, PHY_MIB* p, MAC_MIB *m) : Mac802_11e(upmac, p, m) {}

/*###################################################################
#                                                                   #
# sendHandler: called by mac-timers.cc when a transmission timeout  #
#              expires                                              #
#                                                                   #
###################################################################*/

void Mac802_11eHC::sendHandler() {
	send_timer();
}

/*###################################################################
#                                                                   #
# send_timer: called when a transmission timeout expires, this      #
#             method performs the operations that have to be done   #
#             according to the type of the frame that was sent:     #
#             it can retransmit the frame or discard it             #
#                                                                   #
###################################################################*/

void Mac802_11eHC::send_timer() {

	hdr_mac802_11* dh;
	struct qos_cfpoll_frame* pf;

	if (bursting > 1) {             // In bursting mode, the timeout expiration forces the end of the current burst, and the last
                bursting = 1;           // frame not "acked" must be retransmitted with the normal access procedure
                ptosend = 0;
                ((BS_QueueHandler*)(mac_target_->callback_))->stop_bursting(nq);        // Informs the Interface Queue about the end
        }                                                                               // of this burst

	switch(tx_state_) {

	/*
	 * Sent a RTS, but did not receive a CTS.
	 */

	case MAC_RTS:
		RetransmitRTS();
		break;

	/*
	 * Sent a CTS, but did not receive a DATA packet.
	 */

	case MAC_CTS:
		assert(pktCTRL_);                               // CTS frames are not retransmitted
		Packet::free(pktCTRL_); pktCTRL_ = 0;
		break;

	/*
	 * Sent DATA, but did not receive an ACK packet.
	 */

	case MAC_SEND:
		RetransmitDATA();
		break;

	/*
	 * Sent an ACK, and now ready to resume transmission.
	 */

	case MAC_ACK:
		assert(pktCTRL_);                               // ACK frames are not retransmitted
		Packet::free(pktCTRL_); pktCTRL_ = 0;

		// Pending CFP or CAP [case ACKSENT].
		// We just sent an ACK: frame exchange terminated.
		// If we were waiting to start the CFP or CAP, it can start now
		if((mac_target_->rx_state_ == MAC_IDLE) &&
		   (mac_target_->nav_ < NOW)) {
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) {
		   		if(mac_target_->cfpWaiting_ ||
				   mac_target_->capWaiting_) {
					fprintf(macLog, "[%d] ACK sent, channel idle %f\n", mac_target_->index_, NOW);
				}
			}
#endif //=========================================================================
			// wait PIFS
			if(mac_target_->cfpWaiting_)
				mac_target_->mhCfpStart_.start(mac_target_->pifs_);
			if(mac_target_->capWaiting_)
				mac_target_->mhCapStart_.start(mac_target_->pifs_);

			// We don't want the transmission to be resumed as
			// usual because we are entering the CFP or CAP, so
			// we set the transmission state to MAC_IDLE and then
			// wait for the mhCfpStart_ or mhCapStart_ timer to expire
			set_tx_state(MAC_IDLE);
			return;
		}

		break;

	// Sent a beacon, free the frame

	case MAC_BEACON:
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
			fprintf(macLog, "[%d] Beacon sent\n", mac_target_->index_);
		}
#endif //=========================================================================
		Packet::free(mac_target_->pktMNG_);
		mac_target_->pktMNG_ = 0;
		break;

	// Sent a Contention Control frame,
	// free it and reset the tx_state_

	case MAC_CC:
		Packet::free(mac_target_->pktMNG_);
		mac_target_->pktMNG_ = 0;
		set_tx_state(MAC_IDLE);
		return;
	
	// Sent a CF-End, free the frame 
	// and prepare the HC to go back to CP

	case MAC_CFEND:
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
			fprintf(macLog, "[%d] CF-End sent\n", mac_target_->index_);
		}
#endif //=========================================================================
		Packet::free(mac_target_->pktMNG_);
		mac_target_->pktMNG_ = 0;
		mac_target_->cfp_ = 0;
		mac_target_->cfpWaiting_ = 0;
		mac_target_->txop = -1;
		// Reset the NAV
		if(mac_target_->mhNav_.busy()) {
			mac_target_->mhNav_.stop();
		}
		mac_target_->nav_ = NOW;
		for(int i=0; i<(mac_target_->num_tc); i++) {
			// This is needed because 'tx_state_'
			// can be set to a value different from
			// MAC_IDLE when sending CF-End frame
         if (mac_target_->tc[i]->edcaTXenabled()) {
            mac_target_->tc[i]->set_tx_state(MAC_IDLE);
            mac_target_->tc[i]->resumeTC();
         }
		}
		// tx_resume() already invoked in
		// previous cycle, so return here
		return;

	// Sent a QoS CF-Poll or QoS Data+CF-Poll, resume tx

	case MAC_QOS_CFPOLL:
		// Sent the QoS CF-Poll that ends the Polled TXOP,
		// free the QoS CF-Poll packet
		if((HDR_MAC802_11(pktPOLL_))->dh_duration == 0) {
			Packet::free(pktPOLL_);
			pktPOLL_ = 0;
			break;
		}
		// Update TXOP duration to
		// the remaining TXOP time
		pf = (struct qos_cfpoll_frame*)pktPOLL_->access(hdr_mac::offset_);
		pf->pf_txop_duration = mac_target_->usec(mac_target_->txopLimit_
		                                         + mac_target_->txopDuration_
		                                         - NOW);
		// fall through
	case MAC_QOS_DATA_CFPOLL:
		// Update the MAC header Duration/ID 
		// field to the remaining TXOP duration
		if(pktPOLL_)
			dh = HDR_MAC802_11(pktPOLL_);
		if(pktTx_ && !pktPOLL_)
			dh = HDR_MAC802_11(pktTx_);
		dh->dh_duration = mac_target_->usec(mac_target_->txopLimit_
		                                    + mac_target_->txopDuration_
		                                    - NOW
		                                    + DSSS_SlotTime);
		// fall through
	case MAC_QOS_DATA_CFACK:
	case MAC_QOS_DATA_CFACK_CFPOLL:
#ifdef DEBUGACTIVE //=============================================================
		if(NOW > DEBUGTIME) {
			fprintf(macLog, "[%d.%d] Send timer expired during CFB %f, attempt retransmission ", mac_target_->index_, nq, NOW);
		}
#endif //=========================================================================
		if(tx_state_ != MAC_QOS_CFPOLL) {
			// Set the retry bit
			HDR_MAC802_11(pktTx_)->dh_fc.fc_retry = 1;

			switch (HDR_CMN(pktTx_)->ptype()) {     // Statistics: retransmitted frames
				case PT_TCP:
					++rtx_tcp;
					break;
				case PT_ACK:
					++rtx_ack;
					break;
				case PT_MESSAGE:
					++rtx_message;
					break;
				case PT_ARP:
					++rtx_ARP;
					break;
            case PT_VBR:
				case PT_CBR:
					++rtx_udp;
					break;
				case PT_VOIP:
					++rtx_voip;
					break;	
				case PT_MAC_RTS:
					++rtx_mac_rts;
					break;
				case PT_MAC_CTS:
					++rtx_mac_cts;
					break;
				case PT_MAC_ACK:
					++rtx_mac_ack;
					break;
			}
		}

		set_tx_state(MAC_IDLE);
		mac_target_->cfbHandler();
		return;

	// Sent a QoS CF-Ack, resume tx

	case MAC_QOS_CFACK:
		// QoS CF-Ack packets are freed only
		// when the response frame is received.
		// Return here, it is necessary
		// to keep current tx_state_
		return;
	
	case MAC_IDLE:
		break;
	default:
		assert(0);
	}
	tx_resume();
}

/*###################################################################
#                                                                   #
# cfbSendCFPOLL: prepares a QoS CF-Poll frame for this TC. If the   #
#                CFB flag is equal to -1, then a QoS CF-Poll that   #
#                starts a new CFB is prepared, if the CFB flag is   #
#                already allocated to a TC, a QoS CF-Poll that ends #
#                the current CFB is prepared (with duration = 0 and #
#                destination address matching the HC's MAC address) #
#                                                                   #
#                dst: the destination MAC address                   #
#                                                                   #
###################################################################*/

void Mac802_11eHC::cfbSendCFPOLL(int dst) {

	Packet* p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct qos_cfpoll_frame* pf = (struct qos_cfpoll_frame*)p->access(hdr_mac::offset_);
	
	ch->uid() = 0;
	ch->ptype() = PT_MAC_QOS_CFPOLL;
	ch->mac_tc = (u_int8_t)nq;
	ch->mac_tc_aux = (u_int8_t)(mac_target_->num_tc);
	ch->size() = ETHER_QOS_CFPOLL_LEN;
	ch->iface() = -2;
	ch->error() = 0;
	
	bzero(pf, MAC_HDR_LEN);

	pf->pf_fc.fc_protocol_version = MAC_ProtocolVersion;
	pf->pf_fc.fc_type	= MAC_Type_Data;
	pf->pf_fc.fc_subtype	= MAC_Subtype_QoS_CFPoll;
	pf->pf_fc.fc_to_ds	= 0;
	pf->pf_fc.fc_from_ds	= 0;
	pf->pf_fc.fc_more_frag	= 0;
	pf->pf_fc.fc_retry	= 0;
	pf->pf_fc.fc_pwr_mgt	= 0;
	pf->pf_fc.fc_more_data	= 0;
	pf->pf_fc.fc_wep	= 0;
	pf->pf_fc.fc_order	= 0;

	STORE4BYTE(&dst, (pf->pf_ra));
	STORE4BYTE(&(mac_target_->index_), (HDR_MAC802_11(p)->dh_sa));

	// Store QoS CF-Poll tx time
	// XXX v2.0Start
	ch->txtime() = mac_target_->txtime(ch->size(), mac_target_->dataRate_);
	// XXX v2.0End

   // Store Polled TXOP duration
   pf->pf_txop_duration = mac_target_->usec(mac_target_->txopLimit_
                                            + mac_target_->txopDuration_
                                            - NOW);

   // Calculate QoS CF-Poll duration:
   // If this starts a Polled TXOP, duration is the TXOP duration value.
   // If this ends a Polled TXOP, ra is HC's MAC address and duration is 0.
   // (802.11e - 9.10.2.1)
   if(mac_target_->cfb_ == -1) {
      pf->pf_duration = mac_target_->usec(mac_target_->txopLimit_
                                          + mac_target_->txopDuration_
                                          - NOW
                                          + DSSS_SlotTime);
   }
   else {
      STORE4BYTE(&(mac_target_->index_), (pf->pf_ra));
      pf->pf_duration = 0;
   }

	pktPOLL_ = p;

#ifdef DEBUGACTIVE //=============================================================
	if(NOW > DEBUGTIME) {
		fprintf(macLog, "[%d.%d] send CF-Poll (duration: %d) to [%d.%d] ", mac_target_->index_, nq, pf->pf_duration, (int)ETHER_ADDR(pf->pf_ra), nq);
	}
#endif //=========================================================================

	// Set up the node for the Polled TXOP
	// setting the NAV to the end of the TXOP
	// and silencing all the Traffic Categories.
	// Note: this has already been done
	// at Beacon/CC transmission, but we provide
	// further protection this way
	if(mac_target_->cfb_ == -1) {
		mac_target_->set_nav(pf->pf_duration);
		for(int i=0; i<(mac_target_->num_tc); i++)
			mac_target_->tc[i]->stopTC();
	}

}

/*###################################################################
#                                                                   #
# cfbSendDATA_CFPOLL: prepares a QoS Data+CF-Poll for the           #
#                     current TC                                    #
#                                                                   #
#                     dst: the destination MAC address              #
#                     tc: the Traffic Category of Data              #
#                                                                   #
###################################################################*/

void Mac802_11eHC::cfbSendDATA_CFPOLL(int dst, int tc) {

	struct hdr_cmn* ch = HDR_CMN(pktTx_);
	struct hdr_mac802_11* dh = HDR_MAC802_11(pktTx_);
	
	ch->mac_tc_aux = (u_int8_t)tc;
	dh->dh_fc.fc_subtype = MAC_Subtype_QoS_Data_CFPoll;
	dh->dh_duration = mac_target_->usec(mac_target_->txopLimit_
                                            + mac_target_->txopDuration_
                                            - NOW
                                            + DSSS_SlotTime);

#ifdef DEBUGACTIVE //=============================================================
	if(NOW > DEBUGTIME) {
		fprintf(macLog, "[%d.%d] send Data+CF-Poll (duration: %d) to [%d.%d-%d] ", mac_target_->index_, nq, dh->dh_duration, dst, nq, tc);
	}
#endif //=========================================================================

	// Set up the node for the Polled TXOP
	// setting the NAV to the end of the TXOP
	// and silencing all the Traffic Categories.
	// Note: this has already been done
	// at beacon/CC transmission, but we provide
	// further protection this way
	mac_target_->set_nav(dh->dh_duration);
	for(int i=0; i<(mac_target_->num_tc); i++)
		mac_target_->tc[i]->stopTC();

}

/*###################################################################
#                                                                   #
# cfbSendDATA_CFACK_CFPOLL : prepares the Data+CF-Ack+CF-Poll frame #
#                            for transmission during the Contention #
#                            Free Burst                             #
#                                                                   #
#                            dst: the destination address           #
#                            tc: the Data destination TC            #
#                                                                   #
###################################################################*/

void Mac802_11eHC::cfbSendDATA_CFACK_CFPOLL(int dst, int tc) {

	struct hdr_cmn* ch = HDR_CMN(pktTx_);
	struct hdr_mac802_11* dh = HDR_MAC802_11(pktTx_);
	
	ch->mac_tc_aux = (u_int8_t)tc;
	dh->dh_fc.fc_subtype = MAC_Subtype_QoS_Data_CFAck_CFPoll;

	// Duration value during CFB
	// is the remaining TXOP time
	// (802.11e - 9.10.2.2)
	dh->dh_duration = mac_target_->usec(mac_target_->txopLimit_
	                                    + mac_target_->txopDuration_
					    - NOW);

#ifdef DEBUGACTIVE //=============================================================
	if(NOW > DEBUGTIME) {
		fprintf(macLog, "[%d.%d] send Data+CF-Ack+CF-Poll (duration: %d) to [%d.%d-%d] ", mac_target_->index_, nq, dh->dh_duration, dst, nq, tc);
	}
#endif //=========================================================================

}

/*###################################################################
#                                                                   #
# recvACK: called every time an ACK frame is received, this method  #
#          is responsable for resuming the transmission             #
#                                                                   #
#          p: the ACK frame                                         #
#                                                                   #
###################################################################*/

void Mac802_11eHC::recvACK(Packet *p) {

	struct hdr_cmn *ch = HDR_CMN(p);

	// If we didn't send a DATA frame, the ACK is discarded
	if(tx_state_ != MAC_SEND) {
	        mac_target_->discard(p, DROP_MAC_INVALID_STATE);
	        return;
	}
	
        /*
	 * The successful reception of this ACK packet implies
	 * that our DATA transmission was successful.  Hence,
	 * we can reset the Short/Long Retry Count and the CW.
	 */

	if( (u_int32_t)HDR_CMN(pktTx_)->size() <= macmib_->RTSThreshold )
		ssrc_ = 0;
	else
		slrc_ = 0;

	rst_cw();

	assert(pktTx_);
	Packet::free(pktTx_);
	pktTx_ = 0;       // The transmitted DATA frame can be discarded

	mhSend_.stop();

	assert(mhBackoff_.busy() == 0);

        if (ptosend == 0) {

		// Pending CFP or CAP [case ACKRECV].
		// Received an ACK & no more bursting: frame exchange terminated.
		// If we were waiting to start the CFP or CAP, it can start now
		if(mac_target_->cfpWaiting_ || mac_target_->capWaiting_) {
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) {
				fprintf(macLog, "[%d] ACK received, channel idle %f\n", mac_target_->index_, NOW);
			}
#endif //=========================================================================
			// wait PIFS
			if(mac_target_->cfpWaiting_)
				mac_target_->mhCfpStart_.start(mac_target_->pifs_);
			if(mac_target_->capWaiting_)
				mac_target_->mhCapStart_.start(mac_target_->pifs_);

			mac_target_->mac_log(p);
			// We are entering CFP or CAP, so we don't
			// resume transmission as usual but set
			// transmission to MAC_IDLE and then wait
			// for mhCfpStart_ or mhCapStart_ expiration
			set_tx_state(MAC_IDLE);
			return;
		}

	        mhBackoff_.start(cw_, is_idle());       // We start a new backoff before resuming transmission

        } else {        // Bursting modality

                mac_target_->txop = nq;              // This queue maintains the TXOP

                mac_target_->need_pkt(nq);           // We need the next frame from the Interface Queue for continuing the burst
                set_tx_state(MAC_IDLE);
        }

        if (ptosend == 0) {
	        tx_resume();        
        }
	mac_target_->mac_log(p);
}

/*###################################################################
#                                                                   #
# recvDATA: called every time a DATA frame is received, this method #
#           checks if it is possible to send the ACK frame, then it #
#           passes the DATA frame to the LLC layer.                 #
#                                                                   #
#           p: the DATA frame                                       #
#                                                                   #
###################################################################*/

void Mac802_11eHC::recvDATA(Packet *p) {
   struct hdr_mac802_11 *dh = HDR_MAC802_11(p);
   u_int32_t dst, src, size;

   struct hdr_cmn *ch = HDR_CMN(p);

   dst = ETHER_ADDR(dh->dh_da);
   src = ETHER_ADDR(dh->dh_sa);
   size = ch->size();

 	// Outside CFB, discard QoS Data packets
	if(mac_target_->cfb_ != nq &&
	   dh->dh_fc.fc_subtype == MAC_Subtype_QoS_Data){
		mac_target_->discard(p, "---");
		tx_resume();
		return;
	}

   // During CFB, discard Data packets
   if(mac_target_->cfb_ != -1  && dh->dh_fc.fc_subtype == MAC_Subtype_Data)
   {
      mac_target_->discard(p, "---");
      if(!mhSend_.busy() && !(mac_target_->mhCfp_.busy()))
      {
         set_tx_state(MAC_IDLE);
         mac_target_->cfbHandler();
      }
      return;
   }

   // XXX HCF v3.0 Start
   // Get from the station the remaining data for this TC
  if (((Mac802_11HC*)mac_target_)->scheduler == SCHED_ENHANCED ||
      (((Mac802_11HC*)mac_target_)->scheduler == SCHED_MMFAIR && !((Mac802_11HC*)mac_target_)->second_schedule))
  {
      int poll_index = ((Mac802_11HC*)mac_target_)->cfPollIndex_;
      int poll_number = ((Mac802_11HC*)mac_target_)->cfPollNum_;
      if ((poll_index >= 0) && (poll_index < poll_number))
        ((Mac802_11HC*)mac_target_)->getRemainingData(poll_index, ch->mac_tc, src);
  }
  else if (((Mac802_11HC*)mac_target_)->scheduler == SCHED_MMFAIR &&
           ((Mac802_11HC*)mac_target_)->second_schedule)
     ((Mac802_11HC*)mac_target_)->tspecs_[src].RemainingData -= size-52;  // 52 is the size of the headers
  // XXX HCF v3.0 End

   // XXX v2.0Start
   //Update CFB statistics
   if(mac_target_->cfb_ == nq) {

      switch(ch->ptype()) {
         case PT_TCP:
               cfb_data_fast_recv++;
               break;
         case PT_VOIP:
               cfb_voice_fast_recv++;
               break;
      }
   }
   // XXX v2.0End

   /*
    * Adjust the MAC packet size - ie; strip
    * off the mac header
    */
   ch->size() -= ETHER_HDR_LEN11;
   ch->num_forwards() += 1;

   /*
   *  If we sent a CTS, clean up...
   */
	if((dst != MAC_BROADCAST) && (mac_target_->cfb_ == -1)) {
		if(size >= macmib_->RTSThreshold && ch->burst_no_rts == 0) {
			if (tx_state_ == MAC_CTS) {
				assert(pktCTRL_);
				Packet::free(pktCTRL_); pktCTRL_ = 0;
				mhSend_.stop();

				/*
				 * Our CTS got through.
				 */

			} else {               // If the size of the received DATA frame is smaller
            ++busy_mac_ack;     // than the RTS_Threshold (or in bursting modality it
            mac_target_->discard(p, DROP_MAC_BUSY);    // is a frame following the first) and we didn't send
            return;             // a CTS, discard this frame without sanding an ACK
			}

			sendACK(src);           // Send the ACK frame
			tx_resume();
		}

		/*
		 *  We did not send a CTS and there's no
		 *  room to buffer an ACK.
		 */

		else {
			if(pktCTRL_ || tx_state_ != MAC_IDLE) {          // The MAC is busy, discard the frame
                                ++busy_mac_ack;
				mac_target_->discard(p, DROP_MAC_BUSY);
				return;
			}
			sendACK(src);   // Send the ACK frame

			if(mhSend_.busy() == 0) {
			        tx_resume();
                        }
		}
	}

	// During the CFB, prepare the response
	// and restart the CFB timer, then
	// invokes tx_resume, that resets the
	// tx_state_ and asks for new packets
	if(mac_target_->cfb_ == nq) {
		if(pktPOLL_) {
			Packet::free(pktPOLL_);
			pktPOLL_ = 0;
		}
		// Free CF-Ack if present
		if(pktCTRL_) {
			Packet::free(pktCTRL_);
			pktCTRL_ = 0;
		}
		mhSend_.stop();
		tx_resume();

		// Prepare response packet:
		// Data+CF-Poll+CF-Ack or Data+CF-Ack if there's
		// data to send to the polled QSTA, CF-Ack otherwise
		int dataTc = mac_target_->dataToSend(src, nq);
		if(dataTc != mac_target_->num_tc) {
			if(dh->dh_fc.fc_more_data == 1)
				cfbSendDATA_CFACK_CFPOLL(src, dataTc);
			else
				cfbSendDATA_CFACK(src, dataTc);
		}
		else cfbSendCFACK(src, mac_target_->num_tc);

		mac_target_->mhCfb_.start(mac_target_->sifs_);
	}
	
        mac_target_->upsend(p,nq);      // Pass the DATA frame to the LLC layer
}

/*###################################################################
#                                                                   #
# recvCFACK: called every time a QoS CF-ACK frame is received,      #
#            this method frees the packet that was transmitted      #
#            and stops its send timer, then prepares the response   #
#            QoS CF-Poll frame that ends the CFB and restarts the   #
#            CFB timer                                              #
#                                                                   #
#            p: the CF-Ack frame                                    #
#                                                                   #
###################################################################*/

void Mac802_11eHC::recvCFACK(Packet* p) {

	// Outside CFB, discard this packet
	if(mac_target_->cfb_ != nq) {
		mac_target_->discard(p, "---");
		tx_resume();
		return;
	}

	// If no data was sent discard CF-Ack
	// and resume CFB transmissions
	if(tx_state_ != MAC_QOS_DATA_CFPOLL &&
	   tx_state_ != MAC_QOS_DATA_CFACK &&
	   tx_state_ != MAC_QOS_DATA_CFACK_CFPOLL) {
	        mac_target_->discard(p, DROP_MAC_INVALID_STATE);
		if(!(mac_target_->mhCfb_.busy())) {
			// Stop the send timer
			if(mhSend_.busy())
				mhSend_.stop();
			set_tx_state(MAC_IDLE);
			mac_target_->mhCfb_.start(mac_target_->sifs_);
		}
	        return;
	}

	struct hdr_cmn *ch = HDR_CMN(p);
	int ackTc = ch->mac_tc_aux;

	// If the CF-Ack part of the received
	// frame is directed to a different TC,
	// free pktTx_ for such TC, and restore
	// the previous pktTx_ for this TC
	if(ackTc != nq &&
	   ackTc != mac_target_->num_tc) {
		Packet::free((mac_target_->tc[ackTc])->pktTx_);
		(mac_target_->tc[ackTc])->pktTx_ = 0;
		(mac_target_->tc[ackTc])->tx_resume();
		if(pktBackup_) {
			pktTx_ = pktBackup_;
			pktBackup_ = 0;
		}
		else
			pktTx_ = 0;
	}
	// If the CF-Ack part of the received
	// frame is directed to this TC, free pktTx_
	else {
		Packet::free(pktTx_);
		pktTx_ = 0;
	}

	// Stop the send timer
	mhSend_.stop();

	// Prepare the QoS CF-Poll
	// with duration/ID = 0
	struct hdr_mac802_11 * dh = HDR_MAC802_11(p);
	u_int32_t src = (u_int32_t)ETHER_ADDR(dh->dh_sa);
	cfbSendCFPOLL(src);

	// Ask for a new packet to send
	// and reset tx_state to MAC_IDLE
	tx_resume();

	// Restart the CFB timer
	mac_target_->mhCfb_.start(mac_target_->sifs_);

	// Free the received frame
	mac_target_->mac_log(p);

}

/*###################################################################
#                                                                   #
# recvDATA_CFACK: called every time a QoS Data+CF-ACK frame is      #
#                 received, this method is responsable for          #
#                 receiving the CF-Ack and preparing the answer     #
#                 to the Data with a CF-Ack, Data+CF-Ack or         #
#                 Data+CF-Ack+CF-Poll frame                         #
#                                                                   #
#                 p: the Data+CF-Ack frame                          #
#                                                                   #
###################################################################*/

void Mac802_11eHC::recvDATA_CFACK(Packet* p) {
	
	// Outside CFB, discard this packet
	if(mac_target_->cfb_ != nq) {
		mac_target_->discard(p, "---");
		tx_resume();
		return;
	}

   // extract common-header and data-header
   struct hdr_mac802_11 *dh = HDR_MAC802_11(p);
   struct hdr_cmn *ch = HDR_CMN(p);
   u_int32_t src = ETHER_ADDR(dh->dh_sa);
   u_int32_t size = ch->size();

   // XXX HCF v3.0 Start
   // Get from the station the remaining data for this TC
  if (((Mac802_11HC*)mac_target_)->scheduler == SCHED_ENHANCED ||
      (((Mac802_11HC*)mac_target_)->scheduler == SCHED_MMFAIR && !((Mac802_11HC*)mac_target_)->second_schedule))
   {
      int poll_index = ((Mac802_11HC*)mac_target_)->cfPollIndex_;
      int poll_number = ((Mac802_11HC*)mac_target_)->cfPollNum_;
      if ((poll_index >= 0) && (poll_index < poll_number))
         ((Mac802_11HC*)mac_target_)->getRemainingData(poll_index, ch->mac_tc, src);
   }
   else if (((Mac802_11HC*)mac_target_)->scheduler == SCHED_MMFAIR &&
            ((Mac802_11HC*)mac_target_)->second_schedule)
      ((Mac802_11HC*)mac_target_)->tspecs_[src].RemainingData -= size-52;  // 52 is the size of the headers
   //XXX HCF v3.0 End

	// Free CF-Ack if present
	if(pktCTRL_) {
		Packet::free(pktCTRL_);
		pktCTRL_ = 0;
	}

	// Handle CF-Ack reception
	int ackTc = ch->mac_tc_aux;

	// If the CF-Ack part of the received
	// frame is directed to a different TC,
	// free pktTx_ for such TC, and restore
	// the previous pktTx_ for this TC
	if(ackTc != nq &&
	   ackTc != mac_target_->num_tc) {
		Packet::free((mac_target_->tc[ackTc])->pktTx_);
		(mac_target_->tc[ackTc])->pktTx_ = 0;
		(mac_target_->tc[ackTc])->tx_resume();
		if(pktBackup_) {
			pktTx_ = pktBackup_;
			pktBackup_ = 0;
		}
		else
			pktTx_ = 0;
	}
	// If the CF-Ack part of the received
	// frame is directed to this TC, free pktTx_
	else {
		Packet::free(pktTx_);
		pktTx_ = 0;
	}

	// Stop the send timer
	mhSend_.stop();

	// Ask for a new packet to send
	// and reset tx_state to MAC_IDLE
	tx_resume();

	// Handle Data reception
	// XXX v2.0Start
	// Update CFB statistics
	if(mac_target_->cfb_ == nq) {
		switch(ch->ptype()) {
			case PT_TCP:
					cfb_data_fast_recv++;
				break;
			case PT_VOIP:
					cfb_voice_fast_recv++;
				break;
		}
	}
	// XXX v2.0End

	// Adjust the MAC packet size
	// i.e. strip off the mac header
	ch->size() -= ETHER_HDR_LEN11;
	ch->num_forwards() += 1;

	// Prepare response packet:
	// Data+CF-Poll+CF-Ack or Data+CF-Ack if there's
	// data to send to the polled QSTA, CF-Ack otherwise
	int dataTc = mac_target_->dataToSend(src, nq);
	if(dataTc != mac_target_->num_tc) {
		if(dh->dh_fc.fc_more_data == 1)
			cfbSendDATA_CFACK_CFPOLL(src, dataTc);
		else
			cfbSendDATA_CFACK(src, dataTc);
	}
	else cfbSendCFACK(src, mac_target_->num_tc);
	
        // Pass the Data frame to the LLC layer
	mac_target_->upsend(p, nq);

	// Restart the CFB timer
	mac_target_->mhCfb_.start(mac_target_->sifs_);

}

/*###################################################################
#                                                                   #
# recvNULL: called every time a QoS Null frame is received by HC,   #
#           this method prepares the QoS CF-Poll that ends the      #
#           Polled TXOP and then restarts the CFB timer             #
#                                                                   #
#           p: the Null frame                                       #
#                                                                   #
###################################################################*/

void Mac802_11eHC::recvNULL(Packet* p) {

	// Outside CFB, discard this packet
	if(mac_target_->cfb_ != nq) {
		mac_target_->discard(p, "---");
		tx_resume();
		return;
	}

   // XXX HCF v3.0 Start
   // Here is the check of how much time of the current
   // POLL-TXOP has not been used by the station; this
   // info is used to schedule next POLL-TXOP
   struct hdr_mac802_11 *NULL_dh = HDR_MAC802_11(p);
   u_int32_t src_addr = ETHER_ADDR(NULL_dh->dh_sa);

   if (((Mac802_11HC*)mac_target_)->scheduler == SCHED_ENHANCED)
   {
      double Null_time = mac_target_->txtime(ETHER_NULL_LEN2, mac_target_->basicRate_);
      double time_not_used = (mac_target_->txopLimit_ + mac_target_->txopDuration_) - NOW
                           + DSSS_MaxPropagationDelay + DSSS_SlotTime +
                           mac_target_->sifs_ + Null_time;

      // Note: update the NotUsedTime only if the RemainingData is equal to 0
      if (mac_target_->tspecs_[src_addr].RemainingData = 0)
         mac_target_->tspecs_[src_addr].TimeNotUsed = time_not_used;
      else
         mac_target_->tspecs_[src_addr].TimeNotUsed = 0.0;
      //if (NOW > 50)
      //fprintf(stderr, "Sta %d  Time not used: %f\n", src_addr, time_not_used);
   }
   // XXX HCF v3.0 Start

	// If the Null packet has been received
	// as a response to a QoS CF-Poll,
	// free pktPOLL_ and stop the send timer
	if(pktPOLL_) {
		struct hdr_mac802_11* dh = HDR_MAC802_11(pktPOLL_);
		if(dh->dh_fc.fc_type == MAC_Type_Data &&
		   dh->dh_fc.fc_subtype == MAC_Subtype_QoS_CFPoll) {
			Packet::free(pktPOLL_);
			pktPOLL_ = 0;
			mhSend_.stop();
		}
	}
	// Free CF-Ack if present
	if(pktCTRL_) {
		Packet::free(pktCTRL_);
		pktCTRL_ = 0;
	}

	// Prepare the QoS CF-Poll
	// with duration/ID = 0
	struct hdr_mac802_11 * dh = HDR_MAC802_11(p);
	u_int32_t src = (u_int32_t)ETHER_ADDR(dh->dh_sa);
	cfbSendCFPOLL(src);
	
	// Ask for a new packet to send
	// and reset tx_state to MAC_IDLE
	tx_resume();

	// Restart the CFB timer
	mac_target_->mhCfb_.start(mac_target_->sifs_);

	// Free the received frame
	mac_target_->mac_log(p);

}

/*###################################################################
#                                                                   #
# tx_resume: starts a new access procedure, according to the type   #
#            of the frame that will be sent. If a new frame is not  #
#            available, a request to the Interface Queue is         #
#            advanced. During CAP or CFP, it resets the tx_state_,  #
#            restart the CAP/CFP timer and possibly asks for new    #
#            data to send                                           #
#                                                                   #
###################################################################*/

void Mac802_11eHC::tx_resume() {

	// During the CAP the TC has to resume the Controlled
	// Access Period beahviour of the HC, restarting the
	// mhCap_ timer of the Mac802_11HC class.
	// It also resets the tx_state_ and possibly asks for
	// a new packet
	if((mac_target_->cap_ != -1) && (mac_target_->cfb_ == -1)) {
		// The CAP timer may be already busy
		// at this time due to channel errors:
		// reset the tx_state_ and return
		if((mac_target_->mhCap_).busy()) {
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) {
				fprintf(macLog, "[%d.%d] CAP timer running!\n", mac_target_->index_, nq);
			}
#endif //=========================================================================
			set_tx_state(MAC_IDLE);
			return;
		}
		// Start PIFS timer
		(mac_target_->mhCap_).start(mac_target_->pifs_);
		// No outgoing data packet:
		// ask for a new packet
		if(!pktTx_)
			mac_target_->need_pkt(nq);
		set_tx_state(MAC_IDLE);
		return;
	}

	// During the CFP the TC has to resume the transmission
	// restarting the mhCfp_ timer of the Mac802_11HC class.
	// It also resets the tx_state_ and possibly asks for
	// a new packet
	if(mac_target_->cfp_ && (mac_target_->cfb_ == -1)) {
		if((mac_target_->mhCfp_).busy()) {
#ifdef DEBUGACTIVE //=============================================================
			if(NOW > DEBUGTIME) {
				fprintf(macLog, "[%d.%d] CFP timer running!\n", mac_target_->index_, nq);
			}
#endif //=========================================================================
			(mac_target_->mhCfp_).stop();
		}
		// Start SIFS timer
		(mac_target_->mhCfp_).start(mac_target_->sifs_);
		// No outgoing data packet:
		// ask for a new packet
		if(!pktTx_)
			mac_target_->need_pkt(nq);
		set_tx_state(MAC_IDLE);
		return;
	}

	// During CFB or EDCF, use
	// superclass member function
	Mac802_11e::tx_resume();

}
//---------- HCFend

//----------------------------------------------------------------------------------------------------------------------------//EndEDCFpatch

